#!/usr/bin/python3
"""
 Operations on Autopilot Package Format. Both a tool and a library used.
 Cracks the APCK file, uses the libxdecode.so library to uncompress the chunks
 that are compressed using the Xpress+Huffman compression.
 libxdecode.so is a library ported from Windows for this project.
"""
import sys
import os
import struct
import argparse
import ctypes
import shutil

from tempfile import NamedTemporaryFile
from collections import namedtuple

# if script's parent directory isn't in sys.path, then add it
if '/'.join(sys.path[0].split("/")[:-1]) not in sys.path:
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apsdk import apenv


class ApPackReader(object):
    """
    Read and Extract an Autopilot package file.
    """
    FileInfo = namedtuple('FileInfo', ['packed_crc',
                                       'unpacked_crc',
                                       'last_write_time',
                                       'packed_size',
                                       'unpacked_size',
                                       'file_name',
                                       'offset'])

    _apck_header_v1 = b'Autopilot Pack v1.0'

    def __init__(self, apck_file_name: str):
        """
        Read and extract an Autopilot package.
        Throws OSError if the file is missing or have issue loading dependent libraries.
        """
        self.apck_file_name = apck_file_name

        # Open the file, let the caller handle the exceptions.
        self.pkg = open(apck_file_name, 'rb')

        # Read the APCK File Header
        # 32 - bytes
        # char header[19], char compression_type, UINT32 file_table_size, UINT64 file_table_crc
        header_buff = self.pkg.read(32)
        (header,
         compression_type,
         self.file_table_size,
         file_table_crc) = struct.unpack('=19scIQ', header_buff)

        if header != self._apck_header_v1:
            print("Autopilot Package Version not supported : ", header)

        if compression_type != b'x':
            print("Compression type not supported : ", compression_type)

        # TODO: Check that the file table has a valid structure by counting that the
        # number of bytes in the buffer matches

        self.file_info_list = self.__parse_file_table()
        if len(self.file_info_list) == 0:
            raise ValueError('Could not parse file table of apck file.')

        # Load the xdecode C library. This is a compression lib taken from the windows code base
        # and compiled into a c library. Let the caller handle the exceptions.
        self.xpdecode = ctypes.cdll.LoadLibrary(apenv.get_libxdecode_lib_path())

        # Create a 64k buffer(max uncompress size for xpdecode) for using during
        # uncompressing. We will keep reusing this while unpacking this pack.
        self.uncompress_buf = ctypes.create_string_buffer(b'\000' * 65536)

    def __parse_file_table(self) -> list:
        file_info_list = list()
        apck_header_size = 32

        # Move to the file table
        self.pkg.seek(apck_header_size)

        # Compressed file starts after header and file table
        file_offset_in_pkg = apck_header_size + self.file_table_size

        # Read till all the elements in the File Table is read
        while self.pkg.tell() < apck_header_size + self.file_table_size:
            # Buffer: <ApPackSingleFileHeader><size_of_file_name><file_name><1 byte>
            # ApPackSingleFileHeader = UNIT64 packed_crc, unpacked_crc, last_write_time,
            #                          packed_size, unpacked_size
            (packed_crc,
             unpacked_crc,
             last_write_time,
             packed_size,
             unpacked_size,
             file_name_size) = struct.unpack('=QQQQQb',
                                             self.pkg.read(40 + 1))

            format_str = '=' + str(file_name_size) + 'sx'

            (file_name,) = struct.unpack(format_str,
                                         self.pkg.read(file_name_size + 1))

            print("--->", file_name.decode('ascii'))
            file_info = self.FileInfo(file_name=file_name.decode('ascii'),
                                      packed_crc=packed_crc,
                                      unpacked_crc=unpacked_crc,
                                      last_write_time=last_write_time,
                                      packed_size=packed_size,
                                      unpacked_size=unpacked_size,
                                      offset=file_offset_in_pkg)

            file_info_list.append(file_info)

            # Move the file pointer to the next file offset
            file_offset_in_pkg = file_offset_in_pkg + packed_size

        return file_info_list

    def __get_file_info(self,
                        file_name: str) -> FileInfo:
        for file_info in self.file_info_list:
            if file_info.file_name == file_name:
                return file_info
        return None

    def __uncompress_chunk(self,
                           chunk_compressed_size,
                           chunk_uncompressed_size,
                           chunk_uncompressed_crc,
                           chunk,
                           target_buf) -> int:
        cmp_buf = ctypes.create_string_buffer(chunk)
        try:
            num_bytes = self.xpdecode.XPressHuffmanDecode(ctypes.byref(cmp_buf),
                                                          chunk_compressed_size,
                                                          ctypes.byref(target_buf),
                                                          chunk_uncompressed_size)
        except (OSError, ValueError) as e:
            print("Exception calling XPressHuffmanDecode.", str(e))
            return None

        if num_bytes != chunk_uncompressed_size:
            return None
        else:
            return num_bytes

    # Extract a file from the package.
    # File is stored as chunks with format:
    #  <chunk_header><chunk>
    #  <chunk_header><chunk>
    #  <chunk_header><chunk>
    #  ......
    def extract_file(self,
                     file_name: str,
                     target_dir: str) -> str:
        """
        Extract an apck file in a target directory specified.
        except: IOError, OSError
        """
        if not os.path.isdir(target_dir):
            print(target_dir, "Does not exist.")
            return None

        # TODO: CRC check
        chunk_header_size = 16

        file_info = self.__get_file_info(file_name)
        if file_info is None:
            print("File Not found in package ")
            return None

        # Get the file full path, the file will come in with Windows file path
        # structure if it is under a sub directory.
        file_full_path = os.path.join(target_dir, file_name.replace('\\', '/'))

        # If this file is to be under a directory, create them
        dir_name = os.path.dirname(file_full_path)
        if not os.path.isdir(dir_name):
            os.makedirs(dir_name)

        # Open a temporary file to unpack into
        tmp_file = NamedTemporaryFile(mode='w+b',
                                      dir=target_dir,
                                      prefix=file_name,
                                      delete=False)

        # Go to the file offset
        self.pkg.seek(file_info.offset)

        # Read till we don't go over the file's last byte in the pack
        while self.pkg.tell() < file_info.offset + file_info.packed_size:
            # Read chunk header
            # TODO: Check chunk size, CRC etc.,
            (chunk_compressed_size,
             chunk_uncompressed_size,
             chunk_uncompressed_crc) = struct.unpack('=IIQ',
                                                     self.pkg.read(chunk_header_size))

            # Read the chunk
            chunk_buf = self.pkg.read(chunk_compressed_size)
            if chunk_compressed_size == chunk_uncompressed_size:
                print(file_name, "chunk isn't compressed.")
                tmp_file.write(chunk_buf)
            else:
                num_bytes_decoded = self.__uncompress_chunk(chunk_compressed_size,
                                                            chunk_uncompressed_size,
                                                            chunk_uncompressed_crc,
                                                            chunk_buf,
                                                            self.uncompress_buf)
                if num_bytes_decoded is None:
                    print("ERROR: A chunk could not be uncompressed for file:",
                          file_name)
                    tmp_file.delete = True
                    tmp_file.close()
                    break
                else:
                    tmp_file.write(self.uncompress_buf.raw[:num_bytes_decoded])

        # If file is closed, then there was error unpacking
        if tmp_file.closed is True:
            print("Error unpacking :", file_name)
        else:
            tmp_file.close()
            # We uncompressed to the temporary file. Now rename it.
            if os.path.isfile(file_full_path):
                shutil.os.remove(file_full_path)
            shutil.move(tmp_file.name, file_full_path)

        return file_full_path

    def extract_all(self,
                    target_dir: str) -> list:
        """
        Extract all packages in a directory and returns the list.
        except: IOError, OSError
        """
        extracted_file_list = []

        # Extract all the files in the package
        for file_info in self.file_info_list:
            file_full_path = self.extract_file(file_info.file_name, target_dir)

            if file_full_path is not None:
                extracted_file_list.append(file_full_path)
            else:
                print("File extraction failed: ", file_info)

        return extracted_file_list


def main():
    parser = argparse.ArgumentParser(description="Unpack an APCK file.")
    parser.add_argument('apck', metavar='\"APCK\"', help='APCK File')
    parser.add_argument('dir', metavar='\"DIR\"', help='Target Dir')
    parser.add_argument('--file', '-f', metavar='FILE', help='Target File', default=None)
    parser.add_argument('--all', '-a', metavar='ALL', help='Extract All', default=None)

    args = parser.parse_args()
    apck = args.apck
    target_dir = args.dir
    file = args.file
    all_args = args.all

    reader = ApPackReader(apck)
    if file is None:
        if all_args is None:
            parser.print_usage()
        else:
            import time
            time1 = time.time()
            file_list = reader.extract_all(target_dir)
            print('Extracted File List:', file_list)
            timing = (time.time() - time1) * 1000
            print("Extraction took %0.3f ms" % timing)
    else:
        reader.extract_file(file, target_dir)


if __name__ == '__main__':
    main()
