#!/usr/bin/env python3

"""
Read and cache APSecret KeyVault secrets and keys.
"""

from http.client import HTTPSConnection
import sys
import os
import stat
import ssl
import json

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apcommon.keyvaulthelper import KeyVaultHelper


class APSecretClient:
    APSECRET_SECRET_DIR = "secrets"
    ENDPOINT_TIMEOUT = 5
    APSECRET_RETRIES = 2

    def __init__(self,
                 client_cert_file,
                 client_key_file,
                 endpoints,
                 apsecret_cache="/var/cache/apsecret"):
        self.client_key_file = client_key_file
        self.client_cert_file = client_cert_file
        self.apsecret_endpoints = endpoints
        self.apsecret_cache = apsecret_cache

        self.keyvault_config = ""
        self.tenant = ""
        self.client_id = ""
        self.client_key = ""
        self.vault_uri = ""

    def is_connection_data_initialized(self) -> bool:
        """
        Checks to see if we have connection data needed to access KeyVault.
        :return: Returns True if the KeyVault connection info is present and otherwise False.
        """
        return bool(self.tenant and self.client_id and self.client_key and self.vault_uri)

    def __get_keyvault_connection_json(self):
        """
        Gets the connection data needed to connect to APLinux Prod's KeyVault instance for the
        purpose of accessing bootstrap secrets.
        """
        if self.client_key_file is None or self.client_cert_file is None:
            raise ValueError("Client certificate is required")

        # configure certificate verification
        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

        # retrieve the json config
        eps = self.apsecret_endpoints
        if type(eps) is str:
            endpoints = [eps]
        elif type(eps) is list:
            endpoints = eps
        else:
            raise ValueError("Endpoint parameter must be a str or a list of str")

        for endpoint in endpoints:
            for retry in range(0, self.APSECRET_RETRIES):
                try:
                    conn = HTTPSConnection(endpoint,
                                           key_file=self.client_key_file,
                                           cert_file=self.client_cert_file,
                                           context=ctx,
                                           timeout=self.ENDPOINT_TIMEOUT)
                    conn.putrequest("GET", "/keyvault/aplinux")
                    conn.endheaders()
                    response = conn.getresponse()
                except ssl.SSLError as e:
                    if str(e).startswith("[SSL] PEM lib"):
                        raise NotImplementedError("Invalid client cert format")
                    print("ERROR: SSL error connecting to the APSecret service "
                          "endpoint '{0}': {1}. Trying next endpoint...".format(endpoint, e),
                          file=sys.stderr)
                    continue
                except Exception as e:
                    print("ERROR: Unable to connect to the APSecret service "
                          "endpoint '{0}': {1}. Trying next endpoint...".format(endpoint, e),
                          file=sys.stderr)
                    continue

                if response.status != 200:
                    print("WARNING: APSecret service endpoint '{0}' returned with "
                          "an error: {1} {2}. Trying next endpoint...".format(endpoint,
                                                                              response.status,
                                                                              response.reason),
                          file=sys.stderr)
                    continue

                try:
                    data = response.read()
                    conn.close()
                except Exception as e:
                    print("ERROR: Connection not available to APSecret service "
                          "endpoint '{0}': {1}. Trying next endpoint...".format(endpoint, e),
                          file=sys.stderr)
                    continue

                # successfully connected
                if data:
                    break

        if not data:
            raise ConnectionError("Unable to connect to any working APSecret service endpoints")

        self.keyvault_config = data.decode(encoding="UTF-8")
        self.__parse_keyvault_connection_json()

    def __parse_keyvault_connection_json(self):
        """
        Parses the json received from APSecret containing the KeyVault connection information.
        """
        # parse the json config file
        try:
            config = json.loads(self.keyvault_config)
        except:
            raise ValueError("Connection information retrieved from authority is in a unsupported format.")

        if config is None or len(config) == 0:
            raise ValueError("No connection information available from authority.")
        if [o for o in ("TenantID", "ClientID", "ClientSecret", "VaultURI") if o not in config[0]]:
            raise ValueError("Connection information retrieved from authority is incomplete.")

        self.tenant = config[0]["TenantID"]
        self.client_id = config[0]["ClientID"]
        self.client_key = config[0]["ClientSecret"]
        self.vault_uri = config[0]["VaultURI"]

        if not self.is_connection_data_initialized():
            raise ValueError("Connection information retrieved from authority is invalid.")

    def __create_directory_layout(self):
        """
        Creates the directory structure for the cache.
        """
        if not os.path.isdir(self.apsecret_cache):
            try:
                os.makedirs(self.apsecret_cache, 0o750, exist_ok=True)
            except Exception as e:
                raise IOError("Unable to create directory layout of cache: mkdir recursive {0}: {1}".format(self.apsecret_cache, e))

        subdirectory = os.path.join(self.apsecret_cache, self.APSECRET_SECRET_DIR)
        try:
            if not os.path.isdir(subdirectory):
                os.mkdir(subdirectory, 0o750)
        except Exception as e:
            raise IOError("Unable to create directory layout of cache: mkdir {0}: {1}".format(subdirectory, e))

    def __secure_cache_file(self, filepath: str):
        """
        Changes the permissions of the secret file to protect it from unauthorized access.
        """
        try:
            os.chmod(filepath, stat.S_IRWXU | stat.S_IRGRP)
        except Exception as e:
            raise IOError("Failed secure permissions of cache file {0}: {1}".format(filepath, e))

    def get_keyvault_secret(self, name: str) -> str:
        """
        Connects to KeyVault to retrieve the specified secret. Returns the secret.
        :return: Returns the secret value as a string.
        """
        if not name:
            raise ValueError("Get keyvault secret: Empty name")

        kv = KeyVaultHelper(self.tenant, self.client_id, self.client_key, self.vault_uri)

        try:
            kv.connect()
        except Exception as e:
            raise IOError("Connection failed to KeyVault: {0}".format(e))

        try:
            secret = kv.get_secret(name)
        except Exception as e:
            raise IOError("Failed to get secret '{0}' from KeyVault: {1}".format(name, e))

        return secret

    def __update_cache_secret(self, name: str):
        """
        Updates a locally cached secret.
        """
        if not self.is_connection_data_initialized():
            self.__get_keyvault_connection_json()

        try:
            secret = self.get_keyvault_secret(name)
        except Exception as e:
            raise IOError("Updating secret cache: {0}".format(e))

        # Write the secret to file and secure it
        self.__create_directory_layout()
        filename = os.path.join(self.apsecret_cache, self.APSECRET_SECRET_DIR, name)
        try:
            f = open(filename, "w+")
            f.write(secret)
            f.close()
            self.__secure_cache_file(filename)
        except Exception as e:
            raise IOError("Updating secret cache: Writing secret to cache file {0}: {1}".format(filename, e))

    def read_cache_secret(self, name: str) -> str:
        """
        Gets the value of the specified locally cached secret.
        :return: Returns the secret on success and an exception on error.
        """
        if not name:
            raise ValueError("Read secret cache: Empty name")

        try:
            self.__update_cache_secret(name)
        except Exception as e:
            print("WARNING: Failed to update cache of secret '{0}': {1}".format(name, e), file=sys.stderr)

        filename = os.path.join(self.apsecret_cache, self.APSECRET_SECRET_DIR, name)
        if not os.path.isfile(filename):
            raise KeyError("Read secret cache: '{0}' not cached".format(name))

        try:
            f = open(filename, "r")
            secret = f.read()
            f.close()
        except Exception as e:
            raise IOError("Read secret cache: Unable to read '{0}' from cache: {1}".format(name, e))

        return secret
