#!/usr/bin/env python3

import argparse
import platform
import os.path

if platform.system() == "Linux":
    import apt


# Return if package is installed or not
def apt_get_is_installed(pkg_name, pkg_ver=None):
    apt_cache = apt.cache.Cache()

    if pkg_name not in apt_cache:
        # Let's update the cache and try once more
        apt_cache.update()
        if pkg_name not in apt_cache:
            print("No such package exists :", pkg_name)
            return False

    apt_pkg = apt_cache[pkg_name]

    # Pin the specific version available
    if pkg_ver is not None:
        if pkg_ver in apt_pkg.versions:
            apt_pkg.candidate = apt_pkg.versions[pkg_ver]
        else:
            print("Requested version could not be found for pkg", pkg_name, pkg_ver)
            return False

    if apt_pkg.is_installed:
        print('{0} is installed'.format(pkg_name))
        return True
    else:
        print('{0} is not installed'.format(pkg_name))
        return False


# Return installed package version
def apt_get_version(pkg_name):
    apt_cache = apt.cache.Cache()

    if pkg_name not in apt_cache:
        # Let's update the cache and try once more
        apt_cache.update()
        if pkg_name not in apt_cache:
            print("No such package exists :", pkg_name)
            return ""

    apt_pkg = apt_cache[pkg_name]

    installed_version = apt_pkg.installed
    if installed_version:
        return installed_version.version
    else:
        return ""


def apt_get_install(pkg_name, pkg_ver=None):
    apt_cache = apt.cache.Cache()

    # TODO: Handle dependencies
    if pkg_name not in apt_cache:
        # Let's update the cache and try once more
        apt_cache.update()
        if pkg_name not in apt_cache:
            print("No such package exists :", pkg_name)
            return False

    apt_pkg = apt_cache[pkg_name]

    # Pin the specific version available
    if pkg_ver is not None:
        if pkg_ver in apt_pkg.versions:
            apt_pkg.candidate = apt_pkg.versions[pkg_ver]
        else:
            print("Requested version could not be found for pkg", pkg_name, pkg_ver)
            return False

    # TODO: Handle dependencies
    if apt_pkg.is_installed:
        print('{0} already installed'.format(pkg_name))
        return True
    else:
        apt_pkg.mark_install(auto_fix=True)

        try:
            print("Installing ", pkg_name, pkg_ver)
            apt_cache.commit()
        except Exception as arg:
            print("ERROR: Package installation failed :", str(arg))
            return False
        else:
            return True


def apt_get_upgrade(pkg_name):
    apt_cache = apt.cache.Cache()

    # TODO: Handle dependencies
    if pkg_name not in apt_cache:
        # Let's update the cache and try once more
        apt_cache.update()
        if pkg_name not in apt_cache:
            print("No such package exists :", pkg_name)
            return False

    apt_pkg = apt_cache[pkg_name]

    apt_pkg.mark_upgrade()

    try:
        print("Upgrading ", pkg_name)
        apt_cache.commit()
    except Exception as arg:
        print("ERROR: Package upgrade failed :", str(arg))
        return False
    else:
        return True


def apt_get_remove(pkg_name):
    apt_cache = apt.cache.Cache()

    # TODO: Handle dependencies
    if pkg_name not in apt_cache:
        # Let's update the cache and try once more
        apt_cache.update()
        if pkg_name not in apt_cache:
            print("No such package exists :", pkg_name)
            return False

    apt_pkg = apt_cache[pkg_name]

    if not apt_pkg.is_installed:
        print('{0} not installed'.format(pkg_name))
        return 0
    else:
        apt_pkg.mark_delete(purge=True)

        try:
            apt_cache.commit()
        except Exception as arg:
            print("ERROR: Package installation failed :", str(arg))
            return False
        else:
            return True


def path():
    """ Helper method to provide path to this file for copying to container """
    # Total hack...
    return os.path.realpath(__file__)

if __name__ == '__main__':
    """ Install packages from apt cache or remove installed package. """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--version', dest='version',
        help='Version of package to install')
    parser.add_argument(
        '--install', dest='install',
        help='install package')
    parser.add_argument(
        '--remove', dest='remove',
        help='install package')

    args = parser.parse_args()

    if args.install is not None:
        if not apt_get_install(args.install, args.version):
            exit(1)
    elif args.remove is not None:
        if not apt_get_remove(args.remove):
            exit(1)
