#!/usr/bin/env python3

"""
Interact with the Docker service.
"""

import subprocess


class DockerHelper:
    def __init__(self):
        self.has_initialized = False
        self.network_inspect_bridge_raw = ""
        self.docker_bridge_address = ""

    def __is_initialized(self) -> bool:
        """
        Determine of the __initialize function has been called yet.
        :return: Returns True if the __initialize function has been called, otherwise False.
        """
        return self.has_initialized

    def __initialize(self):
        """
        Initializes all attributes relating to this class.
        """
        self.has_initialized = True
        try:
            self.__inspect_docker_bridge()
            self.__parse_network_inspect_bridge()
        except Exception as e:
            raise RuntimeError("Unable to get docker bridge address: {0}".format(e))

    def __inspect_docker_bridge(self):
        """
        Captures the output of the 'docker network inspect bridge' command.
        """
        try:
            # Returns the running config of the docker bridge in json format
            command = "docker network inspect bridge"
            self.network_inspect_bridge_raw = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired as e:
            raise RuntimeError("Inspect docker: '{0}' execution timed out".format(command))
        except subprocess.CalledProcessError as e:
            raise RuntimeError("Inspect docker: Calling '{0}': {1}".format(command, e))

    def __parse_network_inspect_bridge(self):
        """
        Parses the output of the 'docker network inspect bridge' command and saves interesting
        fields to class attributes.
        """
        for line in self.network_inspect_bridge_raw.strip().split("\n"):
            if 'Gateway' in line:
                field = line.split()
                # Expecting "Gateway": "192.168.43.1"
                if len(field) < 2 or len(field[1]) < 2:
                    raise SyntaxError("Unexpected output")
                self.docker_bridge_address = field[1][1:field[1].rfind('"')]

    def get_docker_bridge(self) -> str:
        """
        Get the docker bridge address.
        :return: The IPv4 address of the docker bridge as a string.
        """
        # Initialize if needed
        if not self.__is_initialized():
            self.__initialize()

        return self.docker_bridge_address
