#!/usr/bin/env python3

"""
Helper functions for reading KeyVault vault contents.
"""

from azure.common.credentials import ServicePrincipalCredentials
from azure.keyvault import KeyVaultClient


class KeyVaultHelper:
    def __init__(self, tenant_id: str, client_id: str, client_secret: str, vault_uri: str):
        self.client_id = client_id
        self.client_secret = client_secret
        self.tenant_id = tenant_id
        self.vault_uri = vault_uri
        self.client = None

    def connect(self):
        try:
            # create the client
            creds = ServicePrincipalCredentials(self.client_id,
                                                self.client_secret,
                                                tenant=self.tenant_id,
                                                resource='https://vault.azure.net')  # Must be https://vault.azure.net
            self.client = KeyVaultClient(creds)
        except Exception as e:
            raise Exception("Error creating KeyVault client: {0}".format(e))

    def list_key_versions(self, name: str) -> list:
        """
        Returns a list of versions available for the specified key.
        :param name: Name of the key.
        :return: A list of versions.
        """
        if not name:
            raise ValueError("List key versions: Empty name")

        versions = self.client.get_key_versions(self.vault_uri, name)

        # transform 'https://apsecretvault.vault.azure.net/keys/NotAKey/5a4864beb9884dfaa87c2b47ca97917a'
        # into '5a4864beb9884dfaa87c2b47ca97917a'
        return [key_id[key_id.rfind('/') + 1:] for key_id in [version.kid for version in versions]]

    def list_secret_versions(self, name: str) -> list:
        """
        Returns a list of versions available for the specified secret.
        :param name: Name of the secret.
        :return: A list of versions.
        """
        if not name:
            raise ValueError("List secret versions: Empty name")

        versions = self.client.get_secret_versions(self.vault_uri, name)

        # transform 'https://apsecretvault.vault.azure.net/secrets/NotASecret/5a4864beb9884dfaa87c2b47ca97917a'
        # into '5a4864beb9884dfaa87c2b47ca97917a'
        return [secret_id[secret_id.rfind('/') + 1:] for secret_id in [version.id for version in versions]]

    def get_secret(self, name: str, version=None) -> str:
        """
        Returns the value of the specified KeyVault secret or an empty string if the secret doesn't exist.
        :param name: Name of the secret to retrieve.
        :param version: The version of the secret. Specify None to retrieve the latest.
        :return: The value of the secret or an empty string if the secret doesn't exist.
        :rtype: str
        """
        if not name:
            raise ValueError("Get secret: Empty name")

        key_id = '{}/secrets/{}'.format(self.vault_uri, name)
        if version:
            key_id = '{}/{}'.format(key_id, version)

        s = self.client.get_secret(key_id)
        return s.value
