#!/usr/bin/env python3

import sys
import os
import parted

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apcommon import shellprocess


class Partitioning(object):
    available_partition_types = ["JBOD", "RAID-0", "RAID-5"]
    block_devices = {}

    def __init__(self, disk_config_type, force_format=False):
        # TODO: Check input of disk_config_type against self.available_partition_types[]
        self.shellprocess = shellprocess.Shellprocess()
        self.disk_config_type = disk_config_type
        self.force_format = force_format

    def find_root_device(self):
        with open('/proc/self/mounts', 'r') as f:
            for line in f:
                if (line.split()[1] == "/"):
                    full_root_device = (line.split()[0])
                    break

        # command = "awk '$2 == \"/\"' /proc/self/mounts | awk {'print $1'}"
        # result = self.shellprocess.run_shell_cmd_and_get_result(command)

        try:
            self.root_device = full_root_device.split('/')[2][0:3]

        except Exception as e:
            print("\npartitioning: Unable to split string to obtain root block device")
            print("partitioning: Exception: " + str(e))
            sys.exit(os.EX_SOFTWARE)  # EX_SOFTWARE = an internal software error was detected

        print("\npartitioning: Root block device is: " + str(self.root_device))

    def check_root_device(self):
        # TODO: This test is useless, replace with checking against result of fstab and UUID link to device name
        command = "mount | awk '$3 == \"/\" {print $1}'"
        result = self.shellprocess.run_shell_cmd_and_get_result(command)

        if (not result[0]):
            sys.exit(os.EX_IOERR)  # an error occurred while doing I/O on some file

        if not (result[1].strip()):
            sys.exit(os.EX_OSFILE)  # EX_OSFILE = some system file did not exist, could not be opened, or had some other kind of error
        else:
            try:
                test_root_device = result[1].splitlines()[0].split('/')[2][0:3]

            except Exception as e:
                print("\npartitioning: Unable to split string to obtain root block device")
                print("partitioning: Exception: " + str(e))
                sys.exit(os.EX_OSFILE)  # EX_OSFILE = some system file did not exist, could not be opened, or had some other kind of error

        # Compare /proc/self/mounts to mount to make sure device is the same
        if self.root_device != test_root_device:
            print("\npartitioning: Unable to match /proc/self/mounts result to mount result for root block device")
            sys.exit(os.EX_OSFILE)  # EX_OSFILE = some system file did not exist, could not be opened, or had some other kind of error

        # Get root UUID from /etc/fstab to verify root is correct
        command = "awk '$2 == \"/\" {print $1}' /etc/fstab | sed 's/UUID=//'"
        result = self.shellprocess.run_shell_cmd_and_get_result(command)

        if not (result[1].strip()):
            print("\npartitioning: Unable to get UUID from /etc/fstab for root block device")
            sys.exit(os.EX_OSFILE)  # EX_OSFILE = some system file did not exist, could not be opened, or had some other kind of error

        else:
            self.root_uuid = result[1].strip()

        # Check if root UUID exists in /dev/disk/by-uuid
        command = "if [ -e /dev/disk/by-uuid/" + self.root_uuid + " ]; then echo \"found\"; fi"
        result = self.shellprocess.run_shell_cmd_and_get_result(command)

        if (result[1].strip() != "found"):
            print("\npartitioning: Unable to match /dev/disk/by-uuid to UUID from /etc/fstab for root block device")
            sys.exit(os.EX_OSFILE)  # EX_OSFILE = some system file did not exist, could not be opened, or had some other kind of error

    def discover_block_devices(self):
        command = "lsblk -S -n -o name"
        result = self.shellprocess.run_shell_cmd_and_get_result(command)
        self.raw_devices = result[1].splitlines()

        # Output discovered block devices
        print("\npartitioning: Found devices:")
        for device in self.raw_devices:
            print("\t" + device)

    def configure_block_devices(self):
        # Determine list of labeled devices
        command = "ls /dev/disk/by-label | grep -E 'HDD|SSD'"
        result = self.shellprocess.get(command)

        acquired_labels = result[1].splitlines()

        labels = {}

        for label in acquired_labels:
            command = "readlink -f /dev/disk/by-label/" + label
            result = self.shellprocess.run_shell_cmd_and_get_result(command)

            if not result[1]:
                print("\npartitioning: Label has no corresponding device symlink!")
                sys.exit(os.EX_OSFILE)  # EX_OSFILE = some system file did not exist, could not be opened, or had some other kind of error
            else:
                linked_device = result[1].split('/')[2][0:3]
                labels[linked_device] = label

        for device in self.raw_devices:
            if (device != self.root_device):
                # Determine whether the device is an SSD or an HDD by probing the rotational parameter
                command = "cat /sys/block/" + device + "/queue/rotational"
                print("\npartitioning: Getting device type information on /dev/" + device)

                result = self.shellprocess.run_shell_cmd_and_get_result(command)
                device_type_value = result[1].splitlines()[0]

                if not device_type_value:
                    print("\npartitioning: Unable to determine HDD vs SSD for device: " + device)
                    sys.exit(os.EX_OSFILE)  # EX_OSFILE = some system file did not exist, could not be opened, or had some other kind of error

                if (device_type_value == "1"):
                    device_type = "HDD"
                else:
                    device_type = "SSD"

                # Determine if the device is already mounted and, if so, where
                command = "mount | awk '$1 ~ /\/dev\/sda/ {print $1" + "\" \"$3}'"
                result = self.shellprocess.run_shell_cmd_and_get_result(command)

                acquired_mounts = result[1].splitlines()

                mounted_locations = []

                for line in acquired_mounts:
                    mounted_locations.append(line.split(' ')[1])

                # Configure Devices with all their associated options
                config_options = {"device_type": device_type, "mounted_location": ','.join(mounted_locations),
                                  "label": labels[device]}
                self.block_devices[device] = config_options

        # Output list of configured devices and their options
        print("\npartitioning: Configured devices:")

        for device, options in sorted(self.block_devices.items()):
            print("\t" + device)
            if isinstance(options, dict):
                for option, value in sorted(options.items()):
                    print("\t\t" + option + " : " + value)

                sys.exit(os.EX_OK)  # EX_OK = everything executed normally

    # def remove_root_from_block_devices(self):
    #     try:
    #         del self.block_devices[self.root_device]
    #     except KeyError:
    #         pass

    def get_block_devices(self):
        return self.block_devices

    def get_disk_config_type(self):
        return self.disk_config_type

    def get_root_device(self):
        return self.root_device

    def sync_disk(self, device):
        # Required to force OS to sync disks so that changes to the drives/partitions will be written immediately
        command = "blockdev --rereadpt /dev/" + device
        print("\npartitioning: Syncing disk /dev/" + device)
        self.shellprocess.run_shell_cmd(command)

    def allocate_disks(self):
        if (self.disk_config_type == "JBOD"):
            for device, device_type in sorted(self.block_devices.items()):
                if (device != self.root_device):
                    current_device = parted.getDevice("/dev/" + device)
                    current_device.clobber()

                    disk = parted.freshDisk(current_device, 'gpt')

                    geometry = parted.Geometry(device=current_device, start=1, length=current_device.getLength() - 1)
                    filesystem = parted.FileSystem(type='ext4', geometry=geometry)
                    partition = parted.Partition(disk=disk, type=parted.PARTITION_NORMAL, fs=filesystem, geometry=geometry)

                    disk.addPartition(partition=partition, constraint=current_device.optimalAlignedConstraint)

                    print("\npartitioning: Creating GPT drive on /dev/" + device)
                    print("\npartitioning: Creating partition on /dev/" + device)
                    disk.commit()

                    self.sync_disk(device)

    def format_partitions(self):
        if (self.disk_config_type == "JBOD"):
            for device, device_type in sorted(self.block_devices.items()):
                if (device != self.root_device):
                    command = "mkfs.ext4 -F /dev/" + device + "1"
                    print("\npartitioning: Formatting partition /dev/" + device + "1")
                    self.shellprocess.run_shell_cmd(command, 30)

    def label_partitions(self):
        if (self.disk_config_type == "JBOD"):
            hdd = 1
            ssd = 1

            for device, device_type in sorted(self.block_devices.items()):
                if (device != self.root_device):
                    if (device_type == "HDD"):
                        label_integer = hdd
                        hdd += 1
                    else:
                        label_integer = ssd
                        ssd += 1

                    command = "e2label /dev/" + device + "1 " + device_type + str(label_integer)
                    print("\npartitioning: Labeling partition /dev/" + device + "1 with Label: " + device_type + str(label_integer))
                    self.shellprocess.run_shell_cmd(command)

    def main(self):
        if self.disk_config_type in self.available_partition_types:
            self.find_root_device()
            self.check_root_device()
            self.discover_block_devices()
            self.configure_block_devices()
            self.allocate_disks()
            self.format_partitions()
            self.label_partitions()

            self.set_partitioning_executed()


if __name__ == '__main__':
    Partitioning.main()
