#!/bin/bash
# Start apnginx container

set -exou pipefail

# Get script dir
SCRIPT_DIR=$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )

# Get the tag from version file.
source ${SCRIPT_DIR}/apnginx.ver

# Set variables. 
# APDOCKER_DOMAIN is set in from the systemd .service file
APLINUX_REGISTRY="aplinux-microsoft.azurecr.io${APDOCKER_DOMAIN}"
APNGINX_IMAGE=${APLINUX_REGISTRY}/apnginx:${APNGINX_IMAGE_TAG}
AP_TOOLS_DIR="/ap/app/aptools"

# Hide ACR creds from logs
set +x

# Pull a fresh ACR connection config every start
# We'll only capture stdout to ACR_CONNECTION_STRING, which will be empty on failure
ACR_CONNECTION_STRING=$(${AP_TOOLS_DIR}/apsecret_refresh_acr.py)
if [ $? -ne 0 ]
then
    echo "WARNING: Could not grab the most recent connection configuration for ACR."
fi

if [ -z "${ACR_CONNECTION_STRING}" ]
then
    echo "ERROR: Could not read the cached connection configuration for ACR."
    exit -1
fi

# Turn the connection string into an array
ACR_CONFIG=( ${ACR_CONNECTION_STRING} )
ACR_CLIENT_ID="${ACR_CONFIG[0]}"
ACR_CLIENT_KEY="${ACR_CONFIG[1]}"

if [ -z "${ACR_CLIENT_ID}" ] || [ -z "${ACR_CLIENT_KEY}" ]
then
    echo "ERROR: Invalid connection configuration cached for ACR."
    exit -1
fi

set -x

# Don't bother logging in and pulling the image if its already been done
if [[ "$(docker images -q ${APNGINX_IMAGE} 2> /dev/null)" == "" ]]
then
    # Login to aplinux registry and pull the image
    set +x
    docker login -u ${ACR_CLIENT_ID} -p ${ACR_CLIENT_KEY} ${APLINUX_REGISTRY}
    set -x
    docker pull ${APNGINX_IMAGE}
    docker logout "${APLINUX_REGISTRY}"
fi


# Just in case the container was stopped but not removed (happens during a HardReboot),
# remove the old container. The old container holds a lock on the name that we'll need
# to get things up and running again. Ignoring errors because this is a "best effort"
# cleanup and the error message for failing to cleanup is very specific.
! docker stop apnginx
! docker rm -f apnginx

# Machine hostname
HOSTNAME=$(hostname)
if [ -z "${HOSTNAME}" ]
then
    echo "ERROR: Unable to get machine hostname"
    exit -1
fi

# Get location of machine function certs
if [ -f /etc/ssl/certs/${HOSTNAME}.mf.crt ] && [ -f /etc/ssl/certs/${HOSTNAME}.mf.key ]
then
    MACHINE_PUB_CERT=/etc/ssl/certs/${HOSTNAME}.mf.crt
    MACHINE_PRI_CERT=/etc/ssl/certs/${HOSTNAME}.mf.key
elif [ -f /ap/data/certs/machine_function/${HOSTNAME}.crt ] && [ -f /ap/data/certs/machine_function/${HOSTNAME}.key ]
then
    MACHINE_PUB_CERT=/ap/data/certs/machine_function/${HOSTNAME}.crt
    MACHINE_PRI_CERT=/ap/data/certs/machine_function/${HOSTNAME}.key
fi

# Make sure machine function certs exist
if [ -z "${MACHINE_PUB_CERT}" ] || [ -z "${MACHINE_PRI_CERT}" ]
then
    echo "ERROR: Machine function certificates not found."
    exit -1
fi

# Copy machine function cert keypair into a secure directory.
# Ignore errors in mkdir because we'll catch the condition of failure in the next line.
# Not using mkdir -p because I don't want recursive directory creation that would mask
# a bigger issue like a typo, which mkdir -p would do its job magnificently and we'd
# have secrets in an unexpected location.
! mkdir ${SCRIPT_DIR}/certs
chown root:root ${SCRIPT_DIR}/certs
chmod 0750 ${SCRIPT_DIR}/certs
cp ${MACHINE_PUB_CERT} ${SCRIPT_DIR}/certs/machine.crt
cp ${MACHINE_PRI_CERT} ${SCRIPT_DIR}/certs/machine.key
chown root:root ${SCRIPT_DIR}/certs/machine.key
chmod 0600 ${SCRIPT_DIR}/certs/machine.key

# Get the container up and automatically clean it up on exit
docker run --rm --name=apnginx -h 127.0.0.1 -p 6798:6798 -p 6698:6698 -v ${SCRIPT_DIR}/certs:/certs:ro -v /ap/data/apnginx:/etc/nginx ${APNGINX_IMAGE}

