#!/usr/bin/env python3

"""
Provides functions to query AP environment information.
"""

import sys
import os
import socket

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apcommon import iniparser

__DATA_ROOT__ = '/ap/data'
__APP_ROOT__ = '/ap/app'


def get_data_root():
    return __DATA_ROOT__


def get_app_root():
    return __APP_ROOT__


def set_data_root(data_root):
    global __DATA_ROOT__
    __DATA_ROOT__ = data_root


def set_app_root(app_root):
    global __APP_ROOT__
    __APP_ROOT__ = app_root


def get_linux_bootstrap_ini_path():
    linux_bootstrap_ini_path = get_data_root() + "/LinuxBootstrapData.ini"
    return linux_bootstrap_ini_path


def get_autopilot_ini_path():
    autopilot_ini_path = get_app_root() + "/autopilot.ini"
    return autopilot_ini_path


def get_app_pkg_root():
    return os.path.join(get_app_root(), 'packedimages')


def get_libxdecode_lib_path():
    libxdecode_lib_path = get_app_root() + "/getandrunservices/libxdecode.so"
    return libxdecode_lib_path


def get_environment_name():
    # Returns AP Environment Name
    autopilot_ini = iniparser.INIParser(get_autopilot_ini_path())
    autopilot_ini.read()

    return autopilot_ini.get('FILES', 'Environment')


def get_cluster_name():
    # Returns AP Cluster Name
    autopilot_ini = iniparser.INIParser(get_autopilot_ini_path())
    autopilot_ini.read()

    return autopilot_ini.get('FILES', 'Cluster')


def get_machine_name():
    # Returns AP Machine Name
    return socket.gethostname()


def get_machine_function():
    # Returns AP Machine Function
    autopilot_ini = iniparser.INIParser(get_linux_bootstrap_ini_path())
    autopilot_ini.read()

    return autopilot_ini.get('Config', 'AutopilotMachineFunction')


def get_fsserver_http_port():
    # Returns the http port fsserver is listening on.
    autopilot_ini = iniparser.INIParser(get_autopilot_ini_path())
    autopilot_ini.read()

    if autopilot_ini.get('FSSClient', 'FSSPORT'):
        return autopilot_ini.get('FSSClient', 'FSSPORT')
    else:
        return None


def get_fsserver_secure_port():
    # Returns the SSL port fsserver is listening on.
    autopilot_ini = iniparser.INIParser(get_autopilot_ini_path())
    autopilot_ini.read()

    if autopilot_ini.get('FSSClient', 'FSSSecurePORT'):
        return autopilot_ini.get('FSSClient', 'FSSSecurePORT')
    else:
        return None


def _get_secure_dm_gateway_url(cluster=''):
    # Returns Secure DM gateway URL for the cluster
    if not cluster:
        cluster = get_cluster_name()

    return 'https://dmgateway-be-vip.autopilot.{0}.ap.gbl/DMGateway'.format(cluster)


def get_machine_cert():
    machine = get_machine_name()
    default_pub_path = "/etc/ssl/certs/%s.crt" % machine
    default_pri_path = "/etc/ssl/certs/%s.key" % machine
    legacy_pub_path = "%s/certs/machine/%s.crt" % (get_data_root(), machine)
    legacy_pri_path = "%s/certs/machine/%s.key" % (get_data_root(), machine)

    if os.path.isfile(default_pub_path) and os.path.isfile(default_pri_path):
        return default_pub_path, default_pri_path
    elif os.path.isfile(legacy_pub_path) and os.path.isfile(legacy_pri_path):
        print('WARNING: legacy path for machine cert is being used', file=sys.stderr)
        return legacy_pub_path, legacy_pri_path

    print('ERROR: machine function cert not found', file=sys.stderr)
    return None


def get_machine_function_cert():
    machine = get_machine_name()
    default_pub_path = "/etc/ssl/certs/%s.mf.crt" % machine
    default_pri_path = "/etc/ssl/certs/%s.mf.key" % machine
    legacy_pub_path = "%s/certs/machine_function/%s.crt" % (get_data_root(), machine)
    legacy_pri_path = "%s/certs/machine_function/%s.key" % (get_data_root(), machine)

    if os.path.isfile(default_pub_path) and os.path.isfile(default_pri_path):
        return default_pub_path, default_pri_path
    elif os.path.isfile(legacy_pub_path) and os.path.isfile(legacy_pri_path):
        print('WARNING: legacy path for machine function cert is being used', file=sys.stderr)
        return legacy_pub_path, legacy_pri_path

    print('ERROR: machine function cert not found', file=sys.stderr)
    return None
