#!/usr/bin/env python3

"""
This module provides DMClient class that can be used to send commands to the
Device Manager in the local cluster.

Example Usage:
    dm_client = DMClient(<dm_url>)
    dm_client.send_dm_request("GatewayWatchdogEcho",
                    "Hello World! Test data for the DM Gateway")
"""
import requests
from typing import DefaultDict
from uuid import uuid4
from socket import gethostname

from apsdk import apenv


class DMClient():
    def __init__(self, url="", use_certs=True):
        if url != "":
            self.dm_url = url
        else:
            self.dm_url = apenv._get_secure_dm_gateway_url()

        self.use_certs = use_certs
        self.hostname = gethostname()

    def _get_dm_headers(self, cmd: str, cluster='', environment='') -> DefaultDict:
        """
        Get DM custom headers. DM needs a set of custom headers sent in the
        request
        """
        dm_header = {
            "Command": "X-APDM-Command",
            "ClientUserName": "X-APDM-ClientUserName",
            "ClientAppName": "X-APDM-ClientAppName",
            "ClientMachineName": "X-APDM-ClientMachineName",
            "ClientEnvironmentName": "X-APDM-ClientEnvironmentName",
            "ClientClusterName": "X-APDM-ClientClusterName",
            "TraceIdentifier": "X-APDM-TraceIdentifier"
        }

        if not cluster:
            cluster = apenv.get_cluster_name()

        if not environment:
            environment = apenv.get_environment_name()

        headers = {
            dm_header["Command"]: cmd,
            dm_header["TraceIdentifier"]: str(uuid4()),
            dm_header["ClientUserName"]: "aplinux",
            dm_header["ClientAppName"]: "DMClientLinux",
            dm_header["ClientMachineName"]: self.hostname,
            dm_header["ClientEnvironmentName"]: environment,
            dm_header["ClientClusterName"]: cluster,
        }

        return headers

    def send_request(self, cmd: str, payload: str, cluster='', environment='') -> object:
        """
        Send a request to device manager
        """
        dm_headers = self._get_dm_headers(cmd, cluster, environment)

        machine_function_cert = None
        if self.use_certs is True:
            machine_function_cert = apenv.get_machine_function_cert()
            if machine_function_cert is None:
                return None

        # TODO: Use logger to print request/response
        try:
            response = requests.post(self.dm_url,
                                     cert=machine_function_cert,
                                     # don't verify host cert yet
                                     verify=False,
                                     headers=dm_headers,
                                     data=payload)

        except requests.exceptions.ConnectTimeout:
            print("ConnectTimeout:", self.dm_url)
            return None

        except requests.exceptions.ConnectionError:
            print("ConnectionError:", self.dm_url)
            return None

        except requests.exceptions.Timeout:
            print("Timeout:", self.dm_url)
            return None

        return response


def main():
    dm_client = DMClient()
    dm_client.send_request("GatewayWatchdogEcho",
                           "Hello World! Test data for the DM Gateway")


if __name__ == "__main__":
    main()
