#!/usr/bin/env python3

import sys
import os
import stat
import argparse
import base64

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apsdk import apenv
from apcommon import apsecretclient


class GenevaSecretRefresher:
    SECRET_MDM_CERT = "Geneva-HotPath-Cert"
    SECRET_MDM_KEY = "Geneva-HotPath-Key"

    SECRET_MDMINT_CERT = "GenevaINT-HotPath-Cert"
    SECRET_MDMINT_KEY = "GenevaINT-HotPath-Key"

    METRIC_ENDPOINT = "https://global.metrics.nsatc.net/"
    METRIC_ENDPOINT_INT = "https://az-int.metrics.nsatc.net/"

    APSECRET_ENDPOINTS = ["apsecret.aplinuxmon-prod-co4.co4.ap.gbl:1443",

                          # Legacy APSecret servers
                          # "aplinuxservices.trafficmanager.net:1443"
                          ]

    def __init__(self, client_key_file, client_cert_file):
        self.mdm_environment = {}
        self.mdm_environment["azlinux-dev-co4"] = "int"
        self.mdm_environment["azlinux-test-co4"] = "int"

        self.default_mdm_environment = "prod"

        self.key_file = client_key_file
        self.cert_file = client_cert_file
        self.kv_client = None

        self.mdm_cert = ""
        self.mdm_key = ""

        self.mdm_endpoint = ""

    def __is_connected(self) -> bool:
        if self.kv_client is None:
            return False

        # TODO: Make this more resilient
        return True

    def __connect(self):
        if self.__is_connected():
            return
        self.kv_client = apsecretclient.APSecretClient(self.key_file,
                                                       self.cert_file,
                                                       self.APSECRET_ENDPOINTS)

    # Determine whether to use "int" or "prod" for the MDM publishing environment based on AP Environment
    # or on the parameter passed manually to script
    def detetermine_environment(self) -> str:
        current_environment = apenv.get_environment_name().lower()

        if current_environment in self.mdm_environment:
            mdm_publish_environment = self.mdm_environment[current_environment]
        else:
            mdm_publish_environment = self.default_mdm_environment

        return mdm_publish_environment

    # Decode a base64 string
    def decode_base64_string(self, base64string: str) -> bytes:
        decoded_string = base64.b64decode(base64string, validate=True)

        return decoded_string

    # Write string to a file; used to write secrets to chosen location
    def write_string_to_file(self, directory: str, file_path: str, string_to_write: str):
        if not os.path.isdir(directory):
            raise ValueError("Download directory '{0}' does not exist".format(directory))
        else:
            try:
                    with open(file_path, "w", newline="\n") as text_file:
                        text_file.write(string_to_write)

            except Exception as e:
                print("apsecret_refresh_geneva_hot.write_secret_string_file: Exception: " + str(e))

    # Set permissions on written files so that the secrets are protected
    def set_permissions_on_file(self, full_file_path: str):
        # Set secret file mode to 0600 so that no one can mess with this but root
        try:
            os.chown(full_file_path, 0, 0)
            os.chmod(full_file_path, stat.S_IRUSR | stat.S_IWUSR)

        except Exception as e:
            print("apsecret_refresh_geneva_hot.set_permissions_on_file: Exception: " + str(e))
            raise ValueError("Failed to set permissions on '{0}'".format(full_file_path))

    # Get the right Geneva certificates to use depending on environment
    def download_geneva_certificates(self, directory: str, environment: str):
        self.__connect()

        # Determine the Environment to use for certificates for MDM
        # If environment is passed in as variable honor it, if not, match it to the dictionary
        # defined in the __init__ function
        if not environment:
            environment = self.detetermine_environment()

        if (environment == "prod"):
            self.mdm_cert = self.kv_client.read_cache_secret(self.SECRET_MDM_CERT)
            self.mdm_key = self.kv_client.read_cache_secret(self.SECRET_MDM_KEY)
            self.mdm_endpoint = self.METRIC_ENDPOINT

        elif (environment == "int"):
            self.mdm_cert = self.kv_client.read_cache_secret(self.SECRET_MDMINT_CERT)
            self.mdm_key = self.kv_client.read_cache_secret(self.SECRET_MDMINT_KEY)
            self.mdm_endpoint = self.METRIC_ENDPOINT_INT

        else:
            raise ValueError("Environment not specified and it is required")

        # Decode MDM Cert from base64
        decoded_mdm_cert = self.decode_base64_string(self.mdm_cert).decode('utf-8')
        # Decode MDM Key from base64
        decoded_mdm_key = self.decode_base64_string(self.mdm_key).decode('utf-8')

        # Make sure the download directory exists. If it does not, then we were
        # just refreshing the secret cache
        if not os.path.isdir(directory):
            return

        # Write MDM .env file for metrics endpoint
        mdm_endpoint_file_path = os.path.join(directory, "mdm_endpoint.env")
        mdm_export_string = 'export METRIC_ENDPOINT="' + self.mdm_endpoint + '"'
        self.write_string_to_file(directory, mdm_endpoint_file_path, mdm_export_string)

        # Write Cert and Key Files
        mdm_cert_file_path = os.path.join(directory, "mdm-cert.pem")
        self.write_string_to_file(directory, mdm_cert_file_path, decoded_mdm_cert)
        self.set_permissions_on_file(mdm_cert_file_path)

        mdm_key_file_path = os.path.join(directory, "mdm-key.pem")
        self.write_string_to_file(directory, mdm_key_file_path, decoded_mdm_key)
        self.set_permissions_on_file(mdm_key_file_path)


def main(geneva_data_directory: str, environment: str=""):
    client_key, client_cert = apenv.get_machine_function_cert()
    refresher = GenevaSecretRefresher(client_key, client_cert)

    try:
        refresher.download_geneva_certificates(geneva_data_directory, environment)

    except Exception as e:
        print("ERROR: {0}".format(e), file=sys.stderr)
        sys.exit(1)

    sys.exit(0)

# Example run command:
# /ap/app/aptools/apsecret_refresh_geneva_hot.py -d /ap/data/geneva/mdm -e prod

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Refreshes secrets for accessing Geneva Warm Path and copies "
                                     "certificates to the specified directory.")
    parser.add_argument("-d", "--directory", metavar="DIRECTORY", help="Directory to copy certificates into")
    parser.add_argument("-e", "--environment", metavar="ENVIRONMENT", help="Environment to use - prod/int")

    args = parser.parse_args()

    if args.directory is None or not os.path.exists(args.directory):
        print("ERROR: Valid Download directory must be specified.", file=sys.stderr)
        parser.print_help()
        sys.exit(1)

    main(args.directory, args.environment)
