#!/usr/bin/env python3

# Class used to help send data to statsd
# Automatically includes things like dimensions to the output to make sending stats easier

import sys
import os
import socket
import json
import statsd
import argparse

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apsdk import apenv
from apcommon import apthelper


class StatsdHelper(object):
    def __init__(self,
                 statsd_port: int,
                 statsd_host: str="localhost",
                 osversion_file: str="/opt/ap/OsVersion",
                 osversion_alt_file: str="/ap/data/OsVersion",
                 test_status: bool=False,
                 test_data: dict={}):

        self.osversion_file = osversion_file
        self.osversion_alt_file = osversion_alt_file
        self.deb_package_to_check = "aplinux-shared-services"
        self.test_status = test_status
        self.statsd_host = statsd_host
        self.statsd_port = statsd_port
        self.test_data = test_data

        self.hostname = socket.gethostname()
        self.metric = statsd.StatsClient(statsd_host, statsd_port)

        # If this is a normal run we get our data from localwatchdogstatus.csv
        # and from apenv functions
        if not self.test_status:
            # Tenant = AP Cluster
            # Role = AP Environment
            # RoleInstance = AP Machine Function
            self.tenant = apenv.get_cluster_name()
            self.role = apenv.get_environment_name()
            self.role_instance = apenv.get_machine_function()

        # If this is a unit test we get our data from test data passed in
        else:
            self.tenant = self.test_data["Tenant"]
            self.role = self.test_data["Role"]
            self.role_instance = self.test_data["RoleInstance"]

        self.osversion = ""
        self.shared_services_version = ""

        # Get the osversion
        if not self.get_os_version():
            exit(1)

        # Get the shared-services version
        if not self.get_package_version(self.deb_package_to_check):
            exit(1)

    # Get the OS version by querying the file placed in the image in /opt/ap/OsVersion
    # Fall back to /ap/data/OsVersion if this is an older image (remove once older images are phased out after RC4)
    # This file not existing is an error since it should be imaged
    def get_os_version(self) -> bool:
        if os.path.isfile(self.osversion_file):
            try:
                with open(self.osversion_file, 'r') as f:
                    first_line = f.readline()
                    f.close()

            except Exception as e:
                print("\nstatsdhelper:get_os_version Exception: " + str(e))
                return False

            # The only supported OS right now is Ubuntu so we match that to make sure it's valid
            if "Ubuntu" in first_line:
                self.osversion = first_line.strip()
            else:
                print("statsdhelper.get_os_version OsVersion does not match Ubuntu")
                return False
        elif os.path.isfile(self.osversion_alt_file):
            try:
                with open(self.osversion_alt_file, 'r') as f:
                    first_line = f.readline()
                    f.close()
            except Exception as e:
                print("\nstatsdhelper:get_os_version Exception: " + str(e))
                return False

            # The only supported OS right now is Ubuntu so we match that to make sure it's valid
            if "Ubuntu" in first_line:
                self.osversion = first_line.strip()
            else:
                print("statsdhelper.get_os_version OsVersion does not match Ubuntu")
                return False
        else:
            print("statsdhelper.get_os_version OsVersion file does not exist")
            return False

        return True

    # Use apthelper.py library to make sure package is installed before checking version
    def check_package_installed(self, package: str) -> bool:
        if apthelper.apt_get_is_installed(package):
            return True
        else:
            return False

    # Use apthelper.py library to query package versions
    def get_package_version(self, package: str) -> bool:
        if self.test_status:
            self.shared_services_version = self.test_data["ServiceVersion"]
            return True

        if self.check_package_installed(package):
            version = apthelper.apt_get_version(package)
            self.shared_services_version = version.strip()
            return True
        else:
            return False

    # Send gauge statistic to statsd with required Dimensions
    # Minimal input required to send a metric
    def send_gauge(self, namespace: str, metric_name: str, metric_value: int, custom_dims: dict={}) -> bool:
        # print("\nSending statsd values:")
        # print("statsdhelper.send_gauge Namespace: " + namespace)
        # print("statsdhelper.send_gauge Metric: " + metric_name)
        # print("statsdhelper.send_gauge Tenant: " + self.tenant)
        # print("statsdhelper.send_gauge Role: " + self.role)
        # print("statsdhelper.send_gauge RoleInstance: " + self.role_instance)
        # print("statsdhelper.send_gauge MachineName: " + self.hostname)
        # print("statsdhelper.send_gauge OSVersion: " + self.osversion)
        # print("statsdhelper.send_gauge ServiceVersion: " + self.shared_services_version)
        # print("statsdhelper.send_gauge Value: " + str(metric_value))

        # Make sure all required dimensions are present and accounted for
        if not all((namespace, metric_name, metric_value, self.tenant, self.role, self.role_instance,
                    self.hostname, self.osversion, self.shared_services_version)):
            print("statsdhelper.send_gauge: Not all of the required dimensions were defined properly")
            return False

        dims_dict = {"Tenant": self.tenant,
                     "Role": self.role,
                     "RoleInstance": self.role_instance,
                     "MachineName": self.hostname,
                     "OSVersion": self.osversion,
                     "ServiceVersion": self.shared_services_version}

        # Add custom dimensions
        if type(custom_dims) is dict:
            dims_dict.update(custom_dims)

        metric_dict = {"Namespace": namespace,
                       "Metric": metric_name,
                       "Dims": dims_dict}

        # Show JSON output sent to statsd
        # print("\nJSON Formatted output for statsd:")
        # print(json.dumps(metric_dict, indent=4, sort_keys=True))

        try:
            metric_json = json.dumps(metric_dict)
        except Exception as e:
            print("statsdhelper.send_gauge: json exception: " + str(e))
            return False

        try:
            self.metric.gauge(metric_json, metric_value)
        except Exception as e:
            print("statsdhelper.send_gauge: statsd exception: " + str(e))
            return False

        return True

# Example run command:
# /ap/app/aptools/statsdhelper.py -p 8125 -n MyNamespace -m ServiceFailures -a 1
# /ap/app/aptools/statsdhelper.py -p 8125 -n MyNamespace -m ServiceFailures -a 1 -c "Cause:DBTimeout,User:dbadmin"

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Used to send metrics to Geneva Hot Path (MDM) "
                                     "with additional, useful Autopilot environment dimensions.")
    parser.add_argument("-s", "--statsdhost", metavar="STATSDHOST", default="localhost",
                        help="statsd host to send metrics through, defaults to localhost")
    parser.add_argument("-p", "--port", metavar="PORT", help="statsd port to send metrics through")
    parser.add_argument("-n", "--namespace", metavar="NAMESPACE", help="MDM namespace")
    parser.add_argument("-m", "--metricname", metavar="METRICNAME", help="metric name")
    parser.add_argument("-a", "--metricvalue", metavar="METRICVALUE", help="metric value (integer)")
    parser.add_argument("-c", "--customdims", metavar="DIMENSIONS",
                        help="comma-delimited list of additional dimensions to send with the "
                             "metric, with the dimension name and value separated by a colon "
                             "(optional)")

    args = parser.parse_args()

    if not (args.statsdhost and args.port and args.namespace and args.metricname and args.metricvalue):
        print("ERROR: Missing required parameters", file=sys.stderr)
        parser.print_help()
        sys.exit(1)

    # parse statsd port number
    try:
        port = int(args.port)
    except:
        print("ERROR: statsd port number must be an integer", file=sys.stderr)
        parser.print_help()
        sys.exit(1)

    if port < 1 or port > 65535:
        print("ERROR: statsd port number must be between 1 and 65535", file=sys.stderr)
        parser.print_help()
        sys.exit(1)

    # parse the metric value
    try:
        metricvalue = int(args.metricvalue)
    except:
        print("ERROR: metric value must be an integer", file=sys.stderr)
        parser.print_help()
        sys.exit(1)

    # parse custom dimensions
    customdims = {}
    if args.customdims:
        for dim in args.customdims.split(","):
            d = dim.split(":")
            if len(d) != 2:
                print("ERROR: syntax error in custom dimension '{0}'".format(dim), file=sys.stderr)
                parser.print_help()
                sys.exit(1)

            k = d[0].strip()
            v = d[1].strip()
            if not (k and v):
                print("ERROR: syntax error in custom dimension '{0}'".format(dim), file=sys.stderr)
                parser.print_help()
                sys.exit(1)

            customdims[k] = v

    metrics = StatsdHelper(statsd_host=args.statsdhost, statsd_port=port)
    if not metrics.send_gauge(args.namespace, args.metricname, metricvalue, customdims):
        print("ERROR: failed to send metric to mdm", file=sys.stderr)
        sys.exit(1)
