#!/usr/bin/env python3

"""
Functions to manipulate time sent in by AP and provide AP specific time formats
"""
from datetime import datetime, timedelta, timezone


def time_to_ap_readable(timestamp):
    """
    AP Needs time in a custom format
    :return:
    :param timestamp:
    :return:
    """
    dt = datetime.fromtimestamp(timestamp)
    return dt.strftime('%Y-%m-%d %H:%M:%S')


def win_filetime_to_posixtime(win_filetime):
    """
    Windows FILETIME = 100 nanoseconds intervals since Jan 1, 1601, 64-bit.
    Posix time = seconds since Jan 1, 1970
    :param win_filetime:
    :return:
    """
    try:
        ms = (int(win_filetime, 16) / 10.)
        ret_datetime = datetime(1601, 1, 1) + timedelta(microseconds=ms)
    except OverflowError:
        # Return a max date on Overflow error since MWD most of the times sends in FFFFFFFFFFFFFFFF
        return datetime(9999, 12, 31)

    return ret_datetime


def posixtime_to_win_filetime(posix_time_secs):
    """
    Windows FILETIME = 100 nanoseconds intervals since Jan 1, 1601, 64-bit.
    Posix time = seconds since Jan 1, 1970
    :param posix_time_secs:
    :return:
    """
    # Nano Seconds since Jan 1, 1601 to Jan 1, 1970
    epoch_diff_ns = 116444736000000000
    mtime_ns = posix_time_secs * 10000000
    win_file_time_ns = int(mtime_ns) + epoch_diff_ns

    hex_str = hex(win_file_time_ns)[2:].upper()

    # Prefix the hex string with 0 to fill 16 bytes
    file_time_str = ''.join((((16 - len(hex_str)) * '0'), hex_str))

    return file_time_str


def get_file_mod_datetime(file_path):
    """
    Get a file's modification time in UTC
    :param file_path:
    :return:
    """
    mod_datetime = datetime.utcfromtimestamp(file_path.stat().st_mtime)
    mod_datetime = mod_datetime.replace(tzinfo=timezone.utc)

    # Don't care about microseconds since we don't get it in HTTP header
    mod_datetime = mod_datetime.replace(microsecond=0)

    return mod_datetime
