#!/usr/bin/env python3
"""
FSSERVER Class to provide files and listings for Data Drive
"""

import aptime
import cherrypy

from csv import DictWriter
from pathlib import PureWindowsPath, PurePosixPath, PosixPath
from statsd import StatsClient
from configparser import ConfigParser
from datetime import datetime, timezone
from io import StringIO
from os import access, R_OK
from re import sub


class DataDrive(object):
    """
    Handle DataDrive requests landing on fsserver.
    """

    def __init__(self, config: ConfigParser):
        """
        Process requests for Data drive D:
        """
        self.config = config
        self.metric = StatsClient('localhost', 9125, prefix='fsserver')
        self.anonymous_paths = ["app",
                                "app/autopilot.ini",
                                "data",
                                "data/servicemanager/servicestatus.csv",
                                "data/watchdogoutput/localwatchdogstatus.csv"]

    @staticmethod
    def index(**kwargs):
        """
        Sometimes an empty GET is done by some watchdogs to see if we are
        accessible.
        """
        return

    def process_data_path_req(self, *args, **kwargs):
        """
        Process requests of path http://<url>/data
        """
        # example /D/Data
        path = 'data/' + kwargs['path_info']
        return self._process_request(path)

    def process_drive_path_req(self, *args, **kwargs):
        """
        Process requests of path http://<url>/drived/
        """
        # example /D/Data
        path = kwargs['path_info']
        return self._process_request(path)

    def process_winpath_req(self, *args, **kwargs):
        """
        Process a path request sent in Windows format
        """
        win_path = kwargs['win_file_path']
        if win_path[0] == '\\':
            # Remove the leading '\'
            path = PureWindowsPath(win_path[1:]).as_posix()
        else:
            path = PureWindowsPath(win_path).as_posix()
        return self._process_request(path)

    def _process_request(self, path):
        """
        Process a request
        """
        if path.lower() not in self.anonymous_paths:
            raise cherrypy.HTTPError(
                status=404,
                message="FSSERVER: File/Dir not allowed without authentication:" + path)

        posix_path = self._get_local_path(path.lower())
        if posix_path is None:
            raise cherrypy.HTTPError(
                status=404,
                message="FSSERVER: Path could not be processed:" + path)

        if not posix_path.exists():
            raise cherrypy.HTTPError(
                status=404,
                message="FSSERVER: File/Dir not found:" + posix_path.name)

        if posix_path.is_file():
            self.metric.incr('request.file')
            return self._process_file_request(posix_path)

        if posix_path.is_dir():
            self.metric.incr('request.dir')
            return self._process_dir_request(posix_path)

    def _set_file_headers(self, file_path):
        """
        Set file headers specific to AP
        """
        self._set_common_headers()

        file_stat = file_path.stat()

        cherrypy.response.headers['X-FileSize'] = file_stat.st_size
        cherrypy.response.headers['FSServer-Content-Length'] = file_stat.st_size

        # Convert to windows file time hex string. This is how AP
        # needs it
        win_file_ctime = aptime.posixtime_to_win_filetime(file_stat.st_ctime)
        cherrypy.response.headers['Created'] = win_file_ctime

        win_file_mtime = aptime.posixtime_to_win_filetime(file_stat.st_mtime)
        cherrypy.response.headers['Last-Modified'] = win_file_mtime
        cherrypy.response.headers['Modified'] = win_file_mtime

        return

    @staticmethod
    def _set_common_headers():
        """
        Dont send Server=CherryPy
        fsserver Windows doesn't send Content-Type, Server headers
        :return:
        """
        # del cherrypy.response.headers['Server']
        # fsserver Windows doesn't send Content-Type, Server headers
        # del cherrypy.response.headers['Content-Type']

        # Connection is not persistent
        cherrypy.response.headers['Connection'] = 'Close'

    def _get_local_path(self, path):
        """
        Get Local Path information
        """
        try:
            app_root = self.config['LINUX']['AppDir']
            data_root = self.config['LINUX']['DataDir']
            # 'data/servicemanager/servicestatus.csv' becomes
            # '/usr/local/ap/data/servicemanager/servicestatus.csv'
            if path.startswith('data'):
                posix_path = PosixPath(data_root).joinpath(
                    PurePosixPath(path).relative_to('data'))
            elif path.startswith('app'):
                posix_path = PosixPath(app_root).joinpath(
                    PurePosixPath(path).relative_to('app'))
            else:
                posix_path = PosixPath('')

        except ValueError as e:
            print("_get_local_path - ValueError :", e)
            return None

        return posix_path

    def _process_file_request(self, posix_path):
        """
        Process a request for a file
        """
        if (access(str(posix_path), R_OK)) is False:
            raise cherrypy.HTTPError(
                status=403,
                message="FSSERVER: No permission to read :" + str(posix_path))

        return self._send_file(posix_path)

    def _process_dir_request(self, posix_path):
        # Build the response in a in-memory csv
        """
        Process a directory listing request. The return text is in AP specific format.
        """
        in_mem_text = StringIO()
        in_mem_text.write('#Fields:')

        fields = ['filename', 'size', 'attributes', 'creationtime',
                  'lastaccesstime', 'lastwritetime']
        writer = DictWriter(in_mem_text, fieldnames=fields)
        writer.writeheader()

        curr_dir_stat = posix_path.stat()
        writer.writerow({
            'filename': '.',
            'size': curr_dir_stat.st_size,
            'attributes': 'D',
            'creationtime': aptime.time_to_ap_readable(curr_dir_stat.st_ctime),
            'lastaccesstime': aptime.time_to_ap_readable(curr_dir_stat.st_atime),
            'lastwritetime': aptime.time_to_ap_readable(curr_dir_stat.st_mtime)
        })

        parent_dir_stat = posix_path.parent.stat()
        writer.writerow({
            'filename': '..',
            'size': parent_dir_stat.st_size,
            'attributes': 'D',
            'creationtime': aptime.time_to_ap_readable(parent_dir_stat.st_ctime),
            'lastaccesstime': aptime.time_to_ap_readable(parent_dir_stat.st_atime),
            'lastwritetime': aptime.time_to_ap_readable(parent_dir_stat.st_mtime)
        })

        for file in posix_path.iterdir():
            fstat = file.stat()
            if file.is_dir():
                attribute = 'D'
            else:
                attribute = 'A'

            writer.writerow({
                'filename': file.name,
                'size': fstat.st_size,
                'attributes': attribute,
                'creationtime': aptime.time_to_ap_readable(fstat.st_ctime),
                'lastaccesstime': aptime.time_to_ap_readable(fstat.st_atime),
                'lastwritetime': aptime.time_to_ap_readable(fstat.st_mtime)
            })
        in_mem_text.write('#EOF')

        return self._send_dir_listing(in_mem_text)

    @staticmethod
    def _handle_if_unmodified_since(file_path, http_date):
        """
        If the file is not modified since the time specified, handle as if it it a normal request.
        If the file has been modified since the time specified, then return a HTTP Status 412
        """
        mod_datetime = aptime.get_file_mod_datetime(file_path)

        try:
            # As per HTTP spec we will get GMT
            header_time = datetime.strptime(http_date, '%a, %d %b %Y %H:%M:%S GMT')
        except (ValueError, OverflowError, OSError):
            raise cherrypy.HTTPError(
                status=400,
                message="FSSERVER: Invalid date format in If-Unmodified-Since HTTP Header - " +
                        http_date)

        header_time = header_time.replace(tzinfo=timezone.utc)

        # If the file was modified after the date specified, return 412
        if header_time < mod_datetime and cherrypy.request.method in ("GET", "HEAD"):
            raise cherrypy.HTTPError(412)
        else:
            return

    @staticmethod
    def _handle_if_modified_since(file_path, http_date):
        """
        If file has not been modified since date specified, return 304. Else handle the request
        as normal.
        """
        mod_datetime = aptime.get_file_mod_datetime(file_path)

        try:
            # As per HTTP spec we will get GMT
            header_time = datetime.strptime(http_date, '%a, %d %b %Y %H:%M:%S GMT')
        except (ValueError, OverflowError, OSError):
            raise cherrypy.HTTPError(
                status=400,
                message="FSSERVER: Invalid date format in If-Modified-Since HTTP Header - " +
                        http_date)

        header_time = header_time.replace(tzinfo=timezone.utc)

        # If the file was not modified after the date specified, return 304
        if header_time > mod_datetime and cherrypy.request.method in ("GET", "HEAD"):
            raise cherrypy.HTTPRedirect([], 304)
        else:
            return

    @staticmethod
    def _handle_ap_if_last_modified(file_path, win_filetime):
        """
        Handle AP custom header. If file has not been modified since date specified, return 304.
        Else handle the request as normal.
        """
        try:
            # As per HTTP spec we will get GMT
            header_time = aptime.win_filetime_to_posixtime(win_filetime)

        except (ValueError, OSError) as e:
            raise cherrypy.HTTPError(
                status=400,
                message="FSSERVER: Invalid date format in X-IfLastModifiedTimestampGreater HTTP Header - " +
                        win_filetime + e)

        header_time = header_time.replace(tzinfo=timezone.utc)
        mod_datetime = aptime.get_file_mod_datetime(file_path)

        # If the file was not modified after the date specified, return 304
        # print ("mod_time; ", mod_datetime)
        # print ("header_time; ", header_time)
        if mod_datetime < header_time:
            raise cherrypy.HTTPRedirect([], 304)
        else:
            return

    def _check_headers(self, file_path):
        """
        Check if any standard http or non-standard ap headers were specified
        """
        if_mod_hdr = cherrypy.request.headers.get('If-Modified-Since')
        if if_mod_hdr:
            return self._handle_if_modified_since(file_path, if_mod_hdr)

        if_unmod_hdr = cherrypy.request.headers.get('If-Unmodified-Since')
        if if_unmod_hdr:
            return self._handle_if_unmodified_since(file_path, if_unmod_hdr)

        # We should never hit this since Machine Watch Dog that uses this param is modified
        # to use standard HTTP headers
        ap_if_last_modified = cherrypy.request.headers.get('X-IfLastModifiedTimestampGreater')
        if ap_if_last_modified:
            # AP Last Modified is sent in as a 64-bit windows time stamp, we can't use it here.
            # This is just a safeguard to ensure we don't fail Machine Watch Dog
            return self._handle_ap_if_last_modified(file_path, ap_if_last_modified)

    def _send_dir_listing(self, stream):
        """
        Send in the directory listing in the stream after adding AP headers
        """
        self._set_common_headers()

        # Go to the beginning of the stream
        stream.seek(0)

        return stream

    def _send_file(self, file_path):
        """
        Send the file back to the client
        """
        self._set_file_headers(file_path)

        # Check if any standard or non-standard ap headers were specified
        self._check_headers(file_path)

        is_cockpit = "autopilot-cockpit" == cherrypy.request.headers.get('User-Agent').lower()
        try:
            file = file_path.open('rb')
            if is_cockpit:
                data = file.read()
                # Can't send binary to cockpit since it is for user display
                if b'\0' in data:
                    msg = "ERROR: " + file_path.name + " is not a text file, cannot be returned"
                    return sub("\r?\n", "\r\n", _cowsay(msg))
                else:
                    return sub(b"\r?\n", b"\r\n", data)
            else:
                return file

        except EOFError:
            self.metric.incr('file.error.404')
            raise cherrypy.HTTPError(
                status=404,
                message="FSSERVER: Empty File, could not be read: EOFError" + file_path.name)

        except IOError:
            self.metric.incr('file.error.403')
            raise cherrypy.HTTPError(
                status=403,
                message="FSSERVER: File exists, but could not be read: IOError" + file_path.name)


def _cowsay(msg):
    return msg + """

     _______
    < gimme >
     -------
        \   ^__^
         \  (oo)\_______
            (__)\       )\/\*
                ||----w |
                ||     ||

 _____
< no? >
 -----
   \.
    \.
        .--.
       |o_o |
       |:_/ |
     .//   \ \.
     (|     | )
   ./'\_   _/`\.
    \___)=(___/

"""
