#!/usr/bin/env python3
"""
FSServer, that serves file and directory requests landing on the machine. This sits behind
AP NGINX and serves files and directories.
"""
import sys
import os
import cherrypy
from argparse import ArgumentParser
from configparser import ConfigParser
from mimetypes import types_map

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from data_drive import DataDrive


def setup_request_dispatcher(config):
    """
    Setup request dispatchers for types of calls landing on fsserver
    :param config:
    :return:
    """
    data_drive_controller = DataDrive(config)

    # cherrypy internally imports routes package
    dispatcher = cherrypy.dispatch.RoutesDispatcher()

    # Handling DataDrive file requests. Order of registration Matters.
    drive_d_alias = ['drived', 'driveD', 'DriveD', 'DRIVED', 'd', 'D']
    for alias in drive_d_alias:
        dispatcher.connect(name='', route='/' + alias + '/{path_info:.*}',
                           controller=data_drive_controller, action="process_drive_path_req")
        dispatcher.connect(name='', route='/' + alias + '{win_file_path}',
                           controller=data_drive_controller, action="process_winpath_req")

        # TODO: We might need to enable /data/ too.
        # dispatcher.connect(name='', route='/data/{path_info:.*}',
        # controller=data_drive_controller, action="process_data_path_req")

    # Return empty result for ServiceManager Commands
    dispatcher.connect(name='', route='/ServiceManager/{path_info:.*}',
                       controller=data_drive_controller, action="index")

    # DataBus dispatch. Not enabled yet
    # dispatcher.connect(name='', route='/DataBus', controller=data_bus, action="process_request")

    # TODO: have a handler for incorrect paths sent
    dispatcher.connect(name='', route='/', controller=data_drive_controller, action="index")

    return dispatcher


def load_fsserver_ini(filename):
    """
    Load the fsserver config ini passed in
    :param filename: Name of the file
    :return: ConfigParser object containing fsserver.ini
    """
    # TODO: Validate Config before loading
    config = ConfigParser()
    config.read(filename)
    return config


def setup_http_listener(config: ConfigParser, bind_address: str):
    """
    Setup an HTTP Listener. We can create as many instances of server as we need.
    All will follow cherrypy's engine life-cycle
    :param config: ConfigParser
    :return: cherrpy._cpserver.Server()
    """
    server = cherrypy._cpserver.Server()
    server.socket_host = bind_address
    server.socket_port = int(config['HTTPHosting']['Port'])
    server.subscribe()

    return server


def setup_cherrypy(config: ConfigParser, bind_address: str):
    """
    Setup CherryPy for starting
    :param config: Config Parser containing fsserver.ini
    :return: Instance of CherryPY
    """
    dispatcher = setup_request_dispatcher(config)

    cp_config = {
        '/': {
            'request.dispatch': dispatcher,
            'log.access_file': config['LINUX']['AccessLog'],
            'log.error_log': config['LINUX']['ErrorLog'],
            'log.access_log_format': '{h} {t} "{r}" {s} {b} "{f}" "{a}"'
        }
    }

    # Setup cherrypy
    cherrypy.tree.mount(root=None, config=cp_config)
    cherrypy.server.unsubscribe()

    # No need to setup SSL Listener since NGINX handles it for us.
    # Just configure a HTTP Listener for us.
    setup_http_listener(config, bind_address)

    cherrypy.engine.signals.subscribe()

    return cherrypy


def start_and_block(cherrypy_engine):
    """
    Start and Block CherryPY server
    :param cherrypy_instance:
    """
    cherrypy_engine.start()
    cherrypy_engine.block()


def main(config_file_name, bind_address):
    """
    The Main
    :param config_file_name: Location of fsserver.ini
    """
    fs_config = load_fsserver_ini(config_file_name)

    # CherryPy uses mimetypes module to determine the content-type.
    types_map['.csv'] = 'text/csv'

    try:
        cherrypy_instance = setup_cherrypy(fs_config, bind_address)

        # Start Cherry Py main engine and wait for requests from NGINX.
        start_and_block(cherrypy_instance.engine)

    except Exception as e:
        print("ERROR: Running FSSEerver threw exception:", e)
        sys.exit(1)


if __name__ == '__main__':
    parser = ArgumentParser(description="Start AP FSServer")
    parser.add_argument('config_file', metavar='CONFIG.INI', help='Config File ini')
    parser.add_argument('bind_address', nargs='?', default='127.0.0.1', metavar='ADDRESS', help='Address to listen on')

    args = parser.parse_args()
    config_file = args.config_file
    bind_address = args.bind_address

    if not os.path.isfile(config_file):
        print("ERROR: %s could not be found" % config_file)
        sys.exit(1)

    main(config_file, bind_address)
