#!/usr/bin/env python3
# GetAndRunServices equivalent in Linux on AP.
# Scheduled once every 5 mins
# 1. Gets the Service List from DM
# 2. Gets ServiceList ini from manifest server if there is a new version for
#    deployment
# 3. If PrepareList field has changed from last time, download the package from DS Server.
#     - Using HTTP
# 4. If 'List' changed, then inform ServiceManager to start the new versions of services.
# 5. Drop the new servicelist.json file for service manager
# 6. Update DM with the new service list version, even if no changes happened, but more the
#    6hrs passed since last update

import argparse
import configparser
import json
import os
import requests
import shutil
import socket
import sys
import tempfile
import time
import typing
from logging import getLogger, config


# if script's parent directory isn't in sys.path, then add it
if '/'.join(sys.path[0].split("/")[:-1]) not in sys.path:
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apsdk import apenv
from apsdk import dmclient
from aptools import statsdhelper
from apcommon import apck

# globals
# Metric MDM namespace to upload metrics
metric_namespace = "APSharedServices"

# New Service list JSON file that GARS writes for informing service manager of List it
# should be running
new_json_file = os.path.join(apenv.get_data_root(), 'servicemanager/new_servicelist.json')

# Active Service list JSON file that Service Manager writes for informing GARS of current List
# it was successfully able to activate
active_json_file = os.path.join(apenv.get_data_root(), 'servicemanager/active_servicelist.json')

# All the service lists are archived in a servicelists folder
service_list_dir = os.path.join(apenv.get_app_root(), 'servicelists')

# Packed images directory contains all the images unpacked by GARS
packed_images_dir = os.path.join(apenv.get_app_root(), 'packedimages')

copy_only_header = '#CopyOnly: '
data_copy_header = '#Data: '

# Csv to ignore when we process the service list
ignore_csv_suffixes = ['.Autopilot-AutopilotClient-VE.csv', '.Autopilot-AutopilotClient-Qualification-VE.csv']

# Force download and install package from dynamic storage each time, even if it was already
# downloaded earlier.
force_install_packages = False

# Setup logger
config.fileConfig(os.path.join(sys.path[0], 'logging.ini'))
log = getLogger(__name__)


def get_current_list(ini_config: configparser.ConfigParser) -> str:
    return ini_config['ServiceManager']['List']


def get_prepare_list(ini_config: configparser.ConfigParser) -> str:
    return ini_config['ServiceManager']['PrepareList']


def send_dm_command(dm_client: dmclient.DMClient,
                    cmd: str, retries=4) -> str:
    """
    Send a command to DM. Handles retries as sometimes DM doesn't respond with
    proper response.
    :return: DM response http text.
    """
    for i in range(0, retries):
        resp = dm_client.send_request(cmd, '')
        if resp:
            reply = resp.text
            resp.connection.close()
            return reply

        time.sleep(2)

    log.error("Sending command to DM failed")
    return None


def get_service_list_from_dm(dm_client: dmclient.DMClient) -> str:
    """
    Gets the service list from Device Manager
    :return: Response to GetServiceList command.
    """
    machine_name = socket.gethostname()
    env_name = apenv.get_environment_name()
    if not env_name:
        log.error("Could not obtain AP environment name.")
        return None

    cmd = 'GetServiceList -s ' + machine_name + ' -w ' + env_name

    log.info("Getting service list from DM: %s", cmd)
    resp = send_dm_command(dm_client, cmd)

    return resp


def update_dm_service_status(dm_client: dmclient.DMClient,
                             current_list: str,
                             prepare_list: str,
                             error_message: str) -> str:
    """
    Update DM with current service status
    :return: DM Response to ServiceStatus command
    """
    # If the service list processing had failed, send the error message to DM
    # indicating the failure. DM will issue a rollback in this case.
    if not error_message:
        # Command to update service list to DM
        cmd = 'ServiceStatus -serviceList ' + current_list + ' -prepareServiceList ' + prepare_list

        # DM expects GARS to report its service list atleast once every 900 mins. If this does not
        # happen, the machine goes to repair mode.
        metrics.send_gauge(metric_namespace,
                           "GARSDMHeartbeatSuccess",
                           1)
    else:
        # Metric: GARS reported an error from service manager to DM. Service Manager reports package
        # processing error to GARS, which gets propagated to DM.
        metrics.send_gauge(metric_namespace,
                           "GARSDMErrorReported",
                           1)
        cmd = 'ServiceStatus -serviceList ' + current_list + ' -prepareServiceList ' + prepare_list + ' -error ' + error_message

    log.info("Updating service status to DM: %s", cmd)
    resp = send_dm_command(dm_client, cmd)

    return resp


def get_manifest_server_list(ini_config: configparser.ConfigParser) -> list:
    """
    Get the manifest server list from an ini config returned
    :return: List of manifest server IPs
    """
    if not ini_config.has_section('Manifest'):
        log.error('DM Reply does not have [Manifest] section : %s', ini_config)
        return None
    else:
        server_list = ini_config['Manifest']['ServerList']
        return server_list.split(':')


def get_env_name_from_svc_list_name(service_list_name: str,
                                    default_env='') -> str:
    """
    Get environment name from a service list name
    """
    service_list_suffix = get_service_list_suffix(service_list_name)

    # Dynamic service lists look like virtual ones, but they belong to
    # the current environment, not VE
    if service_list_suffix and service_list_suffix.lower() == 'dynamic':
        return default_env

    return service_list_suffix


def get_pkgs_from_service_list(service_list_csv_path: str) -> typing.List[str]:
    """
    Get package list from the service list specified
    """
    pkg_list = list()
    try:
        with open(service_list_csv_path) as f:
            for line in f:
                if line.startswith('#'):
                    # if line starts with #, we don't care about it yet
                    continue

                # Line is of format:
                #  APAMLocal.az_compute_2065832_AutoDeployment,0,0,,
                # Get just the package name out of it
                items = line.split(sep=',')[0]
                if len(items) > 4:
                    pkg_name = line.split(sep=',')[0]
                    pkg_list.append(pkg_name)
                else:
                    # We ignore this error since some lines are not parsable
                    log.debug("Package name is not of expected format: %s, in service list: %s",
                              line,
                              service_list_csv_path)

        return pkg_list

    except IOError as e:
        log.error("Service List csv could not be read : %s, error %s", service_list_csv_path, e)
        return None


def get_service_list_suffix(service_list_name: str) -> str:
    """
    Get the service list suffix
    :return: Service list suffix.
    """
    items = service_list_name.split('.')
    count = len(items)
    if (count > 3) and items[count - 1] == 'csv':
        suffix = items[count - 2]
    else:
        suffix = None

    return suffix


def get_apck_url(ip: str,
                 listen_port: str,
                 env_name: str,
                 service_name: str) -> str:
    """
    Get the APCK URL to download the apck file from dynamic storage.
    :return: APCK URL as string.
    """
    if not ip or not listen_port or not env_name or not service_name:
        log.error("Invalid values building apck url: ip=%s, listen_port=%s, env_name=%s, service_name=%s",
                  ip, listen_port, env_name, service_name)
        return None

    url = 'http://' + ip + ':' + str(listen_port) + '//DynamicStorage/?action=DownloadStream&StreamName=stream://' + env_name + '/app/app.retail.amd64/' + service_name + '&version=apck'

    return url


def get_manifest_server_url(server: str,
                            port: str,
                            environment: str,
                            csv_name: str) -> str:
    """
    Get the Manifest server URL
    :return: Manifest server URL as string.
    """
    if not server or not port or not environment or not csv_name:
        log.error("Invalid value building manifest server url: server=%s, port=%s, environment=%s, csv_name=%s",
                  server, port, environment, csv_name)
        return None

    url = 'http://' + server + ':' + port + '/ManifestServer?action=GetServiceList&environment=' + environment + '&name=' + csv_name

    return url


def download_service_list_csv(manifest_servers: typing.List[str],
                              listen_port: str,
                              environment: str,
                              csv_name: str,
                              timeout: int) -> str:
    """
    Download the service list csv file from manifest server
    :return: Service list CSV as string
    """
    # Try from each server in the manifest server list
    for server in manifest_servers:
        url = get_manifest_server_url(server, str(listen_port), environment, csv_name)
        if not url:
            log.error("Could not obtain manifest servier URL")
            return None

        # We will get a "301 - Moved Permanently" from Manifest server that
        # redirects us to the actual DynamicStorage url
        try:
            log.info("Download Service List : %s", url)
            resp = requests.get(url, allow_redirects=True, timeout=timeout)
            if resp.status_code == requests.codes.ok:
                content = resp.text
                resp.close()
                log.debug("Response : %s", content)
                return content

            log.error("Error downloading service list: resp.Stats=%s, resp=%s",
                      resp.status_code, resp.text)

        except requests.exceptions.ConnectTimeout:
            log.error("ConnectTimeout: %s", url)
        except requests.exceptions.ConnectionError:
            log.error("ConnectionError: %s", url)
        except requests.exceptions.Timeout:
            log.error("Timeout: %s", url)

    return None


def download_package(service_name: str,
                     service_list_name: str,
                     image_root: str,
                     manifest_servers: typing.List[str],
                     listen_port: str,
                     timeout=20) -> str:
    """
    Download packages in a service list.
    :return: APCK file full path that was downloaded
    """
    env_name = get_env_name_from_svc_list_name(service_list_name)

    # If this is a physical environment
    if not env_name:
        env_name = apenv.get_environment_name()
        if not env_name:
            log.error("Could not obtain AP environment name")
            return None

    # Create a directory with service_list_name to put the apck files
    service_file_name = os.path.basename(service_list_name)

    # Filename may start with a '.'
    target_dir = os.path.join(image_root, service_file_name)
    if not os.path.exists(target_dir):
        try:
            shutil.os.mkdir(target_dir)
        except OSError as e:
            log.error("Creating directory: %s. error=%s", target_dir, str(e))
            return None

    # Try downloading from each manifest server in the list
    for ip in manifest_servers:
        # Where should then file file be placed
        apck_file_path = os.path.join(target_dir, service_name + '.apck')

        # No need to download again if we have already downloaded this package
        package_dir = get_package_extract_dir(apck_file_path)
        if not force_install_packages and shutil.os.path.exists(package_dir) and shutil.os.path.isdir(package_dir):
            log.debug("Already downloaded : %s", apck_file_path)
            return apck_file_path

        apck_url = get_apck_url(ip, listen_port, env_name, service_name)
        if not apck_url:
            log.error("Could not get apck url for download")
            return None
        log.debug("Downloading Service Pkg from : %s", apck_url)

        # We will get a file to download from DynamicStorage server
        try:
            # Download as a stream as the size might become very large.
            # Stream download ensures memory doesn't blow out, also lets us
            # separate handle header download and content download separately.
            resp = requests.get(apck_url,
                                stream=True,
                                allow_redirects=True,
                                timeout=timeout)

            if resp.status_code != requests.codes.ok:
                log.debug("Error reading pkg file. resp.StatusCode=%s, resp=%s. Trying next server.",
                          resp.status_code, resp.text)
                resp.close()
                continue

            tmp_file_name = write_tmp_pkg(resp, service_name, target_dir)
            if not tmp_file_name:
                log.error("Couldn't download file from url: %s", apck_url)
                return None

            # Everything is fine, rename to original apck file name
            try:
                os.rename(tmp_file_name, apck_file_path)
            except IOError as e:
                log.error("Renaming package from %s to %s. error=%s",
                          tmp_file_name, apck_file_path, str(e))
                return None

            return apck_file_path

        except requests.exceptions.ConnectTimeout:
            log.error("ConnectTimeout: %s", apck_url)
            continue
        except requests.exceptions.ConnectionError:
            log.error("ConnectionError: %s", apck_url)
            continue
        except requests.exceptions.Timeout:
            log.error("Timeout: %s", apck_url)
            continue

    return None


def write_tmp_pkg(resp,
                  service_name: str,
                  target_dir: str) -> str:
    """
    Write the downloaded file to a temporary package name. Temp file should be cleaned
    by caller, this isn't auto cleaned on exit.
    :returns: Temporary file name the package was downloaded to.
    """
    try:
        file_tmp = tempfile.NamedTemporaryFile(mode='wb',
                                               dir=target_dir,
                                               prefix=service_name,
                                               suffix='.downloading',
                                               delete=False)
        # Handle file download now. Since we started this as a stream download,
        # actual file download starts when we do iter_content below
        chunk_size = 1024 * 5
        download_start_time = time.time()
        for chunk in resp.iter_content(chunk_size):
            if chunk:
                file_tmp.write(chunk)
                file_tmp.flush()
        elapsed_time_secs = time.time() - download_start_time
        download_size = file_tmp.tell()

        # Metric: Package download time in seconds. MDM cannot take 0 or negative int as a
        # metric, so reset it to 1, if it is the case.
        if elapsed_time_secs <= 0:
            elapsed_time_secs = 1

        metrics.send_gauge(metric_namespace,
                           "GARSPkgDownloadSecs",
                           elapsed_time_secs,
                           {'PkgName': service_name, 'Size': str(download_size)})

        # release the connection back to the pool
        resp.close()

        # Metric: Package download success
        metrics.send_gauge(metric_namespace,
                           "GARSPkgDownloadSuccess",
                           1,
                           {'PkgName': service_name})

        download_size = file_tmp.tell()
        file_tmp.close()

    except Exception as e:
        # Metric: Package download failure, with size. Sometimes siz is the one causing
        # issue
        metrics.send_gauge(metric_namespace,
                           "GARSPkgDownloadFailure",
                           1,
                           {'PkgName': service_name, 'Size': str(download_size)})
        log.error("Error creating temporary apck file %s", e)
        return None

    # If we didn't receive the full file
    # TODO: check CRC
    # if download_size == 0 or download_size != resp.headers['X-FileSize']:
    if download_size == 0:
        log.error("Error: Didn't receive full file. expected:%d, Downloaded:%d",
                  download_size)
        os.remove(file_tmp.name)
        return None

    return file_tmp.name


def parse_dm_reply(dm_reply: str) -> typing.Union[typing.DefaultDict, typing.List]:
    """
    Parse the DM reply. Will be in Ini format.
     Sample:
     [ServiceManager]
     List=LINUX.SafeDeploy_CL2173222_0.csv:.ap_08_14_e2f41c5_CL2534374_0.Autopilot-AutopilotClient-VE.csv
     PrepareList=LINUX.SafeDeploy_CL2173222_0.csv:.ap_08_14_e2f41c5_CL2534374_0.Autopilot-AutopilotClient-VE.csv
     LastList=LINUX.SafeDeploy_CL2173222_0.csv:.ap_08_14_e2f41c5_CL2534374_0.Autopilot-AutopilotClient-VE.csv
     LastPrepareList=LINUX.SafeDeploy_CL2173222_0.csv:.ap_08_14_e2f41c5_CL2534374_0.Autopilot-AutopilotClient-VE.csv
     LastReportTime=2017-08-15 01:39:53.084
     StaticMachineFunction=LINUX
     StaticScaleunitId=10
     DynamicMachineFunction=LINUX
     DynamicScaleunitId=10
     [Manifest]
     ServerList=25.72.86.152:100.101.41.98:100.101.78.34:25.78.2.209
    :returns: ServiceManager section as dictionary and Manifest IPs as list
    """
    try:
        reply = configparser.ConfigParser()
        reply.read_string(dm_reply)
    except configparser.Error as e:
        log.error("Reading DM reply: %s. Error=%s", dm_reply, str(e))
        return None

    # Verify we got the mandatory fields needed
    sections = ['ServiceManager', 'Manifest']
    for section in sections:
        if not reply.has_section(section):
            log.error('DM Reply does not have [%s] section:%s',
                      section,
                      dm_reply)
            return None

    # Get the ServiceManager section
    service_manager_dict = reply['ServiceManager']

    # Verify we got the mandatory fields
    fields = ['List', 'PrepareList', 'LastList', 'LastPrepareList']
    for field in fields:
        if field not in service_manager_dict:
            log.error(
                'DM Reply does not have field [ServiceManager] %s in reply: ',
                dm_reply)
            return None

    if 'ServerList' not in reply['Manifest']:
        log.error('DM Reply does not have [Manifest].ServerList %s:',
                  dm_reply)
        return None

    # Get the IPs of manifest servers
    manifest_servers = reply['Manifest']['ServerList'].split(':')

    return service_manager_dict, manifest_servers


def download_pkgs_in_csv(csv_file: str,
                         manifest_servers: typing.List[str]) -> typing.List:
    """
    Download the packages listed in csv
    :returns: List of downloaded packages
    """
    log.debug("Downloading packages from : %s", csv_file)

    pkg_list = get_pkgs_from_service_list(csv_file)
    if not pkg_list:
        log.error("Package list couldn't be obtained.")

    if not os.path.isdir(packed_images_dir):
        try:
            shutil.os.mkdir(packed_images_dir)
        except OSError as e:
            log.error("Creating directory: %s. error=%s", packed_images_dir, str(e))
            return None

    downloaded_packages = list()
    for pkg_name in pkg_list:
        log.debug("Downloading Package: %s", pkg_name)
        apck_file_path = download_package(pkg_name,
                                          csv_file,
                                          packed_images_dir,
                                          manifest_servers,
                                          apenv.get_fsserver_http_port(),
                                          timeout=20)
        if not apck_file_path:
            # Metric: Package download failure
            metrics.send_gauge(metric_namespace,
                               "GARSPkgDownloadFailure",
                               1,
                               {'PkgName': pkg_name})
            log.error("Error Could not download : %s", pkg_name)
        else:
            downloaded_packages.append(apck_file_path)
            log.debug("Downloaded apck : %s", apck_file_path)

    # Send total number of packages downloaded
    metrics.send_gauge(metric_namespace,
                       "GARSPkgsDownloaded",
                       len(downloaded_packages))
    return downloaded_packages


def download_csv_from_list(csv_list: str,
                           manifest_servers: typing.List[str]) -> typing.List[str]:
    """
    Download the CSV files containing list of packages from a list sent
    :returns: List of CSV files downloaded from manifest servers
    """
    log.debug("Downloading CSVs from: %s", csv_list)

    if not os.path.isdir(service_list_dir):
        try:
            shutil.os.mkdir(service_list_dir)
        except OSError as e:
            log.error("Creating directory: %s, error=%s", service_list_dir, str(e))
            return None

    # Each list contains a set of csv file names delimited by a :
    csv_list = csv_list.split(':')
    csv_file_list = list()

    for csv_name in csv_list:
        # Special case fo OM - Orchestration manager. We are going to ignore
        # it for now. We will tango with OM later.
        if csv_name.startswith('.OM.'):
            continue

        csv_content = download_service_list_csv(manifest_servers,
                                                apenv.get_fsserver_http_port(),
                                                apenv.get_environment_name(),
                                                csv_name,
                                                timeout=15)
        if not csv_content:
            log.error("Error reading csv from manifest server : %s", csv_name)
            return None

        # Write csv contents to a file
        csv_full_path = os.path.join(service_list_dir, csv_name)
        try:
            with open(csv_full_path, 'w') as f:
                csv_file_list.append(f.name)
                f.write(csv_content)
        except IOError as e:
            log.error("Error writing csv content to file: %s. error=%s",
                      csv_full_path, str(e))
            return None

    return csv_file_list


def get_package_extract_dir(apck_file_path: str) -> str:
    """
    Get the directory path to extract an apck package.
    :returns: Dir path as string
    """

    # We extract in the same directory as the apck file.
    # /ap/app/packedimages/<pkg-name>/
    name_without_ext = os.path.splitext(os.path.basename(apck_file_path))[0]
    target_dir = os.path.join(os.path.dirname(apck_file_path), name_without_ext)

    return target_dir


def extract_service_package(apck_file_path: str) -> bool:
    """
    Extract the apck package sent into the same directory
    :returns: True if extraction succeeded, False otherwise
    """
    if not os.path.exists(apck_file_path):
        log.error("APCK file not found: %s", apck_file_path)
        return False

    try:
        apck_reader = apck.ApPackReader(apck_file_path)
    except ValueError as e:
        log.error("APCK file parsing failed: %s", str(e))
        return False
    except OSError as e:
        log.error("Creating an ApPackReader failed with exception: %s", str(e))
        return False

    # Create the package directory. Package directory is
    # /ap/app/packedimages/<pkg-name>/
    target_dir = get_package_extract_dir(apck_file_path)
    if not os.path.isdir(target_dir):
        try:
            shutil.os.mkdir(target_dir)
        except OSError as e:
            log.error("Creating directory: %s. error=%s", target_dir, str(e))
            return False

    # Extract all files in the apck file in the target directory
    try:
        log.debug("Extracting package : %s", apck_file_path)
        apck_reader.extract_all(target_dir)
        log.debug("Extracting package Complete : %s", apck_file_path)

        # Metric: Package extaction success
        metrics.send_gauge(metric_namespace,
                           "GARSPkgExtractSuccess",
                           1,
                           {'PkgFile': apck_file_path})
    except Exception as e:
        # Metric: Package extaction failure
        metrics.send_gauge(metric_namespace,
                           "GARSPkgExtractFailure",
                           1,
                           {'PkgFile': apck_file_path})
        log.error("Extracting Package %s in directory: %s. error=%s",
                  apck_file_path, target_dir, str(e))
        return False

    return True


def is_ignore_csv(csv_file):
    for ignore_csv in ignore_csv_suffixes:
        if csv_file.endswith(ignore_csv):
            log.info("Csv file : %s. Matches pattern: %s", csv_file, ignore_csv)
            return True

    return False


def process_prepare_list(service_manager_dict,
                         manifest_servers: typing.List[str]) -> str:
    """
    Proceed downloading the packages. This is the Pre-Fetching mechanism used by
    DM/GARS.  We will later get a List entry change that will show which service
    list should service manager be running. Packages already downloaded will be
    skipped.
    :returns: PrepareList value that was successfully processed.
    """
    prepare_list = service_manager_dict['PrepareList']

    # Download the CSV files
    downloaded_csv_files = download_csv_from_list(prepare_list, manifest_servers)
    if not downloaded_csv_files or len(downloaded_csv_files) is 0:
        log.debug('No CSVs were downloaded. No valid CSVs for Linux.')
        return prepare_list

    # Download the packages listed in the csv file
    pkg_lists_downloaded = []
    for csv_file in downloaded_csv_files:
        if is_ignore_csv(csv_file):
            log.info("Ignoring csv file : %s", csv_file)
            continue

        downloaded_pkg_list = download_pkgs_in_csv(csv_file, manifest_servers)
        if not downloaded_pkg_list:
            log.error('APCK Packages could not be downloaded for : %s', csv_file)
            return None

        pkg_lists_downloaded.append(downloaded_pkg_list)
        log.debug("%d packages downloaded. List: %s", len(downloaded_pkg_list), downloaded_pkg_list)

    # Extract all the downloaded APCK packages.
    for package_list in pkg_lists_downloaded:
        for apck_file_path in package_list:
            # No need to download again if we have already downloaded this package
            package_dir = get_package_extract_dir(apck_file_path)
            if not force_install_packages and os.path.isdir(package_dir):
                log.debug("Package already extracted:  %s", apck_file_path)
            else:
                if extract_service_package(apck_file_path) is False:
                    log.error("Extracting package %s", apck_file_path)
                    return None
                # Remove apck file once extracted
                log.debug("Removing apck file after extraction: %s", apck_file_path)
                try:
                    shutil.os.remove(apck_file_path)
                except Exception as e:
                    log.error("Error removing extracted apck file: %s, error = %s", apck_file_path, e)
                    return None

    return prepare_list


def inform_service_manager(service_manager_config: typing.DefaultDict) -> bool:
    """
    Drops a file for service manager to pick up and switch service list
    :returns: Success or failure result of the operation
    """
    # Object returned by the config parser is not a dictionary, though
    # it looks like one, hence isn't JSON serializable. Put it in an object
    # so that we can serialize
    json_serializable = {
        'List': service_manager_config['List'],
        'LastList': service_manager_config['LastList'],
        'PrepareList': service_manager_config['PrepareList'],
        'LastPrepareList': service_manager_config['LastPrepareList']
    }

    try:
        fp = open(new_json_file, 'w')
    except IOError as e:
        log.error("Could not open file for writing: %s. error=%s", new_json_file, str(e))
        return False

    with fp:
        try:
            json.dump(json_serializable, fp, indent=2)
        except TypeError as e:
            log.error("JSON dump error: %s", str(e))
            return False

    return True


def read_active_list(file_name: str) -> typing.DefaultDict:
    """
    Read the active_list json.
    :return: Dictionary of values in active_list file
    """
    if not os.path.exists(file_name):
        log.info('active_servicelist.json not found:%s', active_json_file)
        return None

    try:
        fp = open(file_name, 'r')
    except IOError as e:
        log.error("Could not read file %s. error=%s", file_name, str(e))
        return None

    with fp:
        try:
            return json.load(fp)
        except ValueError as e:
            log.error("JSON syntax error: %s", str(e))

    return None


def report_to_dm(dm_client: dmclient.DMClient,
                 prepare_list: str) -> bool:
    """
    Report current service status to Device Manager
    :return: True or False based on reporting operation to DM
    """

    # Read the current list json file
    active_dict = read_active_list(active_json_file)

    if not active_dict:
        log.info('active_servicelist.json not found or could not be read:%s', active_json_file)
        return False

    # List should come out of active_servicelist.json since this is written by
    # service manager and it does so only when it can process the packages
    # specified in the list.
    log.debug("Updating service status to DM")
    resp = update_dm_service_status(dm_client,
                                    active_dict.get('List'),
                                    prepare_list,
                                    active_dict.get('ErrorMessage', None))

    if not resp:
        log.error('Updating DM Service Status')
        return False

    if not resp.startswith('OK'):
        log.error('DM did not respond with OK to Service Status update : %s', resp)

    return True


def main():
    # Get a new DM Client to talk to the lord(DM)
    dm_client = dmclient.DMClient('')

    # Declare the global metrics object here to enable mocking inside
    # unittests.
    global metrics
    metrics = statsdhelper.StatsdHelper(statsd_port=9125)

    # Get service list from Device Manager
    service_list_reply = get_service_list_from_dm(dm_client)
    if not service_list_reply:
        log.error("Getting service list from DM failed.")
        metrics.send_gauge(metric_namespace,
                           "GARSRunFailure",
                           1,
                           {'Operation': 'ServiceListDownload'})
        exit(1)
    log.debug("DM Reply: %s", service_list_reply)

    if service_list_reply.startswith('No Response'):
        log.error("No Reponse received from DM. This is a valid response.")
        exit(0)

    # Parse DM service list reply and obtain values to work with.
    service_manager_dict, manifest_servers = parse_dm_reply(service_list_reply)
    if not service_manager_dict or not manifest_servers:
        log.error("Could not parse ServiceList ini returned by DM.")
        metrics.send_gauge(metric_namespace,
                           "GARSRunFailure",
                           1,
                           {'Operation': 'ParseServiceListFromDM'})
        exit(2)

    # Process the PrepareList entry in the Service Ini. PrepareList indicates
    # that GARS should pre-fetch the packages.
    processed_prepare_list = process_prepare_list(service_manager_dict, manifest_servers)

    if not processed_prepare_list:
        log.error("Processing prepare list failed. Cannot proceed.")
        metrics.send_gauge(metric_namespace,
                           "GARSRunFailure",
                           1,
                           {'Operation': 'PrepareListProcessing'})
        exit(3)

    # If we successfully downloaded the packages, we should inform
    # service manager to process. Should do this regardless of PrepareList being
    # new or old, since the 'List' entry is what matters to service manager.
    if not inform_service_manager(service_manager_dict):
        # We specifically do not exit here if this operation fails, because doing so
        # will reimage the machine.
        log.error("Could not inform service manager of change.")

    # Send List and PrepareList back to DM informing the versions this machine is at.
    if report_to_dm(dm_client, processed_prepare_list) is False:
        log.error("Reporting to DM failed. Exitting GARS")
        metrics.send_gauge(metric_namespace,
                           "GARSRunFailure",
                           1,
                           {'Operation': 'ReportToDM'})
        exit(4)

    log.info("GARS processing completed.")

    exit(0)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Send a command to getandrunservices service.")
    parser.add_argument("-f", "--force", action='store_true', help="force download and install of current servicelist")

    args = parser.parse_args()
    force_install_packages = args.force

    main()
