#!/usr/bin/env python
# Copyright (c) Microsoft Corporation.  All rights reserved.

import sys
import os
import socket
import stat
import subprocess
import shutil
import time
import argparse
import re
import ssl
import hashlib
import httplib
import urllib
from datetime import datetime
from datetime import timedelta
from M2Crypto import EVP
from M2Crypto import RSA
from M2Crypto import X509
from M2Crypto import m2

# Do not allow to import all symbols when using "from apca_client import *"
__all__ = []

# Globals
rsa_key_bits = 2048
pem_req_header = '-----BEGIN CERTIFICATE REQUEST-----'
pem_req_trailer = '-----END CERTIFICATE REQUEST-----'
pem_cert_header = '-----BEGIN CERTIFICATE-----'
pem_cert_trailer = '-----END CERTIFICATE-----'

# apca endpoints
apca_new_machine_cert_ep = '/APCAOneWay/SignMachineCertInitial'
apca_renew_cert_ep = '/APCAMutual/SignMachineCertRenew'
apca_export_pki_ep = '/APCAOneWay/ExportApPkiStore'
apca_new_mf_cert_ep = '/APCAMutual/SignMachineFunction'

# paths of temporary files used in certificate requests
certreq_rsa_path = '{0}etc/ssl/certs/{1}.rsa'
certreq_pri_path = '{0}etc/ssl/certs/{1}.key.tmp'
certreq_pub_path = '{0}etc/ssl/certs/{1}.crt.tmp'

# paths to certificates
default_apca_path = '{0}usr/share/ca-certificates/ap/'
default_rootca_path = '{0}usr/share/ca-certificates/ap/apca_roots.pem'
default_intermediate_path = '{0}usr/share/ca-certificates/ap/apca_intermediates.pem'
default_intermediatep12_path = '{0}usr/share/ca-certificates/ap/apca_intermediates.p12'
default_machine_public_path = '{0}etc/ssl/certs/{1}.crt'
default_machine_private_path = '{0}etc/ssl/certs/{1}.key'
default_mf_public_path = '{0}etc/ssl/certs/{1}.mf.crt'
default_mf_private_path = '{0}etc/ssl/certs/{1}.mf.key'

# legacy paths to certificates
legacy_intermediate_path = '{0}ap/data/certs/intermediate/ap_pki_store.pem'
legacy_machine_path = '{0}ap/data/certs/machine/{1}'
legacy_mf_path = '{0}ap/data/certs/machine_function/{1}'

# disallowed certificate store
disallowed_store = '{0}usr/share/ca-certificates/ap/disallowed.csv'
disallowed_store_tmp = '{0}usr/share/ca-certificates/ap/disallowed.tmp'

# timeout for http requests
ENDPOINT_TIMEOUT = 5

# settings for APSecret endpoints used to update the disallowed cert store
APSECRET_RETRIES = 2
APSECRET_ENDPOINTS = ["apsecret.aplinuxmon-prod-co4.co4.ap.gbl:1443"]


class HTTPSClient(httplib.HTTPSConnection):
    """ Class to make a HTTPS connection, with support for binding to iface """
    def __init__(self,
                 host,
                 timeout=None,
                 local_ip=None,
                 keyfile=None,
                 certfile=None):
        httplib.HTTPSConnection.__init__(self, host, timeout=timeout)
        self.local_ip = ''
        if local_ip is not None:
            self.local_ip = local_ip
        self.key_file = keyfile
        self.cert_file = certfile

    def connect(self):
        """ Connect to a host on a given (SSL) port. """
        sock = None
        sock = socket.create_connection((self.host, self.port),
                                        self.timeout,
                                        (self.local_ip, 0))
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()
        self.sock = ssl.wrap_socket(sock,
                                    self.key_file,
                                    self.cert_file,
                                    cert_reqs=ssl.CERT_NONE)


def apca_http_request(uri, local_ip, url, method, post_data=None, keyfile=None, certfile=None):
    """ Perform HTTPS request """
    headers = {'content-type': 'text/plain'}
    status = 0
    body = ''

    try:
        # make the request
        conn = HTTPSClient(uri,
                           timeout=ENDPOINT_TIMEOUT,
                           local_ip=local_ip,
                           keyfile=keyfile,
                           certfile=certfile)
        conn.request(method, url=url, body=post_data, headers=headers)
        response = conn.getresponse()

        # get the response
        status = response.status
        body = response.read()
    except Exception as err:
        raise Exception('Unable to make {0} request to {1} '
                        'for {2}: {3}'.format(method, uri, url, err))

    if status != 200:
        raise Exception('Failed {0} request made to {1} for {2}: '
                        'Status {3}'.format(method, uri, url, status),
                        body)

    return body


def get_apca_vip(cluster):
    """ Gets the VIP of the cluster's APCA """
    c = ''
    if cluster:
        c = cluster
    else:
        try:
            c = os.environ['AP_CLUSTER']
        except:
            c = ''

    # default apca vip cluster
    if not c:
        c = 'Co3c'

    return 'APCA-backend.Autopilot.{0}.ap.gbl:443'.format(c)


def set_public_cert_perms(path):
    """ Set the file mode for public certs to 0644 """
    if not os.path.isfile(path):
        print 'Info: File not found: {0}'.format(path)
        return True

    try:
        os.chmod(path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)
    except Exception as err:
        print 'Error: Unable to set permissions ' \
              'for \'{0}\': {1}'.format(path, err)
        return False

    return True

def set_private_cert_perms(path):
    """ Set the file mode for private certs to 0644 because everyone
        on the system needs to use them. TODO: lock this down
        because private keys need to be protected """
    if not os.path.isfile(path):
        print 'Info: File not found: {0}'.format(path)
        return True

    try:
        os.chmod(path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)
    except Exception as err:
        print 'Error: Unable to set permissions ' \
              'for \'{0}\': {1}'.format(path, err)
        return False

    return True


def secure_shred(path):
    """ Shred the file """
    shred_args = [ 'shred', '-u', path ]
    err = ''

    try:
        p = subprocess.Popen(shred_args,
                             shell=False,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        out, err = p.communicate()
    except Exception as e:
        print 'Error: Failed to shred \'{0}\': {1}: {2}'.format(path, err, e)
        return False

    return True


def secure_file_move(old_path, new_path, overwrite=False):
    """ Copies the file from old_path to new_path and then shreds old_path
        If new_path already exists, only shred old_path """
    if not os.path.isfile(old_path):
        return True

    if old_path == new_path:
        return True

    if overwrite or not os.path.exists(new_path):
        try:
            shutil.copyfile(old_path, new_path)
        except Exception as e:
            print 'Error: Failed to copy file from \'{0}\' to ' \
                  '\'{1}\': {2}'.format(old_path, new_path, e)
            return False

    # Skipping the shred because this prevents rolling back to
    # old localcertagent. Security note, mf certs must still be
    # readable by everyone on the machine so securely deleting
    # them from the old location is moot until the mf certs can
    # be properly secured.
    #if not secure_shred(old_path):
    #    return False

    return True


def shred_legacy_secrets(machine_name, root_dir='/'):
    """ Shred the legacy paths to the private machine and machine
        function certificates """

    # This is a separate function to be extra sure all secrets are
    # shredded from the legacy locations
    success = True
    shred_paths = []
    for path in [ 'machine.key', machine_name + '.key', machine_name + '.keyenc' ]:
        shred_paths.append(legacy_machine_path.format(root_dir, path))
        shred_paths.append(legacy_mf_path.format(root_dir, path))

    # Shred machine and machine function certificates in legacy locations.
    # We don't care about the intermediate certificates or the public
    # certificates because they're all public and don't need securing.
    for path in shred_paths:
        if os.path.isfile(path):
            if not secure_shred(path):
                success = False

    return success


def secure_legacy_paths(machine_name, root_dir='/'):
    """ Copies the machine, machine function, and intermidiate
        certificates to secure, default locations in the
        filesystem """
    success = True

    # move machine certs
    if not secure_file_move(legacy_machine_path.format(root_dir, machine_name + '.crt'),
                            default_machine_public_path.format(root_dir, machine_name)):
        success = False
    if not set_public_cert_perms(default_machine_public_path.format(root_dir, machine_name)):
        success = False
    if not secure_file_move(legacy_machine_path.format(root_dir, machine_name + '.key'),
                            default_machine_private_path.format(root_dir, machine_name)):
        success = False
    if not set_private_cert_perms(default_machine_private_path.format(root_dir, machine_name)):
        success = False

    # move machine function certs
    if not secure_file_move(legacy_mf_path.format(root_dir, machine_name + '.crt'),
                            default_mf_public_path.format(root_dir, machine_name)):
        success = False
    if not set_public_cert_perms(default_mf_public_path.format(root_dir, machine_name)):
        success = False
    if not secure_file_move(legacy_mf_path.format(root_dir, machine_name + '.key'),
                            default_mf_private_path.format(root_dir, machine_name)):
        success = False
    if not set_private_cert_perms(default_mf_private_path.format(root_dir, machine_name)):
        success = False

    # Move intermediate certs. Don't worry about the .p12 file because
    # that's only supported on Windows machines. It's left as a raw
    # artifact whenever we refresh the intermediates in case if something
    # goes wrong with the conversion to PEM format.
    if not secure_file_move(legacy_intermediate_path.format(root_dir),
                            default_intermediate_path.format(root_dir)):
        success = False

    # shred everything secret from legacy location
    if not shred_legacy_secrets(machine_name, root_dir):
        success = False

    return success


def is_certificate_root_ca(path):
    """ Returns True if the DER formatted certificate is a root ca and False
        otherwise. If the certificate is not DER formatted, an exception is
        raised. """

    if not path:
        raise ValueError('Path must be provided')
    if not os.path.isfile(path):
        raise Exception('\'{0}\' is not a file'.format(path))

    openssl_args = ['openssl',
                    'x509',
                    '-inform',
                    'DER',
                    '-text',
                    '-noout',
                    '-in',
                    path]
    command = ' '.join(openssl_args)
    out = ''
    err = ''

    try:
        p = subprocess.Popen(openssl_args,
                             shell=False,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        out, err = p.communicate()
    except Exception as e:
        raise Exception('Failed to run \'{0}\': {1}'.format(command, e))
    if p.returncode != 0:
        raise Exception('Non-zero returncode ({0}) running \'{1}\': '
                        '{2}'.format(p.returncode,
                                     command,
                                     err.replace('\n', ' ')))
    if err:
        raise Exception('Unexpected error message executing \'{0}\': '
                        '{1}'.format(command,
                                     err.replace('\n', ' ')))

    # Root certificate authorities have CA:TRUE while intermediates have
    # this in addition to pathlen
    return bool(('CA:TRUE' in out) and not ('CA:TRUE, pathlen:' in out))


def convert_der_to_pem(der_path, pem_path):
    """ Converts a DER formatted certificate file to a PEM file and
        returns the PEM data. On failure, raises an exception. """

    openssl_args = ['openssl',
                    'x509',
                    '-inform',
                    'DER',
                    '-outform',
                    'PEM',
                    '-in',
                    der_path,
                    '-out',
                    pem_path]
    command = ' '.join(openssl_args)
    err = ''

    try:
        p = subprocess.Popen(openssl_args,
                             shell=False,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        out, err = p.communicate()
    except Exception as e:
        raise Exception('Failed to run \'{0}\': {1}'.format(command, e))
    if p.returncode != 0:
        raise Exception('Non-zero returncode ({0}) running \'{1}\': '
                        '{2}'.format(p.returncode,
                                     command,
                                     err.replace('\n', ' ')))
    if err:
        raise Exception('Unexpected error message executing \'{0}\': '
                        '{1}'.format(command,
                                     err.replace('\n', ' ')))

    with open(pem_path, 'r') as pem_file:
        return pem_file.read()


def add_der_certificate_to_root_ca_pem(root_dir, cer_path):
    """ Converts a DER certificate file to a PEM file, then adds it to the
        apca root ca PEM. Returns True on success and False on error. """
    pem_path = cer_path[:-3] + 'pem'
    roots_pem = default_rootca_path.format(root_dir)

    # Convert from der to pem format
    try:
        pem_data = convert_der_to_pem(cer_path, pem_path)
    except Exception as e:
        print 'Warning: Unable to convert root cert from DER format ' \
              'to PEM format of file \'{0}\': {1}'.format(cer_path, e)
        return False

    if not pem_data:
        print 'Warning: Unable to convert root cert from DER format ' \
              'to PEM format of file \'{0}\''.format(cer_path)
        return False

    # append to root ca pem file
    try:
        with open(roots_pem, 'a') as roots:
            roots.write(pem_data)
    except Exception as e:
        print 'Warning: Unable to append root cert ' \
              'to \'{0}\': {1}'.format(roots_pem, e)
        return False

    return True


def cleanup_apca_directory(root_dir):
    """ Removes superfluous intermediate certificates """

    # remove unused apislands directory
    apislands = '{0}/usr/share/ca-certificates/apislands'.format(root_dir)
    if os.path.isdir(apislands):
        shutil.rmtree(apislands)

    apca_dir = default_apca_path.format(root_dir)
    for cer_file in os.listdir(apca_dir):
        cer_path = os.path.join(apca_dir, cer_file)
        remove_file = False

        # delete intermediate certificates that should never have been
        # separated from their original PEM file
        if cer_path.endswith('.crt'):
            remove_file = True

        # Delete DER-formatted root ca certificates not properly marked.
        # Commented out for now so that we can throw the switch later...
        # Requirement to throw the switch:
        #    Verify with apsec2 the only trusted APCA roots are
        #    'AP Root Certificate Authority 2013' and
        #    'AP Root Certificate Authority 2014'
        if cer_path.endswith('.cer'):
            try:
                if is_certificate_root_ca(cer_path):
                    if not add_der_certificate_to_root_ca_pem(root_dir,
                                                              cer_path):
                        print 'Warning: Unable to add root ca to ' \
                              'apca roots pem'
                # else:
                #     remove_file = True
            except Exception as e:
                print 'Error: is_certificate_root_ca: {0}'.format(e)
                # if not str(e).startswith('Failed to execute '):
                #     remove_file = True

        if remove_file:
            os.remove(cer_path)


def download_revoked_certificates(local_ip):
    """ Downloads the list of revoked certificates from APSecret returning
        an empty string on error or the list as a string on success """

    # download list of revoked certificate thumbprints from APSecret
    body = ''
    for endpoint in APSECRET_ENDPOINTS:
        for retry in range(0, APSECRET_RETRIES):
            try:
                body = apca_http_request(endpoint,
                                         local_ip, '/disallowed', 'GET')
                break
            except Exception as e:
                print 'Warning: download_revoked_certificates: {0} ' \
                      'retry {1}/{2}: {3}'.format(endpoint,
                                                  retry + 1,
                                                  APSECRET_RETRIES,
                                                  e)
                body = ''

    return body


def write_revoked_certificates(revoked_certificates, tmp_filename, csv_filename):
    """ Writes the revoked certificates to the specified temp file then
        atomically moves the temp file to overwrite the csv file returning
        True on success and False on error """

    # write to temporary file
    try:
        with open(tmp_filename, 'w') as f:
           f.write(revoked_certificates)
    except Exception as e:
        print 'Error: Unable to write revoked cert list to ' \
              'file \'{0}\': {1}'.format(tmp_filename, e)
        return False

    # atomically switch to new disallowed store
    try:
        os.rename(tmp_filename, csv_filename)
    except Exception as e:
        print 'Error: Unable to atomically move revoked cert file \'{0}\' ' \
              'to \'{1}\': {2}'.format(tmp_filename, csv_filename, e)
        return False

    return True


def refresh_disallowed_store(local_ip, root_dir='/'):
    """ Downloads a list of revoked certificates returning True
        on success and False on error """

    # retrieve the list of revoked certificates
    revoked_certificates = download_revoked_certificates(local_ip)

    # exit if we could not download
    if not revoked_certificates:
        print 'Error: refresh_disallowed_store: Unable to request revoked cert list'
        return False

    # atomically switch to new disallowed store
    if not write_revoked_certificates(revoked_certificates,
                                      disallowed_store_tmp.format(root_dir),
                                      disallowed_store.format(root_dir)):
        print 'Error: refresh_disallowed_store: Unable to save revoked cert list'
        return False

    return True


def read_cert_expiration_date(public_cert):
    """ Get the expiration date of the certificate """
    try:
        cert = X509.load_cert(public_cert)
        cert_expiration = str(cert.get_not_after())
        return datetime.strptime(cert_expiration,
                                 '%b %d %H:%M:%S %Y %Z')
    except Exception as e:
        print 'Error: Unable to get expiration for certificate \'{0}\': ' \
              '{1}'.format(public_cert, e)
    return None


def check_cert_state(cert_expiration_date, days_to_renewal):
    """ Checks to see if certificates need to be renewed or if they're
        expired. Returns -1 on expiration, 0 if renewal is needed, and 1
        when no action is needed. """
    expiration_limit = datetime.utcnow() + timedelta(days=days_to_renewal)
    if datetime.utcnow() > cert_expiration_date:
        return -1
    elif expiration_limit > cert_expiration_date:
        return 0
    return 1


def update_appki_cert_chain(local_ip, uri, root_dir='/'):
    """ Exporting AP PKI store from APCA """
    # apca returns the intermediate certificates in the response body as
    # a single pkcs12 file
    try:
        body = apca_http_request(uri, local_ip, apca_export_pki_ep, 'GET')
    except Exception as e:
        print 'Error: update_appki_cert_chain: {1}'.format(e)
        return False


    pkcs12_file = default_intermediatep12_path.format(root_dir)
    pem_file = default_intermediate_path.format(root_dir)

    # try to write the intermediates to disk in case if we can't use pyopenssl
    try:
        with open(pkcs12_file, 'w') as f:
            f.write(body)
    except Exception as e:
        print 'Warning: Unable to write pkcs12 file ' \
              '\'{0}\': {1}'.format(pkcs12_file, e)
        pkcs12_file = ''

    # try using pyopenssl to convert pkcs12 to pem
    all_pems = convert_pkcs_to_pem_pyopenssl(body)

    if not all_pems:
        # pyopenssl failed so lets fallback to openssl via shell
        if not pkcs12_file:
            print 'Error: No pkcs12 file to attempt openssl conversion ' \
                  'to pem format'
            print 'Error: Failed to store APCA cert intermediate certificates'
            return False

        all_pems = convert_pkcs_to_pem_openssl(pkcs12_file)
        if not all_pems:
            print 'Error: Unable to convert intermediate certificates ' \
                  'to pem format using openssl'
            print 'Error: Failed to store APCA cert intermediate certificates'
            return False

    # write the intermediates in pem format to disk
    try:
        with open(pem_file, 'w') as f:
            f.write(all_pems)
    except IOError as e:
        print 'Error: Unable to write pem content ' \
              'to file \'{0}\': {1}'.format(pem_file, e)
        return False

    return True

    print 'Info: APCA intermediate certificates stored correctly'
    return True


def generate_cert_request_pem(pkey, machine_name):
    """ Creates a X509 certificate sign request """
    cert_request = X509.Request()
    cert_request.set_version(cert_request.get_version())

    # set public key
    cert_request.set_pubkey(pkey)

    # set CN field
    name = X509.X509_Name()
    name.CN = machine_name
    cert_request.set_subject(name)

    # get fingerprint of keypair
    digest = hashlib.sha1(pkey.as_der()).hexdigest().upper()
    fingerprint = ':'.join(digest[pos: pos + 2] for pos in range(0, 40, 2))

    # Extensions array of tuples (name, value, if_critical)
    extensions = [('extendedKeyUsage', 'serverAuth,clientAuth', False),
                  ('keyUsage', 'dataEncipherment, keyEncipherment', True),
                  ('subjectKeyIdentifier', fingerprint, False)]

    # Set extensions
    extstack = X509.X509_Extension_Stack()

    for extension in extensions:
        ext = X509.new_extension(extension[0], extension[1])
        if extension[2]:
            ext.set_critical()
        extstack.push(ext)

    cert_request.add_extensions(extstack)

    # add SHA1 signature
    cert_request.sign(pkey, 'sha1')

    return cert_request.as_pem()


def passphrase_callback(v):
    """ This is a callback that returns a passphrase needed to
        dump a key to file. """
    keystorepass = 'swan'
    return keystorepass


def remove_passprhase(private_rsa_key, private_pem_cert, passprhase):
    """ This is so we can remove the passphrase from the file key.
        On the server the key is not protected. """
    openssl_args = ['openssl',
                 'rsa',
                 '-in',
                 private_rsa_key,
                 '-out',
                 private_pem_cert,
                 '-passin',
                 'pass:' + passprhase]
    command = ' '.join(openssl_args)
    err = ''

    try:
        p = subprocess.Popen(openssl_args,
                             shell=False,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        out, err = p.communicate()
    except Exception as e:
        raise Exception('Failed to run \'{0}\': {1}'.format(command, e))
    if p.returncode != 0:
        raise Exception('Non-zero returncode ({0}) running \'{1}\': '
                        '{2}'.format(p.returncode,
                                     command,
                                     err.replace('\n', ' ')))
    if err.strip() != 'writing RSA key':
        raise Exception('Unexpected error message executing \'{0}\': '
                        '{1}'.format(command,
                                     err.replace('\n', ' ')))


def build_apca_pem_cert_request(root_path, machine_name, is_machine_function, is_renew):
    """ Build a request for a PEM certificate from APCA """
    pri_cert_path = certreq_pri_path.format(root_path, machine_name)
    rsa_key_path = certreq_rsa_path.format(root_path, machine_name)

    # Generate rsa key
    rsa = RSA.gen_key(rsa_key_bits, m2.RSA_F4)
    rsa.save_key(rsa_key_path,
                 cipher='aes_256_cbc',
                 callback=passphrase_callback)

    # Save the key unencrypted (works only if openssl is available)
    try:
        remove_passprhase(rsa_key_path,
                          pri_cert_path,
                          passphrase_callback(None))
    except Exception as e:
        # fallback to M2Crypto functions
        print 'Info: Failed to remove passphrase from private certificate ' \
              'using openssl. Falling back to M2Crypto function'
        fallback_m2 = True

    if fallback_m2:
        rsa.save_key(pri_cert_path,
                     cipher='aes_256_cbc',
                     callback=passphrase_callback)

    # Cleanup rsa key now that we're done with it
    secure_shred(rsa_key_path)

    # Generate certificate keypair
    pkey = EVP.PKey()
    pkey.assign_rsa(rsa)

    # Generate certificate request
    cert_request_pem = generate_cert_request_pem(pkey, machine_name)

    # format the request body
    certreq = cert_request_pem.replace(pem_req_header, '').replace(pem_req_trailer, '')
    post_data = unicode(certreq, 'utf-8')
    return post_data


def convert_pkcs_to_pem_pyopenssl(pkcs12_data):
    """ This function returns all certs in the PKCS12 data converted
        to PEM format using the pyopenssl crypto module or an empty
        string on error """
    allpemContent = ''

    try:
        from OpenSSL import crypto
        pkcs12 = crypto.load_pkcs12(pkcs12_data)
        certs = pkcs12.get_ca_certificates()
        for cert in certs:
            cert_pem = crypto.dump_certificate(crypto.FILETYPE_PEM, cert)
            allpemContent += cert_pem
            # We don't really want to write individual intermediate certs to
            # disk so we'll keep this commented out for now.
            # cert_filename = cert.get_subject().commonName.replace(' ', '_') + '.crt'
            # with open(cdir_path(cert_filename), 'w') as f:
            #    f.write(cert_pem)
    except Exception as e:
        print 'Error: convert_pkcs_to_pem_pyopenssl: {0}'.format(e)
        return ''

    return allpemContent


def convert_pkcs_to_pem_openssl(pkcs12_file):
    """ This function returns all certs in the PKCS12 file converted
        to PEM format using the pyopenssl crypto module or an empty
        string on error """
    openssl_args = ['openssl', 'pkcs12', '-in', pkcs12_file, '-passin', 'pass:']
    command = ' '.join(openssl_args)
    err = ''
    allpemContent = ''

    try:
        p = subprocess.Popen(openssl_args,
                             shell=False,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        out, err = p.communicate()
    except Exception as e:
        print 'Warning: convert_pkcs_to_pem_openssl: Failed to run \'{0}\': ' \
              '{1}'.format(command, e)
    if p.returncode != 0:
        print 'Warning: convert_pkcs_to_pem_openssl: Non-zero returncode ' \
              '({0}) running \'{1}\': {2}'.format(p.returncode,
                                                  command,
                                                  err.replace('\n', ' '))
    if err:
        print 'Warning: convert_pkcs_to_pem_openssl: Unexpected error ' \
              'message executing \'{0}\': {1}'.format(command,
                                                      err.replace('\n', ' '))

    f = None
    cn_hdr = 'subject=/CN='
    curr_cn = None
    for line in out.splitlines():
        if line.startswith(cn_hdr):
            curr_cn = line[len(cn_hdr):].replace(' ', '_')
            continue

        if curr_cn is None or line.startswith(pem_cert_header):
            continue

        # We don't really want to write individual intermediate certs to
        # disk so we'll keep this commented out for now.
        # f = open(cdir_path(curr_cn + '.crt'), 'w')
        # if f is not None:
        #     f.write(line + os.linesep)

        allpemContent += line + os.linesep

        if line.startswith(pem_cert_trailer):
            curr_cn = None
            # if f is not None:
            #     f.close()
            #     f = None

    return allpemContent


def request_certificate(root_path, machine_name, local_ip, apca_uri, is_machine_function, is_renew, ca_thumbprint=''):
    """ Request a certificate from APCA """

    # endpoint - the APCA endpoint for requests
    # machine_key - the private PEM file used for APCA request
    # machine_cert - the public PEM cert used for APCA request
    if is_renew:
        # use machine function certificate for renewals
        endpoint = apca_renew_cert_ep
        machine_key = default_machine_private_path.format(root_path, machine_name)
        machine_cert = default_machine_public_path.format(root_path, machine_name)
    else:
        if is_machine_function:
            # provision machine function cert using machine cert
            endpoint = apca_new_mf_cert_ep
            machine_key = default_machine_private_path.format(root_path, machine_name)
            machine_cert = default_machine_public_path.format(root_path, machine_name)
        else:
            # provision machine cert - can only be done during machine provisioning
            endpoint = apca_new_machine_cert_ep
            machine_key = None
            machine_cert = None

    # Build url for request
    params = {'useAppkiV2': 'true',
              'useShortLifetime': 'false',
              'requestedMachine': machine_name,
              # 'requestedMachineFunctionFqdns': 'APTest04.Autopilot.AzureCloudvSwitch',
              'requestedMachineFunctionFqdns': '',
              'requestedCAThumbprint': ca_thumbprint}
    url = endpoint + '?' + urllib.urlencode(params)

    post_data = build_apca_pem_cert_request(root_path,
                                            machine_name,
                                            is_machine_function,
                                            is_renew)

    # perform request
    try:
        response = apca_http_request(apca_uri,
                                     local_ip,
                                     url,
                                     'POST',
                                     post_data,
                                     machine_key,
                                     machine_cert)
    except Exception as e:
        print 'Error: request_certificate: {0}: {1}'.format(url, e)
        return False

    # add pem header and trailer to the response
    pem_cert = re.sub('(.{64})', '\\1\n', response, 0, re.DOTALL)
    if not pem_cert.startswith(pem_cert_header):
        pem_cert = '{0}\n{1}\n{2}'.format(pem_cert_header,
                                          pem_cert.strip(),
                                          pem_cert_trailer)

    print 'Info: PEM certificate received'
    tmp_pub_path = certreq_pub_path.format(root_path, machine_name)
    try:
        with open(tmp_pub_path, 'w') as pub_cert:
            pub_cert.write(pem_cert)
    except Exception as e:
        print 'Error: Unable to write public cert to ' \
              'file \'{0}\': {1}'.format(tmp_pub_path, e)
        return False

    # Validate certificate can be loaded
    try:
        cert = X509.load_cert(tmp_pub_path)
    except Exception as e:
        if is_machine_function:
            print 'Error: Requested machine function certificate ' \
                  '\'{0}\' corrupted: {1}'.format(tmp_pub_path, e)
        else:
            print 'Error: Requested machine certificate ' \
                  '\'{0}\' corrupted: {1}'.format(tmp_pub_path, e)
        return False

    # update old certificate files with new ones
    if is_machine_function:
        pub_cert_path = default_mf_public_path.format(root_path, machine_name)
        pri_cert_path = default_mf_private_path.format(root_path, machine_name)
    else:
        pub_cert_path = default_machine_public_path.format(root_path, machine_name)
        pri_cert_path = default_machine_private_path.format(root_path, machine_name)
    if not secure_file_move(tmp_pub_path, pub_cert_path, True):
        print 'Error: Failed to replace new public cert \'{0}\' with ' \
              'renewed cert \'{1}\''.format(pub_cert_path, tmp_pub_path)
        return False
    tmp_pri_path = certreq_pri_path.format(root_path, machine_name)
    if not secure_file_move(tmp_pri_path, pri_cert_path, True):
        print 'Error: Failed to replace new private cert \'{0}\' with ' \
              'renewed cert \'{1}\''.format(pri_cert_path, tmp_pri_path)
        return False

    return True


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Keeps the certificates issued by APCA up to date.')
    parser.add_argument('-n', '--name', metavar='MACHINE_NAME',
                        help='Name of the device requesting the certificate. Optional. Defaults to ' \
                             '$AP_MACHINE_NAME variable then to the machine hostname')
    parser.add_argument('-i', '--interface', metavar='INTERFACE',
                        help='Local interface IP address to bound to. Optional. Defaults to any')
    parser.add_argument('-c', '--cluster', metavar='CLUSTER',
                        help='Machine cluster. Optional. Defaults to $AP_CLUSTER variable then to \'C03c\'')
    parser.add_argument('-w', '--daystorenewal', metavar='RENEWAL_DAYS',
                        help='Specifies the number of days prior to cert expiration to request a ' \
                             'renewal. Optional. Defaults to 12')
    parser.add_argument('-b', '--rootdir', metavar='ROOT_DIR',
                        help='Specifies the root directory of the filesystem. Optional. Defaults to \'/\'')

    args = parser.parse_args()

    hostname = args.name
    local_ip_address = args.interface
    cluster = args.cluster
    days_to_renewal = 12
    root_dir = args.rootdir

    force_renew_machine_cert = False
    force_renew_mf_cert = False
    force_update_cert_chain = False

    if not root_dir:
        root_dir = '/'
    elif not os.path.isdir(root_dir):
        parser.print_help()
        sys.exit(2)

    if not hostname:
        try:
            hostname = os.environ['AP_MACHINE_NAME']
        except:
            try:
                hostname = socket.gethostname()
            except:
                print 'Error: No machine name specified and unable to get ' \
                      'machine hostname'
                parser.print_help()
                sys.exit(2)

    if not local_ip_address:
        local_ip_address = ''

    if not cluster:
        try:
            cluster = os.environ['AP_CLUSTER']
        except:
            cluster = 'Co3c'

    if not args.daystorenewal:
        days_to_renewal = 12
    else:
        try:
            days_to_renewal = int(args.daystorenewal)
        except:
            days_to_renewal = 12
    if days_to_renewal < 1:
        days_to_renewal = 1

    apca_vip = get_apca_vip(cluster)
    cert_machine_public = default_machine_public_path.format(root_dir, hostname)
    cert_machine_private = default_machine_private_path.format(root_dir, hostname)
    cert_mf_public = default_mf_public_path.format(root_dir, hostname)
    cert_mf_private = default_mf_private_path.format(root_dir, hostname)

    # secure certificates stored in legacy location by Stage2 of
    # machine provisioning. we'll attempt to continue and just
    # assume the old certificates have been compromised
    if not secure_legacy_paths(hostname, root_dir):
        print 'Warning: Failed to secure legacy certificate locations'

    try:
        cleanup_apca_directory(root_dir)
    except Exception as e:
        print 'Warning: Failed to cleanup apca directory: {0}'.format(e)

    # download list of revoked certificates
    if not refresh_disallowed_store(local_ip_address, root_dir):
        print 'Warning: Failed to refreshed list of revoked certificates'
    else:
        print 'Info: List of revoked certificates has been refreshed'

    # download intermediate certificates if they don't exist
    update_cert_chain = False
    if not os.path.isfile(default_intermediate_path):
        print 'Info: First time provisioning of APCA certificates'
        update_cert_chain = True
    elif force_update_cert_chain:
        print 'Info: Force update of APCA certificates'
        update_cert_chain = True

    if update_cert_chain:
        if not update_appki_cert_chain(local_ip_address, apca_vip, root_dir):
            print 'Error: Unable to provision CA certificates'
            sys.exit(3)

    # check to see if we already have a machine certificate
    has_machine_cert = False
    if os.path.isfile(cert_machine_public) and os.path.isfile(cert_machine_private):
        has_machine_cert = True
    elif (os.path.isfile(legacy_machine_path.format(root_dir, hostname + '.key')) and
          os.path.isfile(legacy_machine_path.format(root_dir, hostname + '.crt'))):
        # let this pass through and pull a new cert and just assume the old
        # certificates have been compromised
        print 'Warning: Machine certificate only exists in an insecure ' \
              'legacy location. Requesting a new certificate. This ' \
              'should only be requested within the first 30 minutes of ' \
              'machine provisioning'
    else:
        print 'Info: Machine certificate not found. Requesting it. This ' \
              'should only be requested within the first 30 minutes of ' \
              'machine provisioning'

    # check to see if we already have a machine function certificate
    has_mf_cert = False
    if os.path.isfile(cert_mf_public) and os.path.isfile(cert_mf_private):
        has_mf_cert = True
    elif (os.path.isfile(legacy_mf_path.format(root_dir, hostname + '.key')) and
          os.path.isfile(legacy_mf_path.format(root_dir, hostname + '.crt'))):
        # let this pass through and pull a new cert and just assume the old
        # certificates have been compromised
        print 'Warning: Machine function certificate only exists in an ' \
              'insecure legacy location. Getting a new certificate...'
    else:
        print 'Info: Generating first machine function certificate'

    # get machine certificate expiration time
    new_machine_cert = False
    renew_machine_cert = False
    if has_machine_cert:
        cert_expiration = read_cert_expiration_date(cert_machine_public)
        if cert_expiration is None:
            cert_expiration = datetime.utcnow()
        days_remaining = (cert_expiration - datetime.utcnow()).days
        print 'Info: Machine certificate expiration date: {0}. Days until ' \
              'expiration: {1}'.format(cert_expiration, days_remaining)

        # -1 if expired, 0 if renewal is needed, 1 if no action required
        machine_expiration_status = check_cert_state(cert_expiration, days_to_renewal)
        if machine_expiration_status < 0:
            print 'Info: Machine certificate has expired. Getting ' \
                  'a new certificate...'
            new_machine_cert = True
        elif machine_expiration_status == 0:
            print 'Info: Machine certificate is near expiration. ' \
                  'Requesting a renewal...'
            renew_machine_cert = True
        elif force_renew_machine_cert:
            print 'Info: Force renewal of machine certificate'
            renew_machine_cert = True
    else:
        new_machine_cert = True

    # get new machine certificate
    if new_machine_cert:
        if not request_certificate(root_dir,
                                   hostname,
                                   local_ip_address,
                                   apca_vip,
                                   False,
                                   False):
            print 'Error: Failed to request machine certificate'
            sys.exit(3)

    # renew machine certificate
    if renew_machine_cert:
        if not request_certificate(root_dir,
                                   hostname,
                                   local_ip_address,
                                   apca_vip,
                                   False,
                                   True):
            print 'Error: Failed to renew machine certificate'
            sys.exit(3)

    # get machine function certificate expiration time
    new_mf_cert = False
    renew_mf_cert = False
    if has_mf_cert:
        cert_expiration = read_cert_expiration_date(cert_mf_public)
        if cert_expiration is None:
            cert_expiration = datetime.utcnow()
        days_remaining = (cert_expiration - datetime.utcnow()).days
        print 'Info: Machine function certificate expiration date: {0}. Days ' \
              'until expiration: {1}'.format(cert_expiration, days_remaining)

        # -1 if expired, 0 if renewal is needed, 1 if no action required
        mf_expiration_status = check_cert_state(cert_expiration, days_to_renewal)
        if mf_expiration_status < 0:
            print 'Info: Machine function certificate has expired. Getting ' \
                  'a new certificate...'
            new_mf_cert = True
        elif mf_expiration_status == 0:
            print 'Info: Machine function certificate is near expiration. ' \
                  'Requesting a renewal...'
            renew_mf_cert = True
        elif force_renew_mf_cert:
            print 'Info: Force renewal of machine function certificate'
            renew_mf_cert = True
    else:
        new_mf_cert = True

    # generate new machine function certificate if we don't have one
    # on the machine
    if new_mf_cert:
        if not request_certificate(root_dir,
                                   hostname,
                                   local_ip_address,
                                   apca_vip,
                                   True,
                                   False):
            print 'Error: Failed to generate machine function certificate'
            sys.exit(3)

    # renew machine function certificate
    if renew_mf_cert:
        if not request_certificate(root_dir,
                                   hostname,
                                   local_ip_address,
                                   apca_vip,
                                   True,
                                   True):
            print 'Error: Failed to renew machine function certificate'
            sys.exit(3)
