#!/usr/bin/env python3

# Class used to aggregate watchdog data and update DM

import sys
import os
import glob
import socket
import time

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apsdk import apenv
from apsdk import dmclient


class MachineLocalWatchdog(object):
    DM_RETRY_COUNT = 4

    def __init__(self,
                 watchdog_csv_dir: str,
                 output_csv_file: str,
                 dm_cmd: str,
                 test_status: bool=False):

        self.test_status = test_status
        self.watchdog_csv_dir = watchdog_csv_dir
        self.output_csv_file = output_csv_file
        self.dm_cmd = dm_cmd

    # Merge all result .csv files from all watchdogs into one file
    def update_watchdogstatus_file(self) -> bool:
        if os.path.exists(self.watchdog_csv_dir):
            # Search for all .csv files in the watchdog csv dir and merge them together
            read_files = glob.glob(self.watchdog_csv_dir + "/*.csv")
            if len(read_files) > 0:
                try:
                    with open(self.output_csv_file, "wb") as outfile:
                        for f in read_files:
                            # print("Reading in: " + f)
                            with open(f, "rb") as infile:
                                outfile.write(infile.read())
                except Exception as e:
                    print("machinelocalwatchdog.update_watchdogstatus_file: Exception: " + str(e))
                    return False
            else:
                if os.path.isfile(self.output_csv_file):
                    # No watchdog files found, probably because this is the initial run after system is configured
                    # Upload a warning to DM to let us know this happened
                    try:
                        with open(self.output_csv_file, "w") as outfile:
                            hostname = socket.gethostname()
                            output = hostname + ",Warning,watchapservice,Watchdog files missing from: " + self.watchdog_csv_dir
                            outfile.write(output)
                    except Exception as e:
                        print("machinelocalwatchdog.update_watchdogstatus_file: Error Writing to output_csv_file: " + str(e))
                        return False
                else:
                    print("machinelocalwatchdog.update_watchdogstatus_file: output_csv_file does not exist")
                    return False
        else:
            print("machinelocalwatchdog.update_watchdogstatus_file: csv_sources dir does not exist")
            return False

        return True

    # Push .csv file with all the machine properties to DM
    def update_dm(self) -> bool:
        if os.path.isfile(self.output_csv_file):
            with open(self.output_csv_file, 'r', newline=None) as infile:
                payload = infile.read()
        else:
            print("machinelocalwatchdog.update_dm: Upload file not found at: " + self.output_csv_file)
            return False

        if not self.test_status:
            dm_client = dmclient.DMClient('')

        # Retry 4 times in case DM isn't responding properly
        count = 0
        while count < self.DM_RETRY_COUNT:
            if not self.test_status:
                resp = dm_client.send_request(self.dm_cmd, payload)
                response = resp.text
            else:
                response = "OK"

            if "OK" in response:
                break
            else:
                count += 1
                time.sleep(10)
        else:
            return False

        return True

    # Execute all the methods in order
    def main(self):
        print("\nStarting machinelocalwatchdog.py")
        print("\nUpdating localwatchdogstatus.csv")

        if not self.update_watchdogstatus_file():
            print("machinelocalwatchdog.main: Could not get csv files from csv_sources dir (main.update_watchdogstatus_file)")
            sys.exit(os.EX_OK)

        if not self.update_dm():
            print("machinelocalwatchdog.main: Could not update DM with latest .csv status file(main.update_dm)")
            sys.exit(os.EX_OK)

        print("\nCompleted machinelocalwatchdog.py")
        sys.exit(os.EX_OK)

if __name__ == '__main__':
    watchdog_csv_dir = apenv.get_data_root() + "/watchdogoutput/csv_sources"
    output_csv_file = apenv.get_data_root() + "/watchdogoutput/localwatchdogstatus.csv"
    dm_cmd = "UpdateMachineProperty"

    MachineLocalWatchdog(watchdog_csv_dir, output_csv_file, dm_cmd).main()
