#!/usr/bin/env python3

# Watchdog used to check certificate expiration and report on it

import sys
import os
import socket
import io
import OpenSSL
import datetime
import dateutil.parser
import platform
import subprocess
from cryptography import x509
from cryptography.hazmat.backends import default_backend

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from aptools import statsdhelper
from apsdk import apenv


class WatchAPCertificates(object):
    def __init__(self,
                 machine_crt_file: str,
                 machine_function_crt_file: str,
                 watchdog_csv_dir: str,
                 watchdog_tmp_file: str,
                 watchdog_csv_file: str,
                 test_status: bool=False):

        # If we are in unit test mode we deal with certain situations
        # which we can't replicate with generic data
        self.test_status = test_status

        # Path to local certificate files
        self.machine_crt_file = machine_crt_file
        self.machine_function_crt_file = machine_function_crt_file

        # machinelocalwatchdog folder and files to use for DM report
        self.watchdog_csv_dir = watchdog_csv_dir
        self.watchdog_tmp_file = watchdog_tmp_file
        self.watchdog_csv_file = watchdog_csv_file

        # Get hostname
        self.hostname = socket.gethostname()

        # Create objects used to obtain information from .crt files
        # Get certificate files read as bytes into memory
        self.machine_crt_bytes = bytes
        self.machine_function_crt_bytes = bytes

        # We store the certificate objects in dictionaries for easy access
        # DO NOT change the key names UNLESS you understand that those names are reported to MDM
        # as actual metric names!  If you change machine or machineFunction to something else
        # the actual metric name pushed to MDM will change also!
        self.cert_objects = {}
        self.cert_objects["machine"] = {}
        self.cert_objects["machineFunction"] = {}
        self.cert_results = {}
        self.cert_results["machine"] = {}
        self.cert_results["machineFunction"] = {}

        # Store the SAN names found
        self.machine_dns_names = ""
        self.machine_function_dns_names = ""

        # Expiration days for each cert
        self.machine_days_remaining = 0
        self.machine_function_days_remaining = 0

        # MDM Namespace to report stats to
        self.namespace = "APSharedServices"

        # Only load statsdhelper for MDM reporting if we're NOT doing unit tests
        if not self.test_status:
            self.metrics = statsdhelper.StatsdHelper(statsd_port=9125)

    # Read in the raw bytes for machine and machine_function certificates
    def read_in_certs(self) -> bool:
        try:
            with open(self.machine_crt_file, 'rb') as content_file:
                self.machine_crt_bytes = content_file.read()

        except Exception as e:
            print("watchapcertificates.read_in_certs: Exception: " + str(e))
            return False

        try:
            with open(self.machine_function_crt_file, 'rb') as content_file:
                self.machine_function_crt_bytes = content_file.read()

        except Exception as e:
            print("watchapcertificates.read_in_certs: Exception: " + str(e))
            return False

        return True

    # Get certificate objects used to obtain data by loading certs into pyopenssl and cryptography modules
    # We can 2 separate objects because the cryptography module does not provide an unambigious Zulu time.
    # Therefore we need to use OpenSSL module instead for expiration.

    # However, OpenSSL module does not provide a way to query SAN properly so we use the cryptography module
    # to get the full SAN information
    def get_cert_objects(self) -> bool:
        try:
            c = OpenSSL.crypto
            self.cert_objects["machine"]["openssl_object"] = c.load_certificate(c.FILETYPE_PEM, self.machine_crt_bytes)
            self.cert_objects["machine"]["crypto_object"] = x509.load_pem_x509_certificate(self.machine_crt_bytes, default_backend())
            self.cert_objects["machine"]["file_path"] = self.machine_crt_file

            c = OpenSSL.crypto
            self.cert_objects["machineFunction"]["openssl_object"] = c.load_certificate(c.FILETYPE_PEM, self.machine_function_crt_bytes)
            self.cert_objects["machineFunction"]["crypto_object"] = x509.load_pem_x509_certificate(self.machine_function_crt_bytes, default_backend())
            self.cert_objects["machineFunction"]["file_path"] = self.machine_function_crt_file

        except Exception as e:
            print("watchapcertificates.get_cert_objects: Exception: " + str(e))
            return False

        return True

    # So, we have a problem where we added the virtual environment and it contains characters that are illegal
    #  per the specification for CN or ALT Names for DNS certificates.  The backslash character is specifically
    # NOT ALLOWED in DNS names per RFC3490. This means that our MACHINE_FUNCTION certificate will fail to parse
    # going forward using the python libraries because python uses IDNA to verify character integrity and the
    # backslash character always fails.
    #
    # Instead, we are going with the less ideal but simpler commandline openssl utility to get the same results.
    # We are STILL continuing to use the python library for certificate expiration days as that functionality can't
    # violate the spec so easily.
    def process_all_dns_alt_names(self) -> bool:
        try:
            if platform.system() != "Windows":
                for key, value in self.cert_objects.items():
                    cert_location = self.cert_objects[key]["file_path"]

                    command = 'openssl x509 -inform PEM -text -noout -in ' + cert_location + ' | grep "DNS:"'

                    try:
                        result = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
                    except subprocess.TimeoutExpired:
                        return False
                    except subprocess.CalledProcessError:
                        return False

                    result_list = result.strip().split(",")

                    result_dns_names = []
                    for item in result_list:
                        result_dns_names.append(item.strip().split("DNS:")[1].upper())

                    self.cert_results[key]["dns_alt_names"] = "|".join(result_dns_names)

            else:
                    self.cert_results["machine"]["dns_alt_names"] = "CO4PHY050071438.CO4|LINUX.AZLINUX-DEV-CO4.CO4|LINUX.AUTOPILOT\AUTOPILOTCLIENT-LINUX.CO4"
                    self.cert_results["machineFunction"]["dns_alt_names"] = "CO4PHY050071438.CO4"
                    return True

        except Exception as e:
            print("watchapcertificates.get_all_dns_alt_names: Exception: " + str(e))
            return False

        return True

    ##########################################################################################################
    # LEFT HERE in case solution presents itself in future for pythonic way to do this without IDNA problems #
    ##########################################################################################################
    # # Use the cryptography module to get the SAN names for the certs
    # def get_all_dns_alt_names(self) -> bool:
    #
    #     if not self.test_status:
    #         san_results_upper = []
    #
    #         for key, value in self.cert_objects.items():
    #             for key2, value2 in value.items():
    #                 if key2 == "crypto_object":
    #                     # Get the SAN extension class with all the info in it
    #                     san_values = self.cert_objects[key][key2].extensions.get_extension_for_oid(ExtensionOID.SUBJECT_ALTERNATIVE_NAME).value
    #                     print(san_values)
    #                     # Get the actual SAN names in a list
    #                     san_results = san_values.get_values_for_type(x509.extensions.GeneralName)
    #                     print(san_results)
    #                     # Uppercase all the names in the list since they show up lowercase for some reason
    #                     san_results_upper = [x.upper() for x in san_results]
    #                     print(san_results_upper)
    #                 elif key2 == "dns_alt_names":
    #                     # Join them with a | so that they are in the correct DM format
    #                     self.cert_results[key][key2] = "|".join(san_results_upper)
    #                     print(self.cert_results)
    #                 else:
    #                     pass
    #     else:
    #         self.cert_results["machine"]["dns_alt_names"] = "CO4PHY050080648.CO4|LINUX.azlinux-Dev-CO4.CO4"
    #         self.cert_results["machineFunction"]["dns_alt_names"] = "CO4PHY050080648.CO4"
    #         return True
    #
    #     return True

    # Return ALL DNS ALT NAMES for unit testing
    def get_all_dns_alt_names(self) -> tuple:
        return (self.cert_results["machine"]["dns_alt_names"], self.cert_results["machineFunction"]["dns_alt_names"])

    # Use the openssl object to get the days remaining on the certs, this module is used in favor of
    # cryptography because the datetime object returned needs to be unambiguous, otherwise we will not be accurately
    # comparing UTC to UTC
    def process_cert_days_remaining(self) -> bool:
        try:
            # Go through each cert object and get the expiration datetime
            for key, value in self.cert_objects.items():
                # Reset expiration time every iteration
                cert_expiry_utc_time = ""

                for key2, value2 in value.items():
                    if key2 == "crypto_object":
                        cert_expiry_utc_time = str(self.cert_objects[key]["openssl_object"].get_notAfter().decode("utf-8"))
                    else:
                        pass

                if cert_expiry_utc_time:
                    cert_expiry_utc_time_parsed = dateutil.parser.parse(cert_expiry_utc_time)
                else:
                    return False

                time_remaining = cert_expiry_utc_time_parsed - datetime.datetime.now(datetime.timezone.utc)

                # Get the days remaining for reporting
                days_remaining = time_remaining.days

                # Make sure the cert hasn't already expired and if it has report a value of 0 and not some crazy
                # negative days or something
                if not days_remaining > 0:
                    days_remaining = 0

                self.cert_results[key]["cert_expiry_utc_time"] = str(cert_expiry_utc_time)
                self.cert_results[key]["days_remaining"] = days_remaining

        except Exception as e:
            print("watchapcertificates.get_cert_days_remaining: Exception: " + str(e))
            return False

        return True

    # Create DM report out of the information returned
    def build_dm_report(self) -> bool:
        self.dm_output = io.StringIO()

        for key, value in self.cert_results.items():
            report = "{Cert is APPKIv2;DNS alt name(s) " + str(self.cert_results[key]["dns_alt_names"]) + ";" + str(
                self.cert_results[key]["days_remaining"]) + " days left for expiry;}"

            # CO4PHY050080648,Ok,AppkiMachineCertStatus,{Cert is APPKIv2;DNS alt name(s) CO4PHY080090121.CO4;59 days left for expiry;}
            # CO4PHY050080648,Ok,AppkiMachineFunctionCertStatus,{Cert is APPKIv2;DNS alt name(s) CO4PHY080090121.CO4|LINUX.azlinux-Dev-CO4.CO4;59 days left for expiry;}
            if key == "machine":
                property = "AppkiMachineCertStatus"
            elif key == "machineFunction":
                property = "AppkiMachineFunctionCertStatus"
            else:
                return False

            line = [self.hostname, "Ok", property, report]

            try:
                self.dm_output.write(",".join(line) + '\r\n')
            except Exception as e:
                print("watchapcertificates.build_dm_report: Exception: " + str(e))
                return False

        return True

    # Make sure the DM report is properly formatted
    def verify_dm_report(self):
        report = self.dm_output.getvalue().split("\r\n")

        for line in list(filter(None, report)):
            results = line.split(",")

            # Each line must have 4 columns separated by commas
            # CO4PHY050080648,Ok,AppkiMachineCertStatus,{Cert is APPKIv2;DNS alt name(s) CO4PHY080090121.CO4;59 days left for expiry;}
            # CO4PHY050080648,Ok,AppkiMachineFunctionCertStatus,{Cert is APPKIv2;DNS alt name(s) CO4PHY080090121.CO4|LINUX.azlinux-Dev-CO4.CO4;59 days left for expiry;}
            if not len(results) == 4:
                print("watchapcertificates.verify_dm_report: Number of expected columns is incorrect")
                return False

        return True

    # Write DM report to a tmp file
    def write_dm_report_to_disk(self) -> bool:
        if not os.path.isdir(self.watchdog_csv_dir):
            print("watchapcertificates.write_dm_report_to_disk: Directory: " + self.watchdog_csv_dir + " not found")
            return False
        else:
            try:
                csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

                with open(csv_file_path, "w") as text_file:
                    text_file.write(self.dm_output.getvalue())

            except Exception as e:
                print("watchapcertificates.write_dm_report_to_disk: Exception: " + str(e))
                return False

        return True

    # Verify that the tmp file on disk is the same as the data that was supposed to be written
    def verify_write_dm_report_to_disk(self) -> bool:
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

        if os.path.exists(csv_file_path):
            with open(csv_file_path, 'r') as content_file:
                content = content_file.read()

                content = content.split("\n")
                dm_output = self.dm_output.getvalue().split("\r\n")

                if list(filter(None, content)) == list(filter(None, dm_output)):
                    return True
                else:
                    print("watchapcertificates.verify_write_dm_report_to_disk: csv content does not match dm_output content")
                    return False

        else:
            return False

    # Atomically rename the .tmp file to a .csv file which will be ingested by machinelocalwatchdog.py
    def atomically_rename_dm_report_on_disk(self) -> bool:
        tmp_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_csv_file)

        if platform.system() == "Windows" and self.test_status:
            return True

        try:
            os.rename(tmp_file_path, csv_file_path)

        except Exception as e:
            print("watchapcertificates.atomically_rename_dm_report_on_disk: Exception: " + str(e))
            return False

        return True

    # Print out the DM report
    def print_dm_report(self):

        output = "\n".join(self.dm_output.getvalue().split("\r\n"))
        self.dm_output.close()

        try:
            print("\nDM Report:")
            print(output)
        except Exception as e:
            print("watchapcertificates.print_dm_report: Exception: " + str(e))
            return False

        return True

    # Send out statsd data to MDM
    def send_statsd_report(self) -> bool:

        # Send multi dim metric
        if not self.test_status:
            for key, value in self.cert_results.items():
                if not self.metrics.send_gauge(self.namespace,
                                               key + "CertExpiration",
                                               self.cert_results[key]["days_remaining"]):
                    return False

        return True

    # Execute all the methods in order
    def main(self):
        if not self.read_in_certs():
            print("watchapcertificates.main: Could not get Read Certificates (main.read_in_certs)")
            sys.exit(1)

        if not self.get_cert_objects():
            print("watchapcertificates.main: Could not get Certificate Objects (main.get_cert_objects)")
            sys.exit(1)

        if not self.process_all_dns_alt_names():
            print("watchapcertificates.main: Could not get Certificate Objects (main.get_all_dns_alt_names)")
            sys.exit(1)

        if not self.process_cert_days_remaining():
            print("watchapcertificates.main: Could not get Cert Expiration Days Remaining (main.get_cert_days_remaining)")
            sys.exit(1)

        if not self.build_dm_report():
            print("watchapcertificates.main: Could not build DM report (main.build_dm_report)")
            sys.exit(1)

        if not self.verify_dm_report():
            print("watchapcertificates.main: Could not verify DM report (main.verify_dm_report)")
            sys.exit(1)

        if not self.write_dm_report_to_disk():
            print("watchapcertificates.main: Could not write DM report to disk (main.write_dm_report_to_disk)")
            sys.exit(1)

        if not self.verify_write_dm_report_to_disk():
            print("watchapcertificates.main: Could not verify DM report written to disk (main.verify_write_dm_report_to_disk)")
            sys.exit(1)

        if not self.atomically_rename_dm_report_on_disk():
            print(
                "watchapcertificates.main: Could not atomically rename dm report file on disk (main.atomically_rename_dm_report_on_disk)")
            sys.exit(1)

        # if not self.print_dm_report():
        #     print("watchapcertificates.main: Could not print DM report (main.print_dm_report)")
        #     sys.exit(1)

        if not self.send_statsd_report():
            print("watchapcertificates.main: Could not send statsd report (main.send_statsd_report)")
            sys.exit(1)

        sys.exit(0)  # EX_OK = everything executed normally

if __name__ == '__main__':
    machine_crt_file, machine_crt_key = apenv.get_machine_cert()
    machine_function_crt_file, machine_function_key = apenv.get_machine_function_cert()
    watchdog_csv_dir = "/ap/data/watchdogoutput/csv_sources"
    watchdog_tmp_file = "watchapcertificates.tmp"
    watchdog_csv_file = "watchapcertificates.csv"

    WatchAPCertificates(machine_crt_file, machine_function_crt_file, watchdog_csv_dir, watchdog_tmp_file, watchdog_csv_file).main()
