#!/usr/bin/env python3

# Watchdog used to check Hard Drives and report on their status

import sys
import os
import socket
import subprocess
import io
import re
import math
import psutil
from collections import OrderedDict


class WatchAPHDD(object):
    DEBUG = False

    def __init__(self, watchdog_csv_dir: str, watchdog_tmp_file: str, watchdog_csv_file: str):
        # machinelocalwatchdog folder and files to use for DM report
        self.watchdog_csv_dir = watchdog_csv_dir
        self.watchdog_tmp_file = watchdog_tmp_file
        self.watchdog_csv_file = watchdog_csv_file

        # Get hostname
        self.hostname = socket.gethostname()

        # Partitions to get % disk usage from
        self.partitions_to_monitor = {"0": {"id": "0", "partition": "/", "percent_free": ""},
                                      "1": {"id": "1", "partition": "/ap", "percent_free": ""}}

        # Store raw devices
        self.raw_devices = []

        # Store processed results
        self.device_results = {}

        # StringIO object to store dm_output data
        self.dm_output = io.StringIO()

    # Use lsblk to determine what block devices are in the system
    def get_block_devices(self) -> str:
        command = "lsblk -n -o name,type | grep disk"

        # Example lsblk output
        # sda  disk
        # sdb  disk
        # sdc  disk
        # nvme0n1  disk
        # nvme1n1  disk
        # nvme2n1  disk
        # nvme3n1  disk

        if self.DEBUG:
            print("\nwatchaphdd:get_block_devices Command: " + command)

        try:
            result = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired:
            raise ValueError('watchaphdd:get_block_devices subprocess.TimeoutExpired')
        except subprocess.CalledProcessError:
            raise ValueError('watchaphdd:get_block_devices subprocess.CalledProcessError')

        return str(result)

    # Use lsblk to determine what block devices are in the system
    def process_block_devices(self) -> bool:
        try:
            result = self.get_block_devices()
        except Exception as e:
            print("watchaphdd.process_block_devices: Exception: " + str(e))
            return False

        raw_devices_list = result.strip().split("\n")
        if self.DEBUG:
            print("\nRaw Block Devices:")
            print(raw_devices_list)

        raw_devices = {}
        for device_data in raw_devices_list:
            raw_devices[device_data.strip().split()[0]] = device_data.strip().split()[1]

        # Make sure devices aren't all null
        device_found = False
        for device, type in raw_devices.items():
            if device:
                device_found = True

        if not device_found:
            return False

        # Make sure devices exclude anything that's not "disk" type
        # Even though we filter this out with original command using grep, this allows us to test additional error
        # conditions in the unit test to make sure non-"disk" types are filtered out
        for device in list(raw_devices):
            device_type = raw_devices[device]
            if device_type != "disk":
                if self.DEBUG:
                    print("Found device: " + device + " of type: " + device_type + " which is not type disk: Removing")
                del raw_devices[device]

        # Make sure more than 1 device is found
        if len(raw_devices) < 2:
            return False
        else:
            self.raw_devices = raw_devices

        if self.DEBUG:
            print("\nStripped Raw Block Devices:")
            print(self.raw_devices)

        return True

    # Get specific block device type for device
    def get_block_device_type(self, device: str) -> str:
        command = "cat /sys/block/" + device + "/queue/rotational"

        if self.DEBUG:
            print("watchaphdd:get_block_device_type " + command)
        try:
            result = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired:
            raise ValueError('main.get_block_device_type exception thrown subprocess.TimeoutExpired')
        except subprocess.CalledProcessError:
            raise ValueError('main.get_block_device_type exception thrown subprocess.CalledProcessError')

        return str(result)

    # Check to see if the device is rotational or not
    def get_types_for_block_devices(self) -> bool:
        for device in list(filter(None, self.raw_devices)):
            # Determine whether the device is an SSD or an HDD by probing the rotational parameter
            if self.DEBUG:
                print("\nwatchaphdd:get_types_for_block_devices Getting device type information on /dev/" + device)

            try:
                result = self.get_block_device_type(device)
            except Exception as e:
                print("watchaphdd.get_types_for_block_devices: Exception: " + str(e))
                return False

            device_type_value = result.split("\n")[0]

            if not device_type_value:
                print("\nwatchaphdd:get_types_for_block_devices Unable to determine HDD vs SSD for device: " + device)
                return False

            if (device_type_value == "1"):
                device_type = "HDD"
                if self.DEBUG:
                    print("watchaphdd:get_types_for_block_devices Found HDD rotational device")
            else:
                device_type = "SSD"
                if self.DEBUG:
                    print("watchaphdd:get_types_for_block_devices Found SSD device")

            self.device_results[device] = {"type": device_type}

        return True

    # Get the name of the device Bus ID using lsscsi for normal bus drives
    def get_lsscsi_device_bus_data(self) -> str:
        command = "lsscsi --verbose | egrep " + r'"^\["' + r" | awk '{ print $1 " + '" "' + r" $NF }'"
        if self.DEBUG:
            print("\nwatchaphdd:get_lsscsi_device_bus Running lsscsi command: " + command)

        try:
            result = str(subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True))
        except subprocess.TimeoutExpired:
            raise ValueError('main.get_lsscsi_device_bus exception thrown subprocess.TimeoutExpired')
        except subprocess.CalledProcessError:
            raise ValueError('main.get_lsscsi_device_bus exception thrown subprocess.CalledProcessError')

        return str(result)

    # Get the name of the pseudo device Bus ID using lsblk for NVMe drives (PCIe)
    def get_nvme_lsblk_pseudo_bus_data(self) -> str:
        command = "lsblk -l | egrep ^nvme[0-9]+n[0-9][^p]+.*disk.*$ | awk '{ print $2 " + '" "' + r" $1 }'"
        if self.DEBUG:
            print("\nwatchaphdd:get_nvme_blkid_pseudo_bus_data Running lsblk command: " + command)

        try:
            result = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired:
            raise ValueError('main.get_nvme_lsblk_pseudo_bus_data exception thrown subprocess.TimeoutExpired')
        except subprocess.CalledProcessError:
            raise ValueError('main.get_nvme_lsblk_pseudo_bus_data exception thrown subprocess.CalledProcessError')

        return str(result)

    # Sort the devices into the dictionary based on their Bus data
    def process_device_bus_data(self) -> bool:
        try:
            result_lsscsi = self.get_lsscsi_device_bus_data()
        except ValueError:
            return False

        try:
            result_nvme = self.get_nvme_lsblk_pseudo_bus_data()
        except ValueError:
            return False

        result = result_lsscsi + result_nvme

        device_list = result.split("\n")

        # Example Device List - all kinds of crazy numbering and bogus values for device names are possible
        # We make sure device names are valid and all 4 bus numbers are reported
        # [0:0:0:0] -
        # [1:0:1:0] /dev/sda
        # [2:0:2:0] /dev/sdb
        # [3:0:3:0] /dev/sdc
        # [0:0:4:0] /dev/sdd
        # [0:0:5:0] /dev/sde
        # [0:0:6:0] /dev/sdf
        # [0:0:7:0] -
        # 259:5 nvme0n1
        # 259:1 nvme1n1
        # 259:0 nvme2n1
        # 259:2 nvme3n1
        # 259:3 nvme4n1
        # 259:4 nvme5n1

        for device in list(filter(None, device_list)):

            # Make sure the bus format is in expected notation, otherwise we return False and error out
            # Format is this: [0:0:0:0]
            raw_bus = device.split(" ")[0]
            raw_device = device.split(" ")[1]

            nvme_pattern = re.compile('^nvme.*$')
            match_nvme = nvme_pattern.search(raw_device)

            if match_nvme:
                # Prepend NVMe devices with /dev/ since they don't have that by default
                raw_device = "/dev/" + raw_device
                # Make sure NVMe bus has 4 octets like the normal scsci buses so they match
                # Append :0 values to each octet if necessary
                raw_bus_append = ""
                raw_bus_octet_count = raw_bus.count(':')
                if raw_bus_octet_count < 4:
                    for i in (1, raw_bus_octet_count):
                        raw_bus_append += ":0"
                raw_bus += raw_bus_append
                raw_bus = "[" + raw_bus + "]"

            if self.DEBUG:
                print(raw_device)
                print("  Raw Bus: " + raw_bus)
                print("  Numeric Bus: " + "".join(raw_bus[1:-1].split(":")))

            # Example Bus Output
            # 0:0:0:0
            # 1:0:0:0
            # 2:0:0:0
            # 3:0:0:0
            # 0:0:14:0
            # 0:0:0:0
            # 0:4:3:0
            # 259:5:0:0
            # 259:1:0:0
            # 259:0:0:0
            # 259:2:0:0
            # 259:3:0:0
            # 259:4:0:0

            bus_pattern = re.compile('^\[[0-9]+:[0-9]+:[0-9]+:[0-9]+\]$')
            match_bus = bus_pattern.search(raw_bus)

            if not raw_bus or not match_bus:
                if self.DEBUG:
                    print("Bus missing or malformed: " + raw_bus)
                return False

            raw_device_pattern = re.compile('\/dev\/[a-zA-Z0-9]+')
            raw_device_match = raw_device_pattern.search(raw_device)

            if raw_device_match:
                # ['', 'dev', 'sda']
                device = raw_device.split("/")[2]
                if self.DEBUG:
                    print("Device: " + device)

                device_pattern = re.compile('[a-zA-Z0-9]+')
                device_match = device_pattern.search(device)

                if device_match:
                    if device.strip() in self.device_results:
                        for key, value in self.device_results.items():
                            if key.strip() == device.strip():
                                if self.DEBUG:
                                    print("Found matching device key: " + key.strip() + " in self.device_results")
                                self.device_results[device]["bus"] = raw_bus
                                # Used to sort by bus when reporting/displaying output
                                self.device_results[device]["numeric_bus"] = "".join(raw_bus[1:-1].split(":"))
                else:
                    if self.DEBUG:
                        print("Skipping device: " + device)
            else:
                if self.DEBUG:
                    print("Skipping raw device: " + raw_device)

        return True

    # Get device SMART health
    def get_smart_status(self, device: str) -> bool:
        command = "smartctl -H /dev/" + device + " | grep PASSED"
        if self.DEBUG:
            print("\nwatchaphdd:get_smartctl_info Running smartctl command: " + command)

        try:
            result_smart = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired as e:
            print("watchaphdd.get_smart_status: Exception: " + str(e))
            return False
        except subprocess.CalledProcessError as e:
            print("watchaphdd.get_smart_status: Exception: " + str(e))
            return False

        result_smart_stripped = result_smart.strip()
        pattern_smart_passed = re.compile('PASSED')
        match_smart_passed = pattern_smart_passed.search(result_smart_stripped)

        if match_smart_passed:
            return True
        else:
            return False

    def get_smart_details(self, device: str) -> str:
        command = "smartctl -i /dev/" + device + " | egrep 'Model|Serial|Firmware'"
        if self.DEBUG:
            print("\nwatchaphdd:get_smartctl_info Running smartctl command: " + command)

        try:
            result_details = subprocess.check_output(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired:
            raise ValueError('main.get_smart_details exception thrown subprocess.TimeoutExpired')
        except subprocess.CalledProcessError:
            raise ValueError('main.get_smart_details exception thrown subprocess.CalledProcessError')

        return str(result_details)

    # use smartctl to get device health and device detailed information like firmware version, etc.
    def get_smartctl_info(self) -> bool:
        for device in list(filter(None, self.raw_devices)):
            device_model = ""
            device_serial = ""
            device_firmware = ""
            smart_health_check_passed = False
            smart_result_details = ""
            smart_error = False

            if self.get_smart_status(device):
                smart_health_check_passed = True
            else:
                smart_health_check_passed = False
                smart_error = True

            # We only do a complete property check on the hard drive if it passes SMART health test
            # since we don't want to lock up the machine if the drive is dying
            if not smart_error:
                try:
                    smart_result_details = self.get_smart_details(device)
                except ValueError:
                    smart_error = True

            # If the device is in error state we mark all properties as being in error because we do not want to parse
            # details on a bad drive as it may lock up the system
            if smart_error or not smart_health_check_passed:
                device_model = "Error"
                device_serial = "Error"
                device_firmware = "Error"

                if self.DEBUG:
                    print("Device: " + device + " is in SMART Error")

                if device.strip() in self.device_results:
                    for key, value in self.device_results.items():
                        if key.strip() == device.strip():
                            # Switch to Fatal once we're sure no other weird device names show up
                            # self.device_results[device]["status"] = "ErrorFatal"
                            self.device_results[device]["status"] = "Warning"
                            self.device_results[device]["model"] = device_model
                            self.device_results[device]["serial"] = device_serial
                            self.device_results[device]["firmware"] = device_firmware
            else:
                if not smart_result_details:
                    device_model = "Error"
                    device_serial = "Error"
                    device_firmware = "Error"

                    if self.DEBUG:
                        print("Device: " + device + " is in SMART Error because it had no SMART results")

                else:
                    # ['Device Model:     SK hynix SH921 2.5 7MM 480GB', 'Serial Number:    1IE7T000110507B0P', Firmware Version: 10011L00', '']
                    raw_device_details = smart_result_details.split("\n")

                    for details in list(filter(None, raw_device_details)):
                        pattern_model = re.compile('Model')
                        pattern_serial = re.compile('Serial')
                        pattern_firmware = re.compile('Firmware')
                        match_model = pattern_model.search(details)
                        match_serial = pattern_serial.search(details)
                        match_firmware = pattern_firmware.search(details)

                        if match_model:
                            device_model = details.split(":")[1].strip()

                        if match_serial:
                            device_serial = details.split(":")[1].strip()

                        if match_firmware:
                            device_firmware = details.split(":")[1].strip()

                        if not match_model and not match_serial and not match_firmware:
                            device_model = "Error"
                            device_serial = "Error"
                            device_firmware = "Error"

                if device.strip() in self.device_results:
                    for key, value in self.device_results.items():
                        if key.strip() == device.strip():
                            if not smart_health_check_passed:
                                # Switch to Fatal once we're sure no other weird device names show up
                                # self.device_results[device]["status"] = "ErrorFatal"
                                self.device_results[device]["status"] = "Warning"
                            else:
                                self.device_results[device]["status"] = "Ok"

                            self.device_results[device]["model"] = device_model
                            self.device_results[device]["serial"] = device_serial
                            self.device_results[device]["firmware"] = device_firmware
                else:
                    return False

        return True

    # Sort all devices in an ordered dictionary and print out the results
    def print_device_results(self):
        try:
            print("\nDevice Results:")
            print(OrderedDict(sorted(self.device_results.items(), key=lambda x: '{0:0>8}'.format(x[1]['numeric_bus']).lower())).items())
        except Exception as e:
            print("watchaphdd.print_device_results: Exception: " + str(e))
            return False

        return True

    # Get partition usage information using psutil for the partitions we're interested in
    def get_partition_usage_data(self) -> dict:
        for id, value in self.partitions_to_monitor.items():
            partition = self.partitions_to_monitor[id]["partition"]

            try:
                percent_used = psutil.disk_usage(partition).percent
            except Exception as e:
                raise ValueError("watchaphdd.get_partition_usage_data using psutil.disk_usage().percent: Exception: " + str(e))

            # Round up free space calculation to the nearest %
            total_percent = float(100)
            percent_free = total_percent - math.ceil(percent_used)
            self.partitions_to_monitor[id]["percent_free"] = percent_free

            if self.DEBUG:
                print(partition + " used space: " + str(percent_used))
                print(partition + " free space: " + str(percent_free))

        return self.partitions_to_monitor

    # Store the partition data into the class dictionary
    def process_partition_usage_data(self) -> bool:
        try:
            self.partitions_to_monitor = self.get_partition_usage_data()
        except ValueError:
            return False

        return True

    # Sort all partitions in an ordered dictionary and print out all the partition usages
    def print_partition_usage_results(self):
        try:
            print("\nPartition Usage Results:")
            print(OrderedDict(sorted(self.partitions_to_monitor.items(), key=lambda x: x[1]['id'])).items())
        except Exception as e:
            print("watchaphdd.print_partition_usage_results: Exception: " + str(e))
            return False

        return True

    # Create DM report out of the information returned
    def build_dm_report(self):
        #
        # SECTION DEVICE HEALTH STATUS
        #
        # CO4PHY080101813,Ok,smartTst:[1:0:0:0],HDD disk [1:0:0:0] model 'TOSHIBA MG03ACA200' serial 'Y4R1KHM1F' firmware 'FL1H' is Ok
        for device, value in OrderedDict(sorted(self.device_results.items(), key=lambda x: '{0:0>8}'.format(x[1]['numeric_bus']).lower())).items():
            device_status_line = [self.hostname,
                                  self.device_results[device]["status"], "smartTst:" +
                                  self.device_results[device]["bus"],
                                  self.device_results[device]["type"] + " disk " +
                                  self.device_results[device]["bus"] + " model '" +
                                  self.device_results[device]["model"] + "' serial '" +
                                  self.device_results[device]["serial"] + "' firmware '" +
                                  self.device_results[device]["firmware"] + "' is " +
                                  self.device_results[device]["status"]]

            try:
                self.dm_output.write(",".join(device_status_line) + '\r\n')
            except Exception as e:
                print("watchaphdd.build_dm_report: Exception: " + str(e))
                return False

        #
        # SECTION PARTITION USAGE
        #
        status_partition_usage = "Ok"
        # CO4PHY080101813,Ok,diskpercentagefree,Drive / has 68 percent free disk space. Drive /ap has 94 percent free disk space. Drive /docker_storage has 58 percent free disk space.
        for key, value in OrderedDict(sorted(self.partitions_to_monitor.items(), key=lambda x: '{0:0>8}'.format(x[1]['id']).lower())).items():
            percent_free = int(self.partitions_to_monitor[key]["percent_free"])

            if percent_free <= 1:
                status_partition_usage = "Error"
            elif percent_free <= 10:
                status_partition_usage = "Warning"

        drive_usage_line = [self.hostname,
                            status_partition_usage,
                            "diskpercentagefree"]

        temp_drive_usage_string = ""
        for key, value in OrderedDict(sorted(self.partitions_to_monitor.items(), key=lambda x: '{0:0>8}'.format(x[1]['id']).lower())).items():
            partition = self.partitions_to_monitor[key]["partition"]
            percent_free = str(self.partitions_to_monitor[key]["percent_free"])
            temp_drive_usage_string += "Drive " + partition + " has " + percent_free + " percent free disk space. "

        drive_usage_line.append(temp_drive_usage_string)

        try:
            self.dm_output.write(",".join(drive_usage_line) + '\r\n')
        except Exception as e:
            print("watchaphdd.build_dm_report: Exception: " + str(e))
            return False

        return True

    # Make sure the DM report is properly formatted
    def verify_dm_report(self):
        report = self.dm_output.getvalue().split("\r\n")

        for line in list(filter(None, report)):
            results = line.split(",")

            # Each line must have 4 columns separated by commas
            # Format: CCPSCH010261732,ok,smartTst:2:0,HDD disk 2:0 model 'MB2000GDMTE' serial 'Z1X0G381' firmware 'HPGC' is OK
            if not len(results) == 4:
                print("watchaphdd.verify_dm_report: Number of expected columns is incorrect")
                return False

        return True

    # Write DM report to a tmp file
    def write_dm_report_to_disk(self) -> bool:
        if not os.path.isdir(self.watchdog_csv_dir):
            print("watchaphdd.write_dm_report_to_disk: Directory: " + self.watchdog_csv_dir + " not found")
            return False
        else:
            try:
                csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

                with open(csv_file_path, "w") as text_file:
                    text_file.write(self.dm_output.getvalue())

            except Exception as e:
                print("watchaphdd.write_dm_report_to_disk: Exception: " + str(e))
                return False

        return True

    # Verify that the tmp file on disk is the same as the data that was supposed to be written
    def verify_write_dm_report_to_disk(self) -> bool:
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)

        if os.path.exists(csv_file_path):
            with open(csv_file_path, 'r') as content_file:
                content = content_file.read()

                content = content.split("\n")
                dm_output = self.dm_output.getvalue().split("\r\n")

                if list(filter(None, content)) == list(filter(None, dm_output)):
                    return True
                else:
                    print("watchaphdd.verify_write_dm_report_to_disk: csv content does not match dm_output content")
                    return False

        else:
            return False

    # Atomically rename the .tmp file to a .csv file which will be ingested by machinelocalwatchdog.py
    def atomically_rename_dm_report_on_disk(self) -> bool:
        tmp_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_tmp_file)
        csv_file_path = os.path.join(self.watchdog_csv_dir, self.watchdog_csv_file)

        # print("\nRenaming: " + tmp_file_path + " to: " + csv_file_path)

        try:
            os.rename(tmp_file_path, csv_file_path)

        except Exception as e:
            print("watchaphdd.atomically_rename_dm_report_on_disk: Exception: " + str(e))
            return False

        return True

    # Print out the DM report
    def print_dm_report(self):

        output = "\n".join(self.dm_output.getvalue().split("\r\n"))
        self.dm_output.close()

        try:
            print("\nDM Report:")
            print(output)
        except Exception as e:
            print("watchaphdd.print_dm_report: Exception: " + str(e))
            return False

        return True

    # Execute all the methods in order
    def main(self):
        if not self.get_block_devices():
            print("watchaphdd.main: Could not discover block devices (main.get_block_devices)")
            sys.exit(1)

        if not self.process_block_devices():
            print("watchaphdd.main: Could not process block devices (main.process_block_devices)")
            sys.exit(1)

        if not self.get_types_for_block_devices():
            print("watchaphdd.main: Could not get block device types (main.get_types_for_block_devices)")
            sys.exit(1)

        if not self.process_device_bus_data():
            print("watchaphdd.main: Could not process lsscsi and blkid information (main.process_device_bus_data)")
            sys.exit(1)

        if not self.get_smartctl_info():
            print("watchaphdd.main: Could not get smartctl information (main.get_smartctl_info)")
            sys.exit(1)

        if not self.process_partition_usage_data():
            print("watchaphdd.main: Could not get partition usage data (main.process_partition_usage_data)")
            sys.exit(1)

        if self.DEBUG:
            if not self.print_device_results():
                print("watchaphdd.main: Could not print device results (main.print_device_results)")
                sys.exit(1)

            if not self.print_partition_usage_results():
                print("watchaphdd.main: Could not print partition usage results (main.print_partition_usage_results)")
                sys.exit(1)

        if not self.build_dm_report():
            print("watchaphdd.main: Could not build DM report (main.build_dm_report)")
            sys.exit(1)

        if not self.verify_dm_report():
            print("watchaphdd.main: Could not verify DM report (main.verify_dm_report)")
            sys.exit(1)

        if not self.write_dm_report_to_disk():
            print("watchaphdd.main: Could not write DM report to disk (main.write_dm_report_to_disk)")
            sys.exit(1)

        if not self.verify_write_dm_report_to_disk():
            print("watchaphdd.main: Could not verify DM report written to disk (main.verify_write_dm_report_to_disk)")
            sys.exit(1)

        if not self.atomically_rename_dm_report_on_disk():
            print("watchaphdd.main: Could not atomically rename dm report file on disk (main.atomically_rename_dm_report_on_disk)")
            sys.exit(1)

        if self.DEBUG:
            if not self.print_dm_report():
                print("watchaphdd.main: Could not print DM report (main.print_dm_report)")
                sys.exit(1)

        sys.exit(0)  # EX_OK = everything executed normally


if __name__ == '__main__':
    watchdog_csv_dir = "/ap/data/watchdogoutput/csv_sources"
    watchdog_tmp_file = "watchaphdd.tmp"
    watchdog_csv_file = "watchaphdd.csv"

    WatchAPHDD(watchdog_csv_dir, watchdog_tmp_file, watchdog_csv_file).main()
