#!/usr/bin/env python3

# Watchdog used to check antimalware last scan and signature age status then report on it

import sys
import os
import io
import socket
import subprocess
import datetime
from dateutil import tz

try:
    from lxml import etree
except ImportError:
    try:
        import xml.etree.cElementTree as etree
    except ImportError:
        import xml.etree.ElementTree as etree


class WatchAntimalware(object):
    def __init__(self,
                 azsecpack_lastscan_file: str="/etc/azsec/lastScan.xml",
                 signature_file: str="/var/lib/azsec-clamav/daily.cvd",
                 report_directory: str="/ap/data/watchdogoutput/csv_sources"):
        self.azsecpack_lastscan_file = azsecpack_lastscan_file
        self.signature_file = signature_file
        self.report_directory = report_directory

        self.watchdog_name = "watchantimalware"
        self.watchdog_tmp_file = self.watchdog_name + ".tmp"
        self.watchdog_csv_file = self.watchdog_name + ".csv"

        # Path of prior clamav signature file before azsecpack moved it
        self.old_signature_file = "/var/lib/clamav/daily.cvd"

        # Fields to be set during runtime
        self.clamav_version_raw = ""
        self.azsecpack_lastscan_raw = ""
        self.hostname = ""

        # Report property values
        self.scan_begin = ""
        self.scan_end = ""
        self.scan_status = ""
        self.signature_version = ""
        self.signature_applied = ""
        self.signature_created = ""
        self.engine = ""
        self.engine_version = ""

    # Get most of the metadata used in the report
    def get_clamav_version(self):
        command = "clamscan -V"
        try:
            result = subprocess.check_output(command, timeout=3, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired:
            raise Exception("WatchAntimalware.get_clamav_version: '{0}' timed out".format(command))
        except subprocess.CalledProcessError:
            raise Exception("WatchAntimalware.get_clamav_version: error running '{0}'".format(command))

        # ClamAV 0.99.2/23229/Wed Mar 22 21:27:26 2017
        self.clamav_version_raw = result.strip()

    def parse_clamav_version(self):
        if not self.clamav_version_raw:
            return

        # split parts
        #   ClamAV 0.99.2
        #   23229
        #   Wed Mar 22 21:27:26 2017
        parts = self.clamav_version_raw.split("/")

        # engine
        engine = parts[0].split(" ")
        self.engine = engine[0]
        if len(engine) > 1:
            self.engine_version = engine[1]

        # signature
        if len(parts) > 1:
            self.signature_version = parts[1]
        if len(parts) > 2:
            try:
                dt = datetime.datetime.strptime(parts[2], "%a %b %d %H:%M:%S %Y")
            except ValueError as e:
                raise Exception("WatchAntimalware.parse_clamav_version: unexpected date format: {0}".format(e))

            # convert to utc
            utc = dt.replace(tzinfo=tz.tzlocal()).astimezone(tz.tzutc())
            self.signature_created = utc.strftime("%Y-%m-%d %H:%M:%S")

    def read_azsecpack_lastscan_file(self):
        if not os.path.isfile(self.azsecpack_lastscan_file):
            raise Exception("WatchAntimalware.read_azsecpack_lastscan_file: not a file")

        try:
            with open(self.azsecpack_lastscan_file, "rb") as f:
                self.azsecpack_lastscan_raw = f.read()
        except Exception:
            raise Exception("WatchAntimalware.read_azsecpack_lastscan_file: error reading file")

    def parse_azsecpack_lastscan_file(self):
        if not self.azsecpack_lastscan_raw:
            self.scan_status = "Never scanned"
            return

        try:
            xml = etree.XML(self.azsecpack_lastscan_raw)
        except Exception:
            raise Exception("WatchAntimalware.parse_azsecpack_lastscan_file: error parsing xml")

        if xml is None or len(xml) < 1 or xml.tag != "last_scans":
            raise Exception("WatchAntimalware.parse_azsecpack_lastscan_file: unsupported format")

        scan_tag = None
        for scan in xml.findall("scan[@name='clamav']"):
            scan_tag = scan

        if scan_tag is None:
            self.scan_status = "Never scanned"
            return

        if "start_time" not in scan_tag.attrib or "end_time" not in scan_tag.attrib:
            raise Exception("WatchAntimalware.parse_azsecpack_lastscan_file: missing fields")

        clamav_start_time = (scan_tag.attrib["start_time"])[:(scan_tag.attrib["start_time"]).find(".")]
        clamav_end_time = (scan_tag.attrib["end_time"])[:(scan_tag.attrib["end_time"]).find(".")]

        # determine if scan is still running
        is_running = False
        if "pid" in scan_tag.attrib:
            is_running = os.path.exists("/proc/{0}".format(scan_tag.attrib["pid"]))

        # determine if scan failed
        is_failed = False
        if "failed" in scan_tag.attrib and scan_tag.attrib["failed"] == "true":
            is_failed = True

        try:
            dt_start = datetime.datetime.strptime(clamav_start_time, "%Y-%m-%dT%H:%M:%S")
            dt_end = datetime.datetime.strptime(clamav_end_time, "%Y-%m-%dT%H:%M:%S")
        except ValueError as e:
            raise Exception("WatchAntimalware.parse_clamav_version: unexpected date format: {0}".format(e))

        # convert datetime to utc
        utc_start = dt_start.replace(tzinfo=tz.tzlocal()).astimezone(tz.tzutc())
        utc_end = dt_end.replace(tzinfo=tz.tzlocal()).astimezone(tz.tzutc())
        self.scan_begin = utc_start.strftime("%Y-%m-%d %H:%M:%S")
        self.scan_end = utc_end.strftime("%Y-%m-%d %H:%M:%S")

        if is_running:
            self.scan_status = "Running"
        elif is_failed:
            self.scan_status = "Failed"
        elif scan_tag.attrib["end_time"] == "0001-01-01T00:00:00Z":
            self.scan_status = "Aborted"
        else:
            self.scan_status = "Finished"

    # Get the date when signatures where downloaded
    def get_signatures_applied(self):
        if os.path.isfile(self.signature_file):
            prefer_signature_file = self.signature_file
        elif os.path.isfile(self.old_signature_file):
            prefer_signature_file = self.old_signature_file
        else:
            raise Exception("WatchAntimalware.get_signatures_applied: signature file not found")

        try:
            t = os.path.getmtime(prefer_signature_file)
            dt = datetime.datetime.fromtimestamp(t)
        except Exception as e:
            raise Exception("WatchAntimalware.get_signatures_applied: {0}".format(e))

        # convert datetime to utc
        utc = dt.replace(tzinfo=tz.tzlocal()).astimezone(tz.tzutc())
        self.signature_applied = utc.strftime("%Y-%m-%d %H:%M:%S")

    # Create DM report out of the information returned
    def build_dm_report(self):
        self.dm_output = io.StringIO()
        if not self.hostname:
            try:
                self.hostname = socket.gethostname()
            except Exception as e:
                print("build_dm_report: Exception: " + str(e))
                return False

        monitored_items = {}
        monitored_items["ApAntimalware_HB_AVScanBegin"] = self.scan_begin
        monitored_items["ApAntimalware_HB_AVScanEnd"] = self.scan_end
        monitored_items["ApAntimalware_HB_AVScanStatus"] = self.scan_status
        monitored_items["ApAntimalware_HB_AVSignatureBaseVersion"] = self.signature_version
        monitored_items["ApAntimalware_HB_AVSignaturesAppliedUTC"] = self.signature_applied
        monitored_items["ApAntimalware_HB_AVSignaturesCreatedUTC"] = self.signature_created
        monitored_items["ApAntimalware_HB_AVSignatureVersion"] = self.signature_version
        monitored_items["ApAntimalware_HB_Engine"] = self.engine
        monitored_items["ApAntimalware_HB_EngineVersion"] = self.engine_version

        for key, value in monitored_items.items():
            status = "Ok"
            if not value:
                status = "Warning"

            line = [self.hostname, status, key, value]

            try:
                self.dm_output.write(",".join(line) + '\r\n')
            except Exception as e:
                print("build_dm_report: Exception: " + str(e))
                return False

        return True

    # Make sure the DM report is properly formatted
    def verify_dm_report(self):
        report = self.dm_output.getvalue().split("\r\n")

        for line in list(filter(None, report)):
            results = line.split(",")

            # Each line must have 4 columns separated by commas
            # Format: CCPSCH010261732,ok,smartTst:2:0,HDD disk 2:0 model 'MB2000GDMTE' serial 'Z1X0G381' firmware 'HPGC' is OK
            if not len(results) == 4:
                print("verify_dm_report: Number of expected columns is incorrect")
                return False

            if results[1] not in {"Ok", "Warning", "Error", "Information"}:
                print("verify_dm_report: Invalid status")
                return False

        return True

    # Write DM report to a tmp file
    def write_dm_report_to_disk(self) -> bool:
        if not os.path.isdir(self.report_directory):
            print("write_dm_report_to_disk: Directory: " + self.report_directory + " not found")
            return False

        tmp_file_path = os.path.join(self.report_directory, self.watchdog_tmp_file)

        try:
            with open(tmp_file_path, "w") as text_file:
                text_file.write(self.dm_output.getvalue())
        except Exception as e:
            print("write_dm_report_to_disk: Exception: " + str(e))
            return False

        return True

    # Verify that the tmp file on disk is the same as the data that was supposed to be written
    def verify_write_dm_report_to_disk(self) -> bool:
        csv_file_path = os.path.join(self.report_directory, self.watchdog_tmp_file)

        if not os.path.exists(csv_file_path):
            return False

        dm_output = self.dm_output.getvalue().split("\r\n")
        with open(csv_file_path, 'r') as content_file:
            content = content_file.read().split("\n")

        if list(filter(None, content)) != list(filter(None, dm_output)):
            print("verify_write_dm_report_to_disk: csv content does not match dm_output content")
            return False

        return True

    # Atomically rename the .tmp file to a .csv file which will be ingested by machinelocalwatchdog.py
    def atomically_rename_dm_report_on_disk(self) -> bool:
        tmp_file_path = os.path.join(self.report_directory, self.watchdog_tmp_file)
        csv_file_path = os.path.join(self.report_directory, self.watchdog_csv_file)

        try:
            os.rename(tmp_file_path, csv_file_path)
        except Exception as e:
            print("atomically_rename_dm_report_on_disk: Exception: " + str(e))
            return False

        return True

    # Print out the DM report
    def print_dm_report(self):

        output = "\n".join(self.dm_output.getvalue().split("\r\n"))
        self.dm_output.close()

        try:
            print("\nDM Report:")
            print(output)
        except Exception as e:
            print("print_dm_report: Exception: " + str(e))
            return False

        return True

    # Execute all the methods in order
    def main(self):
        error_code = 0  # EX_OK = everything executed normally

        try:
            self.get_clamav_version()
            self.parse_clamav_version()
        except Exception as e:
            print("WatchAntimalware.main: {0}".format(e))
            error_code = 1

        try:
            self.read_azsecpack_lastscan_file()
            self.parse_azsecpack_lastscan_file()
        except Exception as e:
            print("WatchAntimalware.main: {0}".format(e))
            error_code = 1

        try:
            self.get_signatures_applied()
        except Exception as e:
            print("WatchAntimalware.main: {0}".format(e))
            error_code = 1

        if not self.build_dm_report():
            print("WatchAntimalware.main: Could not build DM report (main.build_dm_report)")
            sys.exit(1)

        if not self.verify_dm_report():
            print("WatchAntimalware.main: Could not verify DM report (main.verify_dm_report)")
            sys.exit(1)

        if not self.write_dm_report_to_disk():
            print("WatchAntimalware.main: Could not write DM report to disk (main.write_dm_report_to_disk)")
            sys.exit(1)

        if not self.atomically_rename_dm_report_on_disk():
            print("WatchAntimalware.main: Could not atomically rename dm report file on disk (main.atomically_rename_dm_report_on_disk)")
            sys.exit(1)

        sys.exit(error_code)

if __name__ == '__main__':
    WatchAntimalware().main()
