#!/usr/bin/env python3

# Service to Update DNS and NTP entries based on available PS servers

import sys
import os
import tempfile
import subprocess
import dns.resolver

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from apsdk import apenv


class APDNSNTPUpdater(object):
    DEBUG = False
    dns_tmp_file_name = tempfile.mkstemp(suffix='', prefix='tmp', dir=None, text=False)[1]
    ntp_tmp_file_name = tempfile.mkstemp(suffix='', prefix='tmp', dir=None, text=False)[1]

    def __init__(self,
                 dns_cluster_name: str,
                 dns_configuration_file: str="/etc/dnsmasq.d/aplinux.conf",
                 dns_configuration_tmp_file: str=dns_tmp_file_name,
                 ntp_configuration_file: str="/etc/ntp/ap.conf",
                 ntp_configuration_tmp_file: str=ntp_tmp_file_name,
                 dns_prefix_domain: str="all.ps.autopilot",
                 dns_suffix_domain: str="ap.gbl"):

        # DNS and NTP information
        dns_fqdn_array = []
        dns_fqdn_array.append(dns_prefix_domain)
        dns_fqdn_array.append(dns_cluster_name.lower())
        dns_fqdn_array.append(dns_suffix_domain)
        self.dns_fqdn = ".".join(dns_fqdn_array)

        if self.DEBUG:
            print("\nFQDN:")
            print(self.dns_fqdn)

        self.dns_configuration_file = dns_configuration_file
        self.dns_configuration_tmp_file = dns_configuration_tmp_file
        self.ntp_configuration_file = ntp_configuration_file
        self.ntp_configuration_tmp_file = ntp_configuration_tmp_file

        # Create instance of the dnspython class
        self.DNSResolver = dns.resolver.Resolver()

        # Existing Server list
        self.existing_dns_servers = []
        # New Server list
        self.new_dns_server_list = []

    # Get the existing DNS servers used by the system
    def get_existing_dns_servers(self) -> list:
        # Example contents of DNS configuration file
        # server=25.77.192.102
        # server=25.77.193.123
        # server=25.77.193.141
        # server=25.77.193.57
        dns_config_lines = []
        if os.path.isfile(self.dns_configuration_file):
            try:
                with open(self.dns_configuration_file, 'r') as f:
                    dns_config_lines = f.readlines()
            except Exception as e:
                raise ValueError("apdnsntpupdate:get_existing_dns_servers Exception: " + str(e))
        else:
            raise ValueError("apdnsntpupdate:get_existing_dns_servers - " + self.dns_configuration_file + " Not Found")

        # Use list comprehension to remove all empty items in list ("if item" returns False if empty)
        dns_config_lines[:] = [item for item in dns_config_lines if item]
        # Use list comprehension to strip "server=" from all lines
        dns_config_lines[:] = [s.strip("server=") for s in dns_config_lines]
        # Use list comprehension to strip line endings from all lines
        dns_config_lines[:] = [s.strip() for s in dns_config_lines]

        if self.DEBUG:
            print("\nExisting DNS Configuration:")
            print(dns_config_lines)

        return dns_config_lines

    # Query all PS server IPs from DNS
    def get_all_ps_server_ips(self) -> list:
        all_functional_ps_servers = []

        try:
            answers = self.DNSResolver.query(self.dns_fqdn, 'A')
        except Exception as e:
            raise ValueError("apdnsntpupdate:get_all_ps_server_ips Exception: " + str(e))

        if len(answers) > 0:
            for server in answers:
                all_functional_ps_servers.append(str(server))
        else:
            raise ValueError("apdnsntpupdate:get_all_ps_server_ips No IP Addresses Returned by DNS Query")

        if self.DEBUG:
            print("\nPS Servers Obtained from DNS:")
            print(all_functional_ps_servers)

        return all_functional_ps_servers

    # Update the list of servers with the new servers obtained from the DNS lookup
    def update_dns_server_list(self) -> bool:
        # Get the list of existing configured DNS servers
        new_server_list = self.get_all_ps_server_ips()
        self.existing_dns_servers = self.get_existing_dns_servers()

        # We only update the list if the new list is different (we can't tell if different is better (more servers) or
        # not so we try to extend the list here and then later we check again before writing anything to disk)
        if set(new_server_list) != set(self.existing_dns_servers):
            if self.DEBUG:
                print("\nMerging old DNS and new DNS entries since the sets are not equal")

            # Use list comprehension to compare all ps servers we got in the new server list to what is in the previous
            # server list. If new servers are found then we extend the new_server_list with values from the
            # previous_server_list.  This allows us to ensure that the order is always arranged according to the values
            # returned by DNS currently.  Any other old servers that are non-functional at the moment are put at the back
            # of the list as part of the extension
            new_server_list.extend(x for x in self.existing_dns_servers if x not in new_server_list)

        if self.DEBUG:
            print("\nMerged PS Server List:")
            print(new_server_list)

        self.new_dns_server_list = new_server_list

        return True

    # Write temporary files to disk to make sure we have success before we overwrite the old configs
    def write_tmp_files_to_disk(self) -> bool:
        if self.DEBUG:
            print("\nWriting: " + self.dns_configuration_tmp_file)
        # DNS format is:
        # server=25.77.192.102
        # server=25.77.193.123
        # server=25.77.193.141
        # server=25.77.193.57
        try:
            with open(self.dns_configuration_tmp_file, "w") as dns_file:
                for server in self.new_dns_server_list:
                    dns_file.write("server=" + server + "\n")
        except Exception as e:
            raise ValueError("apdnsntpupdate:write_tmp_files_to_disk Exception: " + str(e))

        if self.DEBUG:
            print("\nWriting: " + self.ntp_configuration_tmp_file)

        # NTP format is:
        # server 25.77.192.102
        # server 25.77.193.123
        # server 25.77.193.141
        # server 25.77.193.57
        try:
            with open(self.ntp_configuration_tmp_file, "w") as ntp_file:
                for server in self.new_dns_server_list:
                    ntp_file.write("server " + server + "\n")
        except Exception as e:
            raise ValueError("apdnsntpupdate:write_tmp_files_to_disk Exception: " + str(e))

        return True

    # Atomically rename the .tmp files to the final files
    def atomically_rename_dns_and_ntp_configs(self) -> bool:
        if os.path.isfile(self.dns_configuration_file):
            if self.DEBUG:
                print("\nRenaming: " + self.dns_configuration_tmp_file + " to: " + self.dns_configuration_file)
            try:
                os.rename(self.dns_configuration_tmp_file, self.dns_configuration_file)
            except Exception as e:
                raise ValueError("apdnsntpupdate.atomically_rename_dns_and_ntp_configs: Exception: " + str(e))
        else:
            print("\nFile: " + self.dns_configuration_file + " does not exist!")
            return False

        if os.path.isfile(self.ntp_configuration_file):
            if self.DEBUG:
                print("\nRenaming: " + self.ntp_configuration_tmp_file + " to: " + self.ntp_configuration_file)
            try:
                os.rename(self.ntp_configuration_tmp_file, self.ntp_configuration_file)
            except Exception as e:
                raise ValueError("apdnsntpupdate.atomically_rename_dns_and_ntp_configs: Exception: " + str(e))
        else:
            print("\nFile: " + self.ntp_configuration_file + " does not exist!")
            return False

        return True

    # Reload dnsmasq to pick up DNS server changes
    def reload_dnsmasq_service(self) -> bool:
        command = "systemctl reload dnsmasq.service"

        if self.DEBUG:
            print("\nwatchaphdd:reload_dnsmasq_service Command: " + command)

        try:
            subprocess.run(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired:
            raise ValueError('watchaphdd:reload_dnsmasq_service subprocess.TimeoutExpired')
        except subprocess.CalledProcessError:
            raise ValueError('watchaphdd:reload_dnsmasq_service subprocess.CalledProcessError')

        return True

    # Restart ntpd to pick up NTP server changes
    def restart_ntp_service(self) -> bool:
        command = "systemctl restart ntp.service"

        if self.DEBUG:
            print("\nwatchaphdd:restart_ntp_service Command: " + command)

        try:
            subprocess.run(command, timeout=10, universal_newlines=True, shell=True)
        except subprocess.TimeoutExpired:
            raise ValueError('watchaphdd:restart_ntp_service subprocess.TimeoutExpired')
        except subprocess.CalledProcessError:
            raise ValueError('watchaphdd:restart_ntp_service subprocess.CalledProcessError')

        return True

    # Execute all the methods in order
    def main(self):
        if not self.update_dns_server_list():
            print("apdnsntpupdate.main: Could not get OS Version (main.update_dns_server_list)")
            sys.exit(1)

        # If nothing changed in the server list after we tried to add new servers there's no need to write
        # If the DNS server list on disk has more entries than the one being queried from live DNS then we will
        # try to add to the list but with no changes; hence this condition
        if set(self.new_dns_server_list) != set(self.existing_dns_servers):
            if self.DEBUG:
                print("\nChanging DNS configurations on disk since the sets are changed")

            if not self.write_tmp_files_to_disk():
                print("apdnsntpupdate.main: Could not write tmp files to disk (main.write_tmp_files_to_disk)")
                sys.exit(1)

            if not self.atomically_rename_dns_and_ntp_configs():
                print("apdnsntpupdate.main: Could not atomically rename .tmp files to final files (main.atomically_rename_dns_and_ntp_configs)")
                sys.exit(1)

            if not self.reload_dnsmasq_service():
                print("apdnsntpupdate.main: Could not reload dnsmasq service (main.reload_dnsmasq_service)")
                sys.exit(1)

            if not self.restart_ntp_service():
                print("apdnsntpupdate.main: Could not restart ntp service (main.restart_ntp_service)")
                sys.exit(1)

        sys.exit(0)  # EX_OK = everything executed normally


if __name__ == '__main__':
    dns_cluster_name = apenv.get_cluster_name()

    APDNSNTPUpdater(dns_cluster_name).main()
